//##################################################################################################
//
//  Ceetron Desktop Components
//  QtDemoApp
//  
//  --------------------------------------------------------------------------------------------
//  Copyright (C) 2014, Ceetron AS
//  This is UNPUBLISHED PROPRIETARY SOURCE CODE of Ceetron AS. The contents of this file may 
//  not be disclosed to third parties, copied or duplicated in any form, in whole or in part,
//  without the prior written permission of Ceetron AS.
//##################################################################################################

#include "CuttingPlanesWidget.h"

#include "ui_CuttingPlanesWidget.h"

#include "CeeCore/BoundingBox.h"
#include "CeeVisualization/View.h"

#include "CeeUnstructGrid/CuttingPlane.h"
#include "CeeUnstructGrid/DataSource.h"
#include "CeeUnstructGrid/DataSourceDirectory.h"
#include "CeeUnstructGrid/ModelSpec.h"
#include "CeeUnstructGrid/PartSettings.h"
#include "CeeUnstructGrid/PartSettingsIterator.h"
#include "CeeUnstructGrid/ResultInfo.h"
#include "CeeUnstructGrid/UnstructGridModel.h"

#include "ceeqtUtilsCore.h"

#include <QItemSelectionModel>
#include <QListView>
#include <QModelIndex>
#include <QStringListModel>


//==================================================================================================
/// 
/// Manages cutting planes
/// 
//==================================================================================================

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
CuttingPlanesWidget::CuttingPlanesWidget(QWidget* parent, Qt::WindowFlags f)
:   QWidget(parent, f),
    m_ui(new Ui::CuttingPlanesWidget),
    m_listModel(new QStringListModel(this)),
    m_nextId(-1)
{
    m_ui->setupUi(this);
    m_ui->cuttingPlanesListView->setModel(m_listModel);

    connect(m_ui->cuttingPlanesListView->selectionModel(), 
        SIGNAL(currentRowChanged(const QModelIndex&, const QModelIndex&)), SLOT(applySelection(const QModelIndex&, const QModelIndex&)));
    connect(m_ui->addCuttingPlanePushButton,    SIGNAL(clicked()),                   SLOT(addCuttingPlane()));
    connect(m_ui->removeCuttingPlanePushButton, SIGNAL(clicked()),                   SLOT(removeCuttingPlane()));
    connect(m_ui->visibleCheckBox,              SIGNAL(clicked(bool)),               SLOT(setCuttingPlaneVisible(bool)));
    connect(m_ui->positionLineEdit,             SIGNAL(editingFinished()),           SLOT(applyPosition()));
    connect(m_ui->normalLineEdit,               SIGNAL(editingFinished()),           SLOT(applyNormal()));
    connect(m_ui->mappedScalarResultComboBox,   SIGNAL(currentIndexChanged(int)),    SLOT(setMappedScalarResult(int)));
    connect(m_ui->mappedVectorResultComboBox,   SIGNAL(currentIndexChanged(int)),    SLOT(setMappedVectorResult(int)));

    // Make sure the group box title is disabled if the group box itself is disabled
    QPalette palette;
    palette.setColor(QPalette::Disabled, QPalette::WindowText, QApplication::palette().color(QPalette::Disabled, QPalette::WindowText));
    m_ui->settingsGroupBox->setPalette(palette);

    updatePanel();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
CuttingPlanesWidget::~CuttingPlanesWidget()
{
    delete m_listModel;
}


//--------------------------------------------------------------------------------------------------
/// Set up widget with current meta data and selection from model and view
//--------------------------------------------------------------------------------------------------
void CuttingPlanesWidget::setup(cee::vis::View* view, cee::ug::UnstructGridModel* model)
{
    m_unstructModel = model;
    m_view = view;

    const cee::ug::DataSourceDirectory* directory = m_unstructModel->dataSource()->directory();
    Q_ASSERT(directory);

    // Get scalar results meta data from the models data source
    {
        m_ui->mappedScalarResultComboBox->blockSignals(true);
        m_ui->mappedScalarResultComboBox->clear();

        // Add all available scalar results to scalar combo box
        // Add - none - item to deselect all results
        m_ui->mappedScalarResultComboBox->addItem(tr("- none -"), -1);

        std::vector<cee::ug::ResultInfo> scalarInfos = directory->scalarResultInfos();
        std::vector<cee::ug::ResultInfo>::const_iterator it;
        for (it = scalarInfos.begin(); it != scalarInfos.end(); ++it)
        {
            m_ui->mappedScalarResultComboBox->addItem(cee::qt::UtilsCore::toQString(it->name()), it->id());
        }

        m_ui->mappedScalarResultComboBox->blockSignals(false);
    }

    // Get vector results meta data from the models data source
    {
        m_ui->mappedVectorResultComboBox->blockSignals(true);
        m_ui->mappedVectorResultComboBox->clear();

        // Add all available vector results to vector combo box
        // Add - none - item to deselect all results
        m_ui->mappedVectorResultComboBox->addItem(tr("- none -"), -1);

        std::vector<cee::ug::ResultInfo> vectorInfos = directory->vectorResultInfos();
        std::vector<cee::ug::ResultInfo>::const_iterator it;
        for (it = vectorInfos.begin(); it != vectorInfos.end(); ++it)
        {
            m_ui->mappedVectorResultComboBox->addItem(cee::qt::UtilsCore::toQString(it->name()), it->id());
        }

        m_ui->mappedVectorResultComboBox->blockSignals(false);
    }

    // Populate with all cutting planes from the model
    {
        m_nextId = 1;
        size_t cuttingPlaneCount = m_unstructModel->cuttingPlaneCount();

        QStringList stringList;
        for (size_t i = 0; i < cuttingPlaneCount; ++i)
        {
            stringList += tr("Cutting plane %1").arg(m_nextId++);
        }
        m_listModel->setStringList(stringList);
    }

    updatePanel();
}


//--------------------------------------------------------------------------------------------------
/// A cutting plane was selected from outside the panel. (For instance by picking.)
/// Select cutting plane in list box and update GUI.
//--------------------------------------------------------------------------------------------------
void CuttingPlanesWidget::selectCuttingPlane(size_t cuttingPlaneIndex)
{
    QModelIndex index;
    if (cuttingPlaneIndex != cee::UNDEFINED_SIZE_T)
    {
        index = m_listModel->index(static_cast<int>(cuttingPlaneIndex));
    }

    m_ui->cuttingPlanesListView->setCurrentIndex(index);
}


//--------------------------------------------------------------------------------------------------
/// Update panel settings.
/// Clear if no cutting plane is available/selected.
//--------------------------------------------------------------------------------------------------
void CuttingPlanesWidget::updatePanel()
{
    if (m_unstructModel)
    {
        setEnabled(true);

        int cuttingPlaneIndex = m_ui->cuttingPlanesListView->currentIndex().row();
        if (cuttingPlaneIndex != -1)
        {
            cee::ug::CuttingPlane* cuttingPlane = m_unstructModel->cuttingPlane(cuttingPlaneIndex);

            m_ui->visibleCheckBox->setChecked(cuttingPlane->visible());
            m_ui->positionLineEdit->setText(QString("%1 %2 %3").arg(cuttingPlane->point().x()).arg(cuttingPlane->point().y()).arg(cuttingPlane->point().z()));
            m_ui->normalLineEdit->setText(QString("%1 %2 %3").arg(cuttingPlane->normal().x()).arg(cuttingPlane->normal().y()).arg(cuttingPlane->normal().z()));

            m_ui->mappedScalarResultComboBox->blockSignals(true);
            m_ui->mappedScalarResultComboBox->setCurrentIndex(m_ui->mappedScalarResultComboBox->findData(cuttingPlane->mapScalarResultId()));
            m_ui->mappedScalarResultComboBox->blockSignals(false);

            m_ui->mappedVectorResultComboBox->blockSignals(true);
            m_ui->mappedVectorResultComboBox->setCurrentIndex(m_ui->mappedVectorResultComboBox->findData(cuttingPlane->mapVectorResultId()));
            m_ui->mappedVectorResultComboBox->blockSignals(false);

            // Check to see if we have precomputed cutting planes. 
            // These cannot be deleted, moved or modified.
            // Disable GUI according to this.
            bool isPrecomputed = cuttingPlane->isPrecomputed();
            m_ui->removeCuttingPlanePushButton->setEnabled(!isPrecomputed);
            m_ui->mappedScalarResultComboBox->setEnabled(!isPrecomputed);
            m_ui->mappedVectorResultComboBox->setEnabled(!isPrecomputed);
            m_ui->positionLineEdit->setEnabled(!isPrecomputed);
            m_ui->normalLineEdit->setEnabled(!isPrecomputed);

            m_ui->settingsGroupBox->setEnabled(true);
        }
        else
        {
            m_ui->removeCuttingPlanePushButton->setEnabled(false);

            m_ui->visibleCheckBox->setChecked(false);
            m_ui->positionLineEdit->setText("");
            m_ui->normalLineEdit->setText("");

            m_ui->mappedScalarResultComboBox->blockSignals(true);
            m_ui->mappedScalarResultComboBox->setCurrentIndex(0);
            m_ui->mappedScalarResultComboBox->blockSignals(true);

            m_ui->mappedVectorResultComboBox->blockSignals(true);
            m_ui->mappedVectorResultComboBox->setCurrentIndex(0);
            m_ui->mappedVectorResultComboBox->blockSignals(false);

            m_ui->settingsGroupBox->setEnabled(false);
        }
    }
    else
    {
        setEnabled(false);

        m_ui->visibleCheckBox->setChecked(false);
        m_ui->positionLineEdit->setText("");
        m_ui->normalLineEdit->setText("");

        m_ui->mappedScalarResultComboBox->blockSignals(true);
        m_ui->mappedScalarResultComboBox->setCurrentIndex(0);
        m_ui->mappedScalarResultComboBox->blockSignals(false);

        m_ui->mappedVectorResultComboBox->blockSignals(true);
        m_ui->mappedVectorResultComboBox->setCurrentIndex(0);
        m_ui->mappedVectorResultComboBox->blockSignals(false);
    }
}


//--------------------------------------------------------------------------------------------------
/// Update widgets to show state of selected cutting plane, if any.
/// Emits a signal to tell other that selection has changed and update visualization is needed.
//--------------------------------------------------------------------------------------------------
void CuttingPlanesWidget::applySelection(const QModelIndex& current, const QModelIndex& previous)
{
    if (previous.isValid())
    {
        m_unstructModel->cuttingPlane(static_cast<size_t>(previous.row()))->setHighlight(false);
    }

    if (current.isValid())
    {
        m_unstructModel->cuttingPlane(static_cast<size_t>(current.row()))->setHighlight(true);
        emit selected();
    }

    updatePanel();
}


//--------------------------------------------------------------------------------------------------
/// Add a new cutting plane
//--------------------------------------------------------------------------------------------------
void CuttingPlanesWidget::addCuttingPlane()
{
    cee::PtrRef<cee::ug::CuttingPlane> cuttingPlane = new cee::ug::CuttingPlane();

    // Set initial position of the cutting plane
    cee::BoundingBox boundingbox = m_unstructModel->boundingBox();
    cuttingPlane->setPoint(boundingbox.center());
    cuttingPlane->setNormal(cee::Vec3d(0, 1, 0));

    // Use loaded scalar result (if any) as default on cutting plane
    int fringesResultId = m_unstructModel->modelSpec().fringesResultId();
    if (fringesResultId != -1)
    {
        cuttingPlane->setMapScalarResultId(fringesResultId);
    }

    // Use loaded vector result (if any) as default on cutting plane
    if (m_unstructModel->modelSpec().vectorResultIds().size() > 0) 
    {
        int vectorResultId = m_unstructModel->modelSpec().vectorResultIds()[0];
        cuttingPlane->setMapVectorResultId(vectorResultId);
    }

    // Add the cutting plane to the model and listview
    m_unstructModel->addCuttingPlane(cuttingPlane.get());
    m_listModel->insertRow(m_listModel->rowCount());
    QModelIndex modelIndex = m_listModel->index(m_listModel->rowCount()-1);
    m_listModel->setData(modelIndex, tr("Cutting plane %1").arg(m_nextId++));

    // Set all parts to outline draw style to see cutting plane
    cee::ug::PartSettingsIterator it(m_unstructModel.get());
    while (it.hasNext())
    {
        cee::ug::PartSettings* settings = it.next();
        settings->setDrawStyle(cee::ug::PartSettings::OUTLINE);
    }

    m_ui->cuttingPlanesListView->setCurrentIndex(modelIndex);

    m_unstructModel->updateVisualization();
    m_view->requestRedraw();
}


//--------------------------------------------------------------------------------------------------
/// Remove selected cutting plane
//--------------------------------------------------------------------------------------------------
void CuttingPlanesWidget::removeCuttingPlane()
{
    int cuttingPlaneIndex = m_ui->cuttingPlanesListView->currentIndex().row();
    if (cuttingPlaneIndex == -1) return;

    m_listModel->removeRow(cuttingPlaneIndex);
    m_ui->cuttingPlanesListView->setCurrentIndex(QModelIndex());

    cee::ug::CuttingPlane* cuttingPlane = m_unstructModel->cuttingPlane(static_cast<size_t>(cuttingPlaneIndex));
    m_unstructModel->removeCuttingPlane(cuttingPlane);

    m_unstructModel->updateVisualization(cee::ug::UnstructGridModel::CUTTING_PLANE);
    m_view->requestRedraw();
}


//--------------------------------------------------------------------------------------------------
/// Show/hide cutting plane
//--------------------------------------------------------------------------------------------------
void CuttingPlanesWidget::setCuttingPlaneVisible(bool show)
{
    int cuttingPlaneIndex = m_ui->cuttingPlanesListView->currentIndex().row();
    if (cuttingPlaneIndex == -1) return;

    cee::ug::CuttingPlane* cuttingPlane = m_unstructModel->cuttingPlane(static_cast<size_t>(cuttingPlaneIndex));
    cuttingPlane->setVisible(show);

    m_unstructModel->updateVisualization(cee::ug::UnstructGridModel::VISIBILITY);
    m_view->requestRedraw();
}


//--------------------------------------------------------------------------------------------------
/// Apply position to selected cutting plane
//--------------------------------------------------------------------------------------------------
void CuttingPlanesWidget::applyPosition()
{
    int cuttingPlaneIndex = m_ui->cuttingPlanesListView->currentIndex().row();
    if (cuttingPlaneIndex == -1) return;

    QStringList coordinates = m_ui->positionLineEdit->text().split(" ");
    if (coordinates.size() == 3)
    {
        bool xOk, yOk, zOk;
        double x = coordinates.at(0).toDouble(&xOk);
        double y = coordinates.at(1).toDouble(&yOk);
        double z = coordinates.at(2).toDouble(&zOk);

        if (xOk && yOk && zOk)
        {
            cee::ug::CuttingPlane* cuttingPlane = m_unstructModel->cuttingPlane(static_cast<size_t>(cuttingPlaneIndex));
            cuttingPlane->setPoint(cee::Vec3d(x, y, z));

            m_unstructModel->updateVisualization(cee::ug::UnstructGridModel::CUTTING_PLANE);
            m_view->requestRedraw();
        }
    }
}


//--------------------------------------------------------------------------------------------------
/// Apply normal to selected cutting plane
//--------------------------------------------------------------------------------------------------
void CuttingPlanesWidget::applyNormal()
{
    int cuttingPlaneIndex = m_ui->cuttingPlanesListView->currentIndex().row();
    if (cuttingPlaneIndex == -1) return;

    QStringList normal = m_ui->normalLineEdit->text().split(" ");
    if (normal.size() == 3)
    {
        bool xOk, yOk, zOk;
        double x = normal.at(0).toDouble(&xOk);
        double y = normal.at(1).toDouble(&yOk);
        double z = normal.at(2).toDouble(&zOk);

        if (xOk && yOk && zOk)
        {
            cee::ug::CuttingPlane* cuttingPlane = m_unstructModel->cuttingPlane(static_cast<size_t>(cuttingPlaneIndex));
            cuttingPlane->setNormal(cee::Vec3d(x, y, z));

            m_unstructModel->updateVisualization(cee::ug::UnstructGridModel::CUTTING_PLANE);
            m_view->requestRedraw();
        }
    }
}


//--------------------------------------------------------------------------------------------------
/// Select scalar result to show as fringes on selected cutting plane
//--------------------------------------------------------------------------------------------------
void CuttingPlanesWidget::setMappedScalarResult(int index)
{
    int cuttingPlaneIndex = m_ui->cuttingPlanesListView->currentIndex().row();
    if (cuttingPlaneIndex == -1) return;

    int resultId = m_ui->mappedScalarResultComboBox->itemData(index).toInt();

    cee::ug::CuttingPlane* cuttingPlane = m_unstructModel->cuttingPlane(static_cast<size_t>(cuttingPlaneIndex));
    cuttingPlane->setMapScalarResultId(resultId);

    m_unstructModel->updateVisualization();
    m_view->requestRedraw();
}


//--------------------------------------------------------------------------------------------------
/// Select vector result to show on selected cutting plane
//--------------------------------------------------------------------------------------------------
void CuttingPlanesWidget::setMappedVectorResult(int index)
{
    int cuttingPlaneIndex = m_ui->cuttingPlanesListView->currentIndex().row();
    if (cuttingPlaneIndex == -1) return;

    int resultId = m_ui->mappedVectorResultComboBox->itemData(index).toInt();

    cee::ug::CuttingPlane* cuttingPlane = m_unstructModel->cuttingPlane(static_cast<size_t>(cuttingPlaneIndex));
    cuttingPlane->setMapVectorResultId(resultId);

    m_unstructModel->updateVisualization();
    m_view->requestRedraw();
}
