//##################################################################################################
//
//  Ceetron Desktop Components
//  QtDemoApp
//  
//  --------------------------------------------------------------------------------------------
//  Copyright (C) 2014, Ceetron AS
//  This is UNPUBLISHED PROPRIETARY SOURCE CODE of Ceetron AS. The contents of this file may 
//  not be disclosed to third parties, copied or duplicated in any form, in whole or in part,
//  without the prior written permission of Ceetron AS.
//##################################################################################################

#include "IsosurfacesWidget.h"

#include "ui_IsosurfacesWidget.h"

#include "CeeVisualization/View.h"

#include "CeeUnstructGrid/DataSource.h"
#include "CeeUnstructGrid/DataSourceDirectory.h"
#include "CeeUnstructGrid/DataStateSpec.h"
#include "CeeUnstructGrid/DataSourceInterface.h"
#include "CeeUnstructGrid/Isosurface.h"
#include "CeeUnstructGrid/ModelSpec.h"
#include "CeeUnstructGrid/PartSettings.h"
#include "CeeUnstructGrid/PartSettingsIterator.h"
#include "CeeUnstructGrid/ResultInfo.h"
#include "CeeUnstructGrid/ScalarSettings.h"
#include "CeeUnstructGrid/UnstructGridModel.h"

#include "ceeqtUtilsCore.h"

#include <QListView>
#include <QModelIndex>
#include <QStringListModel>


//==================================================================================================
//
// Manages isosurfaces
//
//==================================================================================================

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
IsosurfacesWidget::IsosurfacesWidget(QWidget* parent, Qt::WindowFlags f)
:   QWidget(parent, f),
    m_ui(new Ui::IsosurfacesWidget),
    m_listModel(new QStringListModel(this)),
    m_nextId(-1)
{
    m_ui->setupUi(this);
    m_ui->isosurfacesListView->setModel(m_listModel);

    connect(m_ui->isosurfacesListView->selectionModel(), 
        SIGNAL(currentRowChanged(const QModelIndex&, const QModelIndex&)), SLOT(applySelection(const QModelIndex&, const QModelIndex&)));
    connect(m_ui->addIsosurfacePushButton,        SIGNAL(clicked()),                   SLOT(addIsosurface()));
    connect(m_ui->removeIsosurfacePushButton,     SIGNAL(clicked()),                   SLOT(removeIsosurface()));
    connect(m_ui->visibleCheckBox,                SIGNAL(clicked(bool)),               SLOT(setIsosurfaceVisible(bool)));
    connect(m_ui->isoScalarResultComboBox,        SIGNAL(currentIndexChanged(int)),    SLOT(setIsoScalarResult(int)));
    connect(m_ui->isoValueLineEdit,               SIGNAL(editingFinished()),           SLOT(applyIsoValue()));
    connect(m_ui->mappedScalarResultComboBox,     SIGNAL(currentIndexChanged(int)),    SLOT(setMappedScalarResult(int)));
    connect(m_ui->mappedVectorResultComboBox,     SIGNAL(currentIndexChanged(int)),    SLOT(setMappedVectorResult(int)));

    // Make sure the group box title is disabled if the group box itself is disabled
    QPalette palette;
    palette.setColor(QPalette::Disabled, QPalette::WindowText, QApplication::palette().color(QPalette::Disabled, QPalette::WindowText));
    m_ui->settingsGroupBox->setPalette(palette);

    updatePanel();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
IsosurfacesWidget::~IsosurfacesWidget()
{
    delete m_listModel;
}


//--------------------------------------------------------------------------------------------------
/// Set up widget with current meta data and selection from model and view
//--------------------------------------------------------------------------------------------------
void IsosurfacesWidget::setup(cee::vis::View* view, cee::ug::UnstructGridModel* model)
{
    m_unstructModel = model;
    m_view  = view;

    const cee::ug::DataSourceDirectory* directory = m_unstructModel->dataSource()->directory();
    Q_ASSERT(directory);

    // Get scalar results meta data for the models data source
    {
        m_ui->isoScalarResultComboBox->blockSignals(true);
        m_ui->isoScalarResultComboBox->clear();

        m_ui->mappedScalarResultComboBox->blockSignals(true);
        m_ui->mappedScalarResultComboBox->clear();

        // Add all available scalar results to scalar combo boxes
        std::vector<cee::ug::ResultInfo> scalarInfos = directory->scalarResultInfos();
        std::vector<cee::ug::ResultInfo>::const_iterator it;
        for (it = scalarInfos.begin(); it != scalarInfos.end(); ++it)
        {
            m_ui->isoScalarResultComboBox->addItem(cee::qt::UtilsCore::toQString(it->name()), it->id());
            m_ui->mappedScalarResultComboBox->addItem(cee::qt::UtilsCore::toQString(it->name()), it->id());
        }

        m_ui->isoScalarResultComboBox->blockSignals(false);
        m_ui->mappedScalarResultComboBox->blockSignals(false);
    }

    // Get vector results meta data for the models data source
    {
        m_ui->mappedVectorResultComboBox->blockSignals(true);
        m_ui->mappedVectorResultComboBox->clear();

        // Add all available vector results to vector combo box
        // Add - none - item to deselect all results
        m_ui->mappedVectorResultComboBox->addItem(tr("- none -"), -1);

        std::vector<cee::ug::ResultInfo> vectorInfos = directory->vectorResultInfos();
        std::vector<cee::ug::ResultInfo>::const_iterator it;
        for (it = vectorInfos.begin(); it != vectorInfos.end(); ++it)
        {
            m_ui->mappedVectorResultComboBox->addItem(cee::qt::UtilsCore::toQString(it->name()), it->id());
        }

        m_ui->mappedVectorResultComboBox->blockSignals(false);
    }

    // Populate with all isosurfaces from the model
    {
        m_nextId = 1;
        size_t isosurfaceCount = m_unstructModel->isosurfaceCount();

        QStringList stringList;
        for (size_t i = 0; i < isosurfaceCount; ++i)
        {
            stringList += tr("Isosurface %1").arg(m_nextId++);
        }
        m_listModel->setStringList(stringList);
    }

    updatePanel();
}


//--------------------------------------------------------------------------------------------------
/// An isosurface was selected from outside the panel. (For instance by picking.)
/// Select isosurface in list box and update GUI.
//--------------------------------------------------------------------------------------------------
void IsosurfacesWidget::selectIsosurface(size_t isosurfaceIndex)
{
    QModelIndex index;
    if (isosurfaceIndex != cee::UNDEFINED_SIZE_T)
    {
        index = m_listModel->index(static_cast<int>(isosurfaceIndex));
    }

    m_ui->isosurfacesListView->setCurrentIndex(index);
}


//--------------------------------------------------------------------------------------------------
/// Update panel settings.
/// Clear if no isosurface is available/selected.
//--------------------------------------------------------------------------------------------------
void IsosurfacesWidget::updatePanel()
{
    if (m_unstructModel)
    {
        setEnabled(true);

        int isosurfaceIndex = m_ui->isosurfacesListView->currentIndex().row();
        if (isosurfaceIndex != -1)
        {
            cee::ug::Isosurface* isosurface = m_unstructModel->isosurface(isosurfaceIndex);

            m_ui->visibleCheckBox->setChecked(isosurface->visible());

            m_ui->isoScalarResultComboBox->blockSignals(true);
            m_ui->isoScalarResultComboBox->setCurrentIndex(m_ui->isoScalarResultComboBox->findData(isosurface->isoScalarResultId()));
            m_ui->isoScalarResultComboBox->blockSignals(false);

            m_ui->isoValueLineEdit->setText(QString::number(isosurface->isoValue()));

            m_ui->mappedScalarResultComboBox->blockSignals(true);
            m_ui->mappedScalarResultComboBox->setCurrentIndex(m_ui->mappedScalarResultComboBox->findData(isosurface->mapScalarResultId()));
            m_ui->mappedScalarResultComboBox->blockSignals(false);

            m_ui->mappedVectorResultComboBox->blockSignals(true);
            m_ui->mappedVectorResultComboBox->setCurrentIndex(m_ui->mappedVectorResultComboBox->findData(isosurface->mapVectorResultId()));
            m_ui->mappedVectorResultComboBox->blockSignals(false);

            // Check to see if we have precomputed isosurfaces. 
            // These cannot be deleted, moved or modified.
            // Disable GUI according to this.
            bool isPrecomputed = isosurface->isPrecomputed();
            m_ui->removeIsosurfacePushButton->setEnabled(!isPrecomputed);
            m_ui->isoScalarResultComboBox->setEnabled(!isPrecomputed);
            m_ui->isoValueLineEdit->setEnabled(!isPrecomputed);
            m_ui->mappedScalarResultComboBox->setEnabled(!isPrecomputed);
            m_ui->mappedVectorResultComboBox->setEnabled(!isPrecomputed);

            m_ui->settingsGroupBox->setEnabled(true);
        }
        else
        {
            m_ui->removeIsosurfacePushButton->setEnabled(false);

            m_ui->visibleCheckBox->setChecked(false);
            m_ui->isoValueLineEdit->setText("");

            m_ui->mappedScalarResultComboBox->blockSignals(true);
            m_ui->mappedScalarResultComboBox->setCurrentIndex(0);
            m_ui->mappedScalarResultComboBox->blockSignals(false);

            m_ui->mappedVectorResultComboBox->blockSignals(true);
            m_ui->mappedVectorResultComboBox->setCurrentIndex(0);
            m_ui->mappedVectorResultComboBox->blockSignals(false);

            m_ui->settingsGroupBox->setEnabled(false);
        }
    }
    else
    {
        setEnabled(false);

        m_ui->visibleCheckBox->setChecked(false);
        m_ui->isoValueLineEdit->setText("");

        m_ui->mappedScalarResultComboBox->blockSignals(true);
        m_ui->mappedScalarResultComboBox->setCurrentIndex(0);
        m_ui->mappedScalarResultComboBox->blockSignals(false);

        m_ui->mappedVectorResultComboBox->blockSignals(true);
        m_ui->mappedVectorResultComboBox->setCurrentIndex(0);
        m_ui->mappedVectorResultComboBox->blockSignals(false);
    }
}


//--------------------------------------------------------------------------------------------------
/// Update widgets to show state of selected isosurface, if any.
/// Emits a signal to tell other that selection has changed and update visualization is needed.
//--------------------------------------------------------------------------------------------------
void IsosurfacesWidget::applySelection(const QModelIndex& current, const QModelIndex& previous)
{
    if (previous.isValid())
    {
        m_unstructModel->isosurface(static_cast<size_t>(previous.row()))->setHighlight(false);
    }

    if (current.isValid())
    {
        m_unstructModel->isosurface(static_cast<size_t>(current.row()))->setHighlight(true);
        emit selected();
    }

    updatePanel();
}


//--------------------------------------------------------------------------------------------------
/// Add a new isosurface
//--------------------------------------------------------------------------------------------------
void IsosurfacesWidget::addIsosurface()
{
    cee::PtrRef<cee::ug::Isosurface> isosurface = new cee::ug::Isosurface();

    // Default: Use the scalar and vector result currently shown on the model
    int fringesResultId = m_unstructModel->modelSpec().fringesResultId();
    if (fringesResultId == -1)
    {
        // Pick first available scalar if non already set by model
        std::vector<cee::ug::ResultInfo> results = m_unstructModel->dataSource()->directory()->scalarResultInfos();
        if (results.size() > 0)
        {
            fringesResultId = results[0].id();
        }
    }

    if (fringesResultId != -1)
    {
        isosurface->setIsoScalarResultId(fringesResultId);
        isosurface->setMapScalarResultId(fringesResultId);

        // Result needs to be loaded to get scalar values
        {
            Q_ASSERT(dynamic_cast<cee::ug::DataSourceInterface*>(m_unstructModel->dataSource()));
            cee::ug::DataSourceInterface* dataSource = static_cast<cee::ug::DataSourceInterface*>(m_unstructModel->dataSource());

            std::vector<int> resultIds;
            resultIds.push_back(fringesResultId);

            cee::ug::DataStateSpec spec;
            spec.setScalarResultIds(resultIds);

            Q_ASSERT(m_unstructModel->modelSpec().stateIds().size() > 0);
            dataSource->loadState(m_unstructModel->modelSpec().stateIds()[0], spec, NULL);
        }

        double minimum, maximum;
        m_unstructModel->scalarRange(fringesResultId, &minimum, &maximum);
        isosurface->setIsoValue(minimum + ((maximum - minimum) / 2.0));
    }

    if (m_unstructModel->modelSpec().vectorResultIds().size() > 0) 
    {
        isosurface->setMapVectorResultId(m_unstructModel->modelSpec().vectorResultIds()[0]);
    }

    // Add the isosurface to the model and listview
    m_unstructModel->addIsosurface(isosurface.get());
    m_listModel->insertRow(m_listModel->rowCount());
    QModelIndex modelIndex = m_listModel->index(m_listModel->rowCount()-1);
    m_listModel->setData(modelIndex, tr("Isosurface %1").arg(m_nextId++));

    // Set all parts to outline draw style to see cutting plane
    cee::ug::PartSettingsIterator it(m_unstructModel.get());
    while (it.hasNext())
    {
        cee::ug::PartSettings* settings = it.next();
        settings->setDrawStyle(cee::ug::PartSettings::OUTLINE);
    }

    m_ui->isosurfacesListView->setCurrentIndex(modelIndex);

    m_unstructModel->updateVisualization();
    m_view->requestRedraw();
}


//--------------------------------------------------------------------------------------------------
/// Remove selected isosurface
//--------------------------------------------------------------------------------------------------
void IsosurfacesWidget::removeIsosurface()
{
    int isosurfaceIndex = m_ui->isosurfacesListView->currentIndex().row();
    if (isosurfaceIndex == -1) return;

    m_listModel->removeRow(isosurfaceIndex);
    m_ui->isosurfacesListView->setCurrentIndex(QModelIndex());

    cee::ug::Isosurface* isosurface = m_unstructModel->isosurface(static_cast<size_t>(isosurfaceIndex));
    m_unstructModel->removeIsosurface(isosurface);

    m_unstructModel->updateVisualization(cee::ug::UnstructGridModel::ISOSURFACE);
    m_view->requestRedraw();
}


//--------------------------------------------------------------------------------------------------
/// Show/hide isosurface
//--------------------------------------------------------------------------------------------------
void IsosurfacesWidget::setIsosurfaceVisible(bool show)
{
    int isosurfaceIndex = m_ui->isosurfacesListView->currentIndex().row();
    if (isosurfaceIndex == -1) return;

    cee::ug::Isosurface* isosurface = m_unstructModel->isosurface(static_cast<size_t>(isosurfaceIndex));
    isosurface->setVisible(show);

    m_unstructModel->updateVisualization(cee::ug::UnstructGridModel::VISIBILITY);
    m_view->requestRedraw();
}


//--------------------------------------------------------------------------------------------------
/// Select scalar result used to calculate isosurface
//--------------------------------------------------------------------------------------------------
void IsosurfacesWidget::setIsoScalarResult(int index)
{
    int isosurfaceIndex = m_ui->isosurfacesListView->currentIndex().row();
    if (isosurfaceIndex == -1) return;

    int resultId = m_ui->isoScalarResultComboBox->itemData(index).toInt();

    cee::ug::Isosurface* isosurface = m_unstructModel->isosurface(static_cast<size_t>(isosurfaceIndex));
    isosurface->setIsoScalarResultId(resultId);

    m_unstructModel->updateVisualization();
    m_view->requestRedraw();
}


//--------------------------------------------------------------------------------------------------
/// Set iso value used to calculate isosurface
//--------------------------------------------------------------------------------------------------
void IsosurfacesWidget::applyIsoValue()
{
    int isosurfaceIndex = m_ui->isosurfacesListView->currentIndex().row();
    if (isosurfaceIndex == -1) return;

    bool ok = false;
    double isoValue = m_ui->isoValueLineEdit->text().toDouble(&ok);

    if (ok)
    {
        cee::ug::Isosurface* isosurface = m_unstructModel->isosurface(static_cast<size_t>(isosurfaceIndex));
        isosurface->setIsoValue(isoValue);

        m_unstructModel->updateVisualization(cee::ug::UnstructGridModel::ISOSURFACE);
        m_view->requestRedraw();
    }
}


//--------------------------------------------------------------------------------------------------
/// Select scalar result to show as fringes on selected isosurface
//--------------------------------------------------------------------------------------------------
void IsosurfacesWidget::setMappedScalarResult(int index)
{
    int isosurfaceIndex = m_ui->isosurfacesListView->currentIndex().row();
    if (isosurfaceIndex == -1) return;

    int resultId = m_ui->mappedScalarResultComboBox->itemData(index).toInt();

    cee::ug::Isosurface* isosurface = m_unstructModel->isosurface(static_cast<size_t>(isosurfaceIndex));
    isosurface->setMapScalarResultId(resultId);

    m_unstructModel->updateVisualization();
    m_view->requestRedraw();
}


//--------------------------------------------------------------------------------------------------
/// Select vector result to show on selected isosurface
//--------------------------------------------------------------------------------------------------
void IsosurfacesWidget::setMappedVectorResult(int index)
{
    int isosurfaceIndex = m_ui->isosurfacesListView->currentIndex().row();
    if (isosurfaceIndex == -1) return;

    int resultId = m_ui->mappedVectorResultComboBox->itemData(index).toInt();

    cee::ug::Isosurface* isosurface = m_unstructModel->isosurface(static_cast<size_t>(isosurfaceIndex));
    isosurface->setMapVectorResultId(resultId);

    m_unstructModel->updateVisualization();
    m_view->requestRedraw();
}
