//##################################################################################################
//
//  Ceetron Desktop Components
//  QtDemoApp
//  
//  --------------------------------------------------------------------------------------------
//  Copyright (C) 2014, Ceetron AS
//  This is UNPUBLISHED PROPRIETARY SOURCE CODE of Ceetron AS. The contents of this file may 
//  not be disclosed to third parties, copied or duplicated in any form, in whole or in part,
//  without the prior written permission of Ceetron AS.
//##################################################################################################

#include "IsovolumesWidget.h"

#include "ui_IsovolumesWidget.h"

#include "CeeVisualization/View.h"

#include "CeeUnstructGrid/DataSource.h"
#include "CeeUnstructGrid/DataSourceDirectory.h"
#include "CeeUnstructGrid/DataStateSpec.h"
#include "CeeUnstructGrid/DataSourceInterface.h"
#include "CeeUnstructGrid/Isovolume.h"
#include "CeeUnstructGrid/ModelSpec.h"
#include "CeeUnstructGrid/PartSettings.h"
#include "CeeUnstructGrid/PartSettingsIterator.h"
#include "CeeUnstructGrid/ResultInfo.h"
#include "CeeUnstructGrid/ScalarSettings.h"
#include "CeeUnstructGrid/UnstructGridModel.h"

#include "ceeqtUtilsCore.h"

#include <QListView>
#include <QModelIndex>
#include <QStringListModel>


//==================================================================================================
//
// Manages isovolumes
//
//==================================================================================================

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
IsovolumesWidget::IsovolumesWidget(QWidget* parent, Qt::WindowFlags f)
:   QWidget(parent, f),
    m_ui(new Ui::IsovolumesWidget),
    m_listModel(new QStringListModel(this)),
    m_nextId(-1)
{
    m_ui->setupUi(this);
    m_ui->isovolumesListView->setModel(m_listModel);

    connect(m_ui->isovolumesListView->selectionModel(), 
        SIGNAL(currentRowChanged(const QModelIndex&, const QModelIndex&)), SLOT(applySelection(const QModelIndex&, const QModelIndex&)));
    connect(m_ui->addIsovolumePushButton,        SIGNAL(clicked()),                   SLOT(addIsovolume()));
    connect(m_ui->removeIsovolumePushButton,     SIGNAL(clicked()),                   SLOT(removeIsovolume()));
    connect(m_ui->visibleCheckBox,               SIGNAL(clicked(bool)),               SLOT(setIsovolumeVisible(bool)));
    connect(m_ui->isoScalarResultComboBox,       SIGNAL(currentIndexChanged(int)),    SLOT(setIsoScalarResult(int)));
    connect(m_ui->rangeMinimumLineEdit,          SIGNAL(editingFinished()),           SLOT(applyRange()));
    connect(m_ui->rangeMaximumLineEdit,          SIGNAL(editingFinished()),           SLOT(applyRange()));
    connect(m_ui->mappedScalarResultComboBox,    SIGNAL(currentIndexChanged(int)),    SLOT(setMappedScalarResult(int)));
    connect(m_ui->mappedVectorResultComboBox,    SIGNAL(currentIndexChanged(int)),    SLOT(setMappedVectorResult(int)));

    // Make sure the group box title is disabled if the group box itself is disabled
    QPalette palette;
    palette.setColor(QPalette::Disabled, QPalette::WindowText, QApplication::palette().color(QPalette::Disabled, QPalette::WindowText));
    m_ui->settingsGroupBox->setPalette(palette);

    updatePanel();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
IsovolumesWidget::~IsovolumesWidget()
{
    delete m_listModel;
}


//--------------------------------------------------------------------------------------------------
/// Set up widget with current meta data and selection from model and view
//--------------------------------------------------------------------------------------------------
void IsovolumesWidget::setup(cee::vis::View* view, cee::ug::UnstructGridModel* model)
{
    m_unstructModel = model;
    m_view = view;

    const cee::ug::DataSourceDirectory* directory = m_unstructModel->dataSource()->directory();
    Q_ASSERT(directory);

    // Get scalar meta data for the models data source
    {
        m_ui->isoScalarResultComboBox->blockSignals(true);
        m_ui->isoScalarResultComboBox->clear();

        m_ui->mappedScalarResultComboBox->blockSignals(true);
        m_ui->mappedScalarResultComboBox->clear();

        // Add all available scalar results to scalar combo boxes
        std::vector<cee::ug::ResultInfo> scalarInfos = directory->scalarResultInfos();
        std::vector<cee::ug::ResultInfo>::const_iterator it;
        for (it = scalarInfos.begin(); it != scalarInfos.end(); ++it)
        {
            m_ui->isoScalarResultComboBox->addItem(cee::qt::UtilsCore::toQString(it->name()), it->id());
            m_ui->mappedScalarResultComboBox->addItem(cee::qt::UtilsCore::toQString(it->name()), it->id());
        }

        m_ui->isoScalarResultComboBox->blockSignals(false);
        m_ui->mappedScalarResultComboBox->blockSignals(false);
    }

    // Get vector meta data for the models data source
    {
        m_ui->mappedVectorResultComboBox->blockSignals(true);
        m_ui->mappedVectorResultComboBox->clear();

        // Add all available vector results to vector combo box
        // Add - none - item to deselect all results
        m_ui->mappedVectorResultComboBox->addItem(tr("- none -"), -1);

        std::vector<cee::ug::ResultInfo> vectorInfos = directory->vectorResultInfos();
        std::vector<cee::ug::ResultInfo>::const_iterator it;
        for (it = vectorInfos.begin(); it != vectorInfos.end(); ++it)
        {
            m_ui->mappedVectorResultComboBox->addItem(cee::qt::UtilsCore::toQString(it->name()), it->id());
        }

        m_ui->mappedVectorResultComboBox->blockSignals(false);
    }

    // Populate with all isovolumes from the model
    {
        m_nextId = 1;
        size_t isovolumeCount = m_unstructModel->isovolumeCount();

        QStringList stringList;
        for (size_t i = 0; i < isovolumeCount; ++i)
        {
            stringList += tr("Isovolume %1").arg(m_nextId++);
        }
        m_listModel->setStringList(stringList);
    }

    updatePanel();
}


//--------------------------------------------------------------------------------------------------
/// An isovolume was selected from outside the panel. (For instance by picking.)
/// Select isovolume in list box and set up GUI.
//--------------------------------------------------------------------------------------------------
void IsovolumesWidget::selectIsovolume(size_t isovolumeIndex)
{
    QModelIndex index;
    if (isovolumeIndex != cee::UNDEFINED_SIZE_T)
    {
        index = m_listModel->index(static_cast<int>(isovolumeIndex));
    }

    m_ui->isovolumesListView->setCurrentIndex(index);
}


//--------------------------------------------------------------------------------------------------
/// Update panel settings.
/// Clear if no isovolume is available/selected.
//--------------------------------------------------------------------------------------------------
void IsovolumesWidget::updatePanel()
{
    if (m_unstructModel)
    {
        setEnabled(true);

        int isovolumeIndex = m_ui->isovolumesListView->currentIndex().row();
        if (isovolumeIndex != -1)
        {
            cee::ug::Isovolume* isovolume = m_unstructModel->isovolume(isovolumeIndex);

            m_ui->visibleCheckBox->setChecked(isovolume->visible());

            m_ui->isoScalarResultComboBox->blockSignals(true);
            m_ui->isoScalarResultComboBox->setCurrentIndex(m_ui->isoScalarResultComboBox->findData(isovolume->isoScalarResultId()));
            m_ui->isoScalarResultComboBox->blockSignals(false);

            m_ui->rangeMinimumLineEdit->setText(QString::number(isovolume->minimumIsoValue()));
            m_ui->rangeMaximumLineEdit->setText(QString::number(isovolume->maximumIsoValue()));

            m_ui->mappedScalarResultComboBox->blockSignals(true);
            m_ui->mappedScalarResultComboBox->setCurrentIndex(m_ui->mappedScalarResultComboBox->findData(isovolume->mapScalarResultId()));
            m_ui->mappedScalarResultComboBox->blockSignals(false);

            m_ui->mappedVectorResultComboBox->blockSignals(true);
            m_ui->mappedVectorResultComboBox->setCurrentIndex(m_ui->mappedVectorResultComboBox->findData(isovolume->mapVectorResultId()));
            m_ui->mappedVectorResultComboBox->blockSignals(false);

            // Check to see if we have precomputed isovolumes. 
            // These cannot be deleted, moved or modified.
            // Disable GUI according to this.
            bool isPrecomputed = isovolume->isPrecomputed();
            m_ui->removeIsovolumePushButton->setEnabled(!isPrecomputed);
            m_ui->isoScalarResultComboBox->setEnabled(!isPrecomputed);
            m_ui->rangeMinimumLineEdit->setEnabled(!isPrecomputed);
            m_ui->rangeMaximumLineEdit->setEnabled(!isPrecomputed);
            m_ui->mappedScalarResultComboBox->setEnabled(!isPrecomputed);
            m_ui->mappedVectorResultComboBox->setEnabled(!isPrecomputed);

            m_ui->settingsGroupBox->setEnabled(true);
        }
        else
        {
            m_ui->removeIsovolumePushButton->setEnabled(false);

            m_ui->visibleCheckBox->setChecked(false);
            m_ui->rangeMinimumLineEdit->setText("");
            m_ui->rangeMaximumLineEdit->setText("");

            m_ui->mappedScalarResultComboBox->blockSignals(true);
            m_ui->mappedScalarResultComboBox->setCurrentIndex(0);
            m_ui->mappedScalarResultComboBox->blockSignals(false);

            m_ui->mappedVectorResultComboBox->blockSignals(true);
            m_ui->mappedVectorResultComboBox->setCurrentIndex(0);
            m_ui->mappedVectorResultComboBox->blockSignals(false);

            m_ui->settingsGroupBox->setEnabled(false);
        }
    }
    else
    {
        setEnabled(false);

        m_ui->visibleCheckBox->setChecked(false);
        m_ui->rangeMinimumLineEdit->setText("");
        m_ui->rangeMaximumLineEdit->setText("");

        m_ui->mappedScalarResultComboBox->blockSignals(true);
        m_ui->mappedScalarResultComboBox->setCurrentIndex(0);
        m_ui->mappedScalarResultComboBox->blockSignals(false);

        m_ui->mappedVectorResultComboBox->blockSignals(true);
        m_ui->mappedVectorResultComboBox->setCurrentIndex(0);
        m_ui->mappedVectorResultComboBox->blockSignals(false);
    }
}


//--------------------------------------------------------------------------------------------------
/// Update widgets to show state of selected isovolume, if any.
/// Emits a signal to tell other that selection has changed and update visualization is needed.
//--------------------------------------------------------------------------------------------------
void IsovolumesWidget::applySelection(const QModelIndex& current, const QModelIndex& previous)
{
    if (previous.isValid())
    {
        m_unstructModel->isovolume(static_cast<size_t>(previous.row()))->setHighlight(false);
    }

    if (current.isValid())
    {
        m_unstructModel->isovolume(static_cast<size_t>(current.row()))->setHighlight(true);
        emit selected();
    }

    updatePanel();
}


//--------------------------------------------------------------------------------------------------
/// Add a new isovolume
//--------------------------------------------------------------------------------------------------
void IsovolumesWidget::addIsovolume()
{
    cee::PtrRef<cee::ug::Isovolume> isovolume = new cee::ug::Isovolume();

    // Default: Use the scalar currently shown on the model
    int fringesResultId = m_unstructModel->modelSpec().fringesResultId();
    if (fringesResultId == -1)
    {
        // Pick first available scalar if non already set by model
        std::vector<cee::ug::ResultInfo> results = m_unstructModel->dataSource()->directory()->scalarResultInfos();
        if (results.size() > 0)
        {
            fringesResultId = results[0].id();
        }
    }

    if (fringesResultId != -1)
    {
        isovolume->setIsoScalarResultId(fringesResultId);
        isovolume->setMapScalarResultId(fringesResultId);

        // Result needs to be loaded to get min and max values
        {
            Q_ASSERT(dynamic_cast<cee::ug::DataSourceInterface*>(m_unstructModel->dataSource()));    // VTFx data source
            cee::ug::DataSourceInterface* dataSource = static_cast<cee::ug::DataSourceInterface*>(m_unstructModel->dataSource());

            std::vector<int> resultIds;
            resultIds.push_back(fringesResultId);

            cee::ug::DataStateSpec spec;
            spec.setScalarResultIds(resultIds);

            Q_ASSERT(m_unstructModel->modelSpec().stateIds().size() > 0);
            dataSource->loadState(m_unstructModel->modelSpec().stateIds()[0], spec, NULL);
        }

        // Set full range
        double minimum, maximum;
        m_unstructModel->scalarRange(fringesResultId, &minimum, &maximum);
        isovolume->setMinimumIsoValue(minimum + (maximum - minimum)/2.0);
        isovolume->setMaximumIsoValue(maximum);
    }

    if (m_unstructModel->modelSpec().vectorResultIds().size() > 0) 
    {
        isovolume->setMapVectorResultId(m_unstructModel->modelSpec().vectorResultIds()[0]);
    }

    // Add the isovolume to the model and listview
    m_unstructModel->addIsovolume(isovolume.get());
    m_listModel->insertRow(m_listModel->rowCount());
    QModelIndex modelIndex = m_listModel->index(m_listModel->rowCount()-1);
    m_listModel->setData(modelIndex, tr("Isovolume %1").arg(m_nextId++));

    // Set all parts to outline draw style to see cutting plane
    cee::ug::PartSettingsIterator it(m_unstructModel.get());
    while (it.hasNext())
    {
        cee::ug::PartSettings* settings = it.next();
        settings->setDrawStyle(cee::ug::PartSettings::OUTLINE);
    }

    // Select the newly added isovolume
    m_ui->isovolumesListView->setCurrentIndex(modelIndex);

    m_unstructModel->updateVisualization();
    m_view->requestRedraw();
}


//--------------------------------------------------------------------------------------------------
/// Remove selected isovolume
//--------------------------------------------------------------------------------------------------
void IsovolumesWidget::removeIsovolume()
{
    int isovolumeIndex = m_ui->isovolumesListView->currentIndex().row();
    if (isovolumeIndex == -1) return;

    m_listModel->removeRow(isovolumeIndex);
    m_ui->isovolumesListView->setCurrentIndex(QModelIndex());

    cee::ug::Isovolume* isovolume = m_unstructModel->isovolume(static_cast<size_t>(isovolumeIndex));
    m_unstructModel->removeIsovolume(isovolume);

    m_unstructModel->updateVisualization(cee::ug::UnstructGridModel::ISOVOLUME);
    m_view->requestRedraw();
}


//--------------------------------------------------------------------------------------------------
/// Show/hide isovolume
//--------------------------------------------------------------------------------------------------
void IsovolumesWidget::setIsovolumeVisible(bool show)
{
    int isovolumeIndex = m_ui->isovolumesListView->currentIndex().row();
    if (isovolumeIndex == -1) return;

    cee::ug::Isovolume* isovolume = m_unstructModel->isovolume(static_cast<size_t>(isovolumeIndex));
    isovolume->setVisible(show);

    m_unstructModel->updateVisualization(cee::ug::UnstructGridModel::VISIBILITY);
    m_view->requestRedraw();
}


//--------------------------------------------------------------------------------------------------
/// Select scalar result used to calculate isovolume
//--------------------------------------------------------------------------------------------------
void IsovolumesWidget::setIsoScalarResult(int index)
{
    int isovolumeIndex = m_ui->isovolumesListView->currentIndex().row();
    if (isovolumeIndex == -1) return;

    int resultId = m_ui->isoScalarResultComboBox->itemData(index).toInt();

    cee::ug::Isovolume* isovolume = m_unstructModel->isovolume(static_cast<size_t>(isovolumeIndex));
    isovolume->setIsoScalarResultId(resultId);

    m_unstructModel->updateVisualization();
    m_view->requestRedraw();
}


//--------------------------------------------------------------------------------------------------
/// Set iso range used to calculate isovolume
//--------------------------------------------------------------------------------------------------
void IsovolumesWidget::applyRange()
{
    int isovolumeIndex = m_ui->isovolumesListView->currentIndex().row();
    if (isovolumeIndex == -1) return;

    bool okMin = false;
    double minimum = m_ui->rangeMinimumLineEdit->text().toDouble(&okMin);

    bool okMax = false;
    double maximum = m_ui->rangeMaximumLineEdit->text().toDouble(&okMax);

    if (okMin && okMax && (minimum <= maximum))
    {
        cee::ug::Isovolume* isovolume = m_unstructModel->isovolume(static_cast<size_t>(isovolumeIndex));
        isovolume->setMinimumIsoValue(minimum);
        isovolume->setMaximumIsoValue(maximum);

        m_unstructModel->updateVisualization(cee::ug::UnstructGridModel::ISOVOLUME);
        m_view->requestRedraw();
    }
}


//--------------------------------------------------------------------------------------------------
/// Select scalar result to show as fringes on selected isovolume
//--------------------------------------------------------------------------------------------------
void IsovolumesWidget::setMappedScalarResult(int index)
{
    int isovolumeIndex = m_ui->isovolumesListView->currentIndex().row();
    if (isovolumeIndex == -1) return;

    int resultId = m_ui->mappedScalarResultComboBox->itemData(index).toInt();

    cee::ug::Isovolume* isovolume = m_unstructModel->isovolume(static_cast<size_t>(isovolumeIndex));
    isovolume->setMapScalarResultId(resultId);

    m_unstructModel->updateVisualization();
    m_view->requestRedraw();
}


//--------------------------------------------------------------------------------------------------
/// Select vector result to show on selected isovolume
//--------------------------------------------------------------------------------------------------
void IsovolumesWidget::setMappedVectorResult(int index)
{
    int isovolumeIndex = m_ui->isovolumesListView->currentIndex().row();
    if (isovolumeIndex == -1) return;

    int resultId = m_ui->mappedVectorResultComboBox->itemData(index).toInt();

    cee::ug::Isovolume* isovolume = m_unstructModel->isovolume(static_cast<size_t>(isovolumeIndex));
    isovolume->setMapVectorResultId(resultId);

    m_unstructModel->updateVisualization();
    m_view->requestRedraw();
}
