//##################################################################################################
//
//  Ceetron Desktop Components
//  QtDemoApp
//  
//  --------------------------------------------------------------------------------------------
//  Copyright (C) 2014, Ceetron AS
//  This is UNPUBLISHED PROPRIETARY SOURCE CODE of Ceetron AS. The contents of this file may 
//  not be disclosed to third parties, copied or duplicated in any form, in whole or in part,
//  without the prior written permission of Ceetron AS.
//##################################################################################################

#include "ModelSpecificationWidget.h"

#include "ui_ModelSpecificationWidget.h"

#include "CeeVisualization/View.h"

#include "CeeUnstructGrid/DataSource.h"
#include "CeeUnstructGrid/DataSourceDirectory.h"
#include "CeeUnstructGrid/ModelSpec.h"
#include "CeeUnstructGrid/ResultInfo.h"
#include "CeeUnstructGrid/StateInfo.h"
#include "CeeUnstructGrid/UnstructGridModel.h"
#include "CeeUnstructGrid/ScalarSettings.h"

#include "ceeqtUtilsCore.h"

#include <QComboBox>
#include <QSpinBox>
#include <QToolButton>


//==================================================================================================
//
// Set up model specification
//
//==================================================================================================

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
ModelSpecificationWidget::ModelSpecificationWidget(QWidget* parent, Qt::WindowFlags f)
:   QWidget(parent, f),
    m_ui(new Ui::ModelSpecificationWidget)
{
    m_ui->setupUi(this);

    connect(m_ui->stateComboBox,                SIGNAL(currentIndexChanged(int)), SLOT(setState(int)));
    connect(m_ui->scalarResultComboBox,         SIGNAL(currentIndexChanged(int)), SLOT(setScalarResult(int)));
    connect(m_ui->vectorResultComboBox,         SIGNAL(currentIndexChanged(int)), SLOT(setVectorResult(int)));
    connect(m_ui->displacementResultComboBox,   SIGNAL(currentIndexChanged(int)), SLOT(setDisplacementResult(int)));

    // Make sure the group box title is disabled if the group box itself is disabled
    QPalette palette;
    palette.setColor(QPalette::Disabled, QPalette::WindowText, QApplication::palette().color(QPalette::Disabled, QPalette::WindowText));
    m_ui->currentSpecificationGroupBox->setPalette(palette);

    updatePanel();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
ModelSpecificationWidget::~ModelSpecificationWidget()
{
}


//--------------------------------------------------------------------------------------------------
/// Set up widget with current meta data and selection from model and view
//--------------------------------------------------------------------------------------------------
void ModelSpecificationWidget::setup(cee::vis::View* view, cee::ug::UnstructGridModel* model)
{
    m_unstructModel = model;
    m_view = view;

    const cee::ug::DataSourceDirectory* directory = m_unstructModel->dataSource()->directory();
    Q_ASSERT(directory);

    // Get state meta data for the models data source
    {
        m_ui->stateComboBox->blockSignals(true);
        m_ui->stateComboBox->clear();

        // Add all available states to state combo box
        std::vector<cee::ug::StateInfo> stateInfos = directory->stateInfos();
        std::vector<cee::ug::StateInfo>::const_iterator it;
        for (it = stateInfos.begin(); it != stateInfos.end(); ++it)
        {
            m_ui->stateComboBox->addItem(cee::qt::UtilsCore::toQString(it->name()), it->id());
        }

        m_ui->stateComboBox->blockSignals(false);
    }

    // Get scalar meta data for the models data source
    {
        m_ui->scalarResultComboBox->blockSignals(true);
        m_ui->scalarResultComboBox->clear();

        // Add all available scalar results to scalar combo box
        // Add - none - item to deselect all results
        m_ui->scalarResultComboBox->addItem(tr("- none -"), -1);

        std::vector<cee::ug::ResultInfo> scalarInfos = directory->scalarResultInfos();
        std::vector<cee::ug::ResultInfo>::const_iterator it;
        for (it = scalarInfos.begin(); it != scalarInfos.end(); ++it)
        {
            m_ui->scalarResultComboBox->addItem(cee::qt::UtilsCore::toQString(it->name()), it->id());
        }

        m_ui->scalarResultComboBox->blockSignals(false);
    }

    // Get vector meta data for the models data source
    {
        m_ui->vectorResultComboBox->blockSignals(true);
        m_ui->vectorResultComboBox->clear();

        // Add all available vector results to vector combo box
        // Add - none - item to deselect all results
        m_ui->vectorResultComboBox->addItem(tr("- none -"), -1);

        std::vector<cee::ug::ResultInfo> vectorInfos = directory->vectorResultInfos();
        std::vector<cee::ug::ResultInfo>::const_iterator it;
        for (it = vectorInfos.begin(); it != vectorInfos.end(); ++it)
        {
            m_ui->vectorResultComboBox->addItem(cee::qt::UtilsCore::toQString(it->name()), it->id());
        }

        m_ui->vectorResultComboBox->blockSignals(false);
    }


    // Get displacement meta data for the models data source
    {
        m_ui->displacementResultComboBox->blockSignals(true);
        m_ui->displacementResultComboBox->clear();

        // Add all available vector results to displacement combo box
        // Add - none - item to deselect all results
        m_ui->displacementResultComboBox->addItem(tr("- none -"), -1);

        std::vector<cee::ug::ResultInfo> displacementInfos = directory->displacementResultInfos();
        std::vector<cee::ug::ResultInfo>::const_iterator it;
        for (it = displacementInfos.begin(); it != displacementInfos.end(); ++it)
        {
            m_ui->displacementResultComboBox->addItem(cee::qt::UtilsCore::toQString(it->name()), it->id());
        }

        m_ui->displacementResultComboBox->blockSignals(false);
    }

    // Current state
    if (m_unstructModel->modelSpec().stateIds().size() > 0)
    {
        int stateId = m_unstructModel->modelSpec().stateIds()[0];
        int itemIndex = m_ui->stateComboBox->findData(stateId);
        Q_ASSERT(itemIndex != -1);

        m_ui->stateComboBox->blockSignals(true);
        m_ui->stateComboBox->setCurrentIndex(itemIndex);
        m_ui->stateComboBox->blockSignals(false);
    }

    // Current scalar
    {
        int scalarId = m_unstructModel->modelSpec().fringesResultId();
        int itemIndex = m_ui->scalarResultComboBox->findData(scalarId);

        m_ui->scalarResultComboBox->blockSignals(true);
        m_ui->scalarResultComboBox->setCurrentIndex(itemIndex);
        m_ui->scalarResultComboBox->blockSignals(false);
    }

    // Current vector
    if (m_unstructModel->modelSpec().vectorResultIds().size() > 0)
    {
        int vectorId = m_unstructModel->modelSpec().vectorResultIds()[0];
        int itemIndex = m_ui->vectorResultComboBox->findData(vectorId);

        m_ui->vectorResultComboBox->blockSignals(true);
        m_ui->vectorResultComboBox->setCurrentIndex(itemIndex);
        m_ui->vectorResultComboBox->blockSignals(false);
    }



    // Current displacement

    int displacementId = m_unstructModel->modelSpec().displacementResultId();
    int itemIndex = m_ui->displacementResultComboBox->findData(displacementId);

    m_ui->displacementResultComboBox->blockSignals(true);
    m_ui->displacementResultComboBox->setCurrentIndex(itemIndex);
    m_ui->displacementResultComboBox->blockSignals(false);


    updatePanel();
}


//--------------------------------------------------------------------------------------------------
/// Update panel
//--------------------------------------------------------------------------------------------------
void ModelSpecificationWidget::updatePanel()
{
    if (m_unstructModel)
    {
        setEnabled(true);
    }
    else
    {
        setEnabled(false);
    }
}


//--------------------------------------------------------------------------------------------------
/// Select new state
//--------------------------------------------------------------------------------------------------
void ModelSpecificationWidget::setState(int index)
{
    int stateId = m_ui->stateComboBox->itemData(index).toInt();
    m_unstructModel->modelSpec().setStateId(stateId);

    m_unstructModel->updateVisualization();
    m_view->requestRedraw();
}


//--------------------------------------------------------------------------------------------------
/// Select new scalar result
//--------------------------------------------------------------------------------------------------
void ModelSpecificationWidget::setScalarResult(int index)
{
    int scalarId = m_ui->scalarResultComboBox->itemData(index).toInt();
    m_unstructModel->modelSpec().setFringesResultId(scalarId);

    cee::ug::ScalarSettings* settings = m_unstructModel->scalarSettings(scalarId);
    if (settings)
    {
        // We have a bright background, so set the legend color to black
        m_unstructModel->scalarSettings(scalarId)->setLegendTextColor(cee::Color3f(0,0,0));
        m_unstructModel->scalarSettings(scalarId)->setLegendLineColor(cee::Color3f(0,0,0));
    }

    m_unstructModel->updateVisualization();
    m_view->requestRedraw();
}


//--------------------------------------------------------------------------------------------------
/// Select new vector result
//--------------------------------------------------------------------------------------------------
void ModelSpecificationWidget::setVectorResult(int index)
{
    int vectorId = m_ui->vectorResultComboBox->itemData(index).toInt();
    m_unstructModel->modelSpec().setVectorResultId(vectorId);

    m_unstructModel->updateVisualization();
    m_view->requestRedraw();
}



//--------------------------------------------------------------------------------------------------
/// Select new displacement result
//--------------------------------------------------------------------------------------------------
void ModelSpecificationWidget::setDisplacementResult(int index)
{
    int displacementId = m_ui->displacementResultComboBox->itemData(index).toInt();
    m_unstructModel->modelSpec().setDisplacementResultId(displacementId);

    m_unstructModel->updateVisualization();
    m_view->requestRedraw();
}