//##################################################################################################
//
//  Ceetron Desktop Components
//  QtDemoApp
//  
//  --------------------------------------------------------------------------------------------
//  Copyright (C) 2014, Ceetron AS
//  This is UNPUBLISHED PROPRIETARY SOURCE CODE of Ceetron AS. The contents of this file may 
//  not be disclosed to third parties, copied or duplicated in any form, in whole or in part,
//  without the prior written permission of Ceetron AS.
//##################################################################################################

#include "ParticleTracesWidget.h"

#include "ui_ParticleTracesWidget.h"

#include "MouseClickEventFilter.h"

#include "CeeVisualization/View.h"

#include "CeeUnstructGrid/DataSource.h"
#include "CeeUnstructGrid/DataSourceDirectory.h"
#include "CeeUnstructGrid/ParticleTraceGroup.h"
#include "CeeUnstructGrid/ResultInfo.h"
#include "CeeUnstructGrid/UnstructGridModel.h"

#include "ceeqtUtilsCore.h"
#include "ceeqtUtilsGui.h"

#include <QColorDialog>
#include <QComboBox>
#include <QModelIndex>
#include <QStandardItem>
#include <QStandardItemModel>
#include <QTableView>

#include <vector>


//==================================================================================================
//
// Manages particle traces
//
//==================================================================================================

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
ParticleTracesWidget::ParticleTracesWidget(QWidget* parent, Qt::WindowFlags f)
:   QWidget(parent, f),
    m_ui(new Ui::ParticleTracesWidget),
    m_listModel(new QStandardItemModel(0, 3, this)),
    m_mouseClickEventFilter(new MouseClickEventFilter)
{
    m_ui->setupUi(this);
    m_ui->seedPointsTableView->setModel(m_listModel);
    m_ui->selectFixedColorFrame->installEventFilter(m_mouseClickEventFilter);

    m_ui->visualizationStyleComboBox->addItem(tr("Cylinders"), cee::ug::ParticleTraceGroup::CYLINDERS);
    m_ui->visualizationStyleComboBox->addItem(tr("Spheres"),   cee::ug::ParticleTraceGroup::SPHERES);
    m_ui->visualizationStyleComboBox->addItem(tr("Ribbons"),   cee::ug::ParticleTraceGroup::RIBBONS);
    m_ui->visualizationStyleComboBox->addItem(tr("Comets"),    cee::ug::ParticleTraceGroup::COMETS);
    m_ui->visualizationStyleComboBox->addItem(tr("Lines"),     cee::ug::ParticleTraceGroup::LINES);

    m_listModel->setHeaderData(0, Qt::Horizontal, tr("X"));
    m_listModel->setHeaderData(1, Qt::Horizontal, tr("Y"));
    m_listModel->setHeaderData(2, Qt::Horizontal, tr("Z"));
    m_ui->seedPointsTableView->horizontalHeader()->setSectionResizeMode(QHeaderView::Stretch);

    connect(m_ui->visibleCheckBox,               SIGNAL(clicked(bool)),               SLOT(setParticleTracesVisible(bool)));
    connect(m_ui->vectorResultComboBox,          SIGNAL(currentIndexChanged(int)),    SLOT(setVectorResult(int)));
    connect(m_ui->mappedScalarResultComboBox,    SIGNAL(currentIndexChanged(int)),    SLOT(setMappedScalarResult(int)));
    connect(m_ui->visualizationStyleComboBox,    SIGNAL(currentIndexChanged(int)),    SLOT(setVisualizationStyle(int)));
    connect(m_ui->numberOfPulsesLineEdit,        SIGNAL(editingFinished()),           SLOT(applyNumberOfPulses()));
    connect(m_ui->useRelativeScalingCheckBox,    SIGNAL(clicked(bool)),               SLOT(enableRelativeScaling(bool)));
    connect(m_ui->primitiveRadiusLineEdit,       SIGNAL(editingFinished()),           SLOT(applyPrimitiveRadius()));
    connect(m_mouseClickEventFilter,             SIGNAL(clicked()),                   SLOT(selectFixedColor()));
    connect(m_ui->addSeedPointPushButton,        SIGNAL(clicked()),                   SLOT(addSeedPoint()));
    connect(m_ui->removeSeedPointPushButton,     SIGNAL(clicked()),                   SLOT(removeSeedPoint()));
    connect(m_ui->removeAllSeedPointsPushButton, SIGNAL(clicked()),                   SLOT(removeAllSeedPoints()));
    connect(m_listModel,                         SIGNAL(itemChanged(QStandardItem*)), SLOT(setSeedPoint(QStandardItem*)));
    connect(m_ui->seedPointsTableView,           SIGNAL(clicked(const QModelIndex&)), SLOT(applySelection(const QModelIndex&)));

    // Make sure the group box title is disabled if the group box itself is disabled
    QPalette palette;
    palette.setColor(QPalette::Disabled, QPalette::WindowText, QApplication::palette().color(QPalette::Disabled, QPalette::WindowText));
    m_ui->settingsGroupBox->setPalette(palette);

    updatePanel();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
ParticleTracesWidget::~ParticleTracesWidget()
{
    delete m_listModel;

    m_ui->selectFixedColorFrame->removeEventFilter(m_mouseClickEventFilter);
    delete m_mouseClickEventFilter;
}


//--------------------------------------------------------------------------------------------------
/// Set up widget with current meta data and selection from model and view
//--------------------------------------------------------------------------------------------------
void ParticleTracesWidget::setup(cee::vis::View* view, cee::ug::UnstructGridModel* model)
{
    m_unstructModel = model;
    m_view = view;

    const cee::ug::DataSourceDirectory* directory = m_unstructModel->dataSource()->directory();
    Q_ASSERT(directory);

    // Get vector meta data for the models data source
    {
        m_ui->vectorResultComboBox->blockSignals(true);
        m_ui->vectorResultComboBox->clear();

        // Add all available vector results to vector combo box
        // Add - none - item to deselect all results
        m_ui->vectorResultComboBox->addItem(tr("- none -"), -1);

        std::vector<cee::ug::ResultInfo> vectorInfos = directory->vectorResultInfos();
        std::vector<cee::ug::ResultInfo>::const_iterator it;
        for (it = vectorInfos.begin(); it != vectorInfos.end(); ++it)
        {
            m_ui->vectorResultComboBox->addItem(cee::qt::UtilsCore::toQString(it->name()), it->id());
        }

        m_ui->vectorResultComboBox->blockSignals(false);
    }

    // Get scalar meta data for the models data source
    {
        m_ui->mappedScalarResultComboBox->blockSignals(true);
        m_ui->mappedScalarResultComboBox->clear();

        // Add all available scalar results to scalar combo box
        // Add - none - item to deselect all results
        m_ui->mappedScalarResultComboBox->addItem(tr("- none -"), -1);

        std::vector<cee::ug::ResultInfo> scalarInfos = directory->scalarResultInfos();
        std::vector<cee::ug::ResultInfo>::const_iterator it;
        for (it = scalarInfos.begin(); it != scalarInfos.end(); ++it)
        {
            m_ui->mappedScalarResultComboBox->addItem(cee::qt::UtilsCore::toQString(it->name()), it->id());
        }

        m_ui->mappedScalarResultComboBox->blockSignals(false);
    }

    updatePanel();
}


//--------------------------------------------------------------------------------------------------
/// Add a particle trace with given seed point to model and GUI
//--------------------------------------------------------------------------------------------------
void ParticleTracesWidget::addSeedPoint(const cee::Vec3d& seedPoint)
{
    // Ensure that we have a particle trace group
    if (m_unstructModel->particleTraceGroupCount() == 0)
    {
        m_unstructModel->addParticleTraceGroup(new cee::ug::ParticleTraceGroup());
    }

    cee::ug::ParticleTraceGroup* group = m_unstructModel->particleTraceGroup(0);
    Q_ASSERT(group);

    // Select first vector if no vector is set in group
    if (group->vectorResultId() == -1)
    {
        std::vector<cee::ug::ResultInfo> results = m_unstructModel->dataSource()->directory()->vectorResultInfos();
        if (results.size() > 0)
        {
            group->setVectorResultId(results[0].id());
        }
    }

    std::vector<cee::Vec3d> seedPoints = group->seedPoints();
    seedPoints.push_back(seedPoint);
    group->setSeedPoints(seedPoints);

    m_unstructModel->updateVisualization();
    m_view->requestRedraw();

    updatePanel();
}


//--------------------------------------------------------------------------------------------------
/// Update panel
//--------------------------------------------------------------------------------------------------
void ParticleTracesWidget::updatePanel()
{
    enableDisableItems();

    if (m_unstructModel && (m_unstructModel->particleTraceGroupCount() > 0))
    {
        cee::ug::ParticleTraceGroup* group = m_unstructModel->particleTraceGroup(0);
        Q_ASSERT(group);

        m_ui->visibleCheckBox->setChecked(group->visible());

        m_ui->vectorResultComboBox->blockSignals(true);
        m_ui->vectorResultComboBox->setCurrentIndex(m_ui->vectorResultComboBox->findData(group->vectorResultId()));
        m_ui->vectorResultComboBox->blockSignals(false);

        m_ui->mappedScalarResultComboBox->blockSignals(true);
        m_ui->mappedScalarResultComboBox->setCurrentIndex(m_ui->mappedScalarResultComboBox->findData(group->mapScalarResultId()));
        m_ui->mappedScalarResultComboBox->blockSignals(false);

        m_ui->visualizationStyleComboBox->blockSignals(true);
        m_ui->visualizationStyleComboBox->setCurrentIndex(m_ui->visualizationStyleComboBox->findData(group->visualizationStyle()));
        m_ui->visualizationStyleComboBox->blockSignals(false);

        cee::ug::ParticleTraceGroup::VisualizationStyle visualizationStyle = group->visualizationStyle();
        bool canSetNumberOfPulses = (visualizationStyle == cee::ug::ParticleTraceGroup::COMETS || visualizationStyle == cee::ug::ParticleTraceGroup::SPHERES);
        m_ui->numberOfPulsesLabel->setEnabled(canSetNumberOfPulses);
        m_ui->numberOfPulsesLineEdit->setEnabled(canSetNumberOfPulses);
        m_ui->numberOfPulsesLineEdit->setText(QString::number(group->pointSamplingNumberOfPulses()));
        m_ui->primitiveRadiusLineEdit->setText(QString::number(group->scaleFactor()));
        m_ui->useRelativeScalingCheckBox->setChecked(group->scaleMode() == cee::ug::ParticleTraceGroup::RELATIVE_SCALING);

        QPalette palette = m_ui->selectFixedColorFrame->palette();
        palette.setColor(m_ui->selectFixedColorFrame->backgroundRole(), cee::qt::UtilsGui::toQColor(group->color()));
        m_ui->selectFixedColorFrame->setPalette(palette);

        m_ui->removeSeedPointPushButton->setEnabled(m_ui->seedPointsTableView->currentIndex().isValid());
        m_ui->removeAllSeedPointsPushButton->setEnabled(group->seedPoints().size() > 0);

        m_listModel->setRowCount(0);

        std::vector<cee::Vec3d> seedPoints = group->seedPoints();
        std::vector<cee::Vec3d>::const_iterator it;
        for (it = seedPoints.begin(); it != seedPoints.end(); ++it)
        {
            cee::Vec3d seedPoint = *it;

            QList<QStandardItem*> coordinates;
            coordinates.append(new QStandardItem(QString::number(seedPoint.x())));
            coordinates.append(new QStandardItem(QString::number(seedPoint.y())));
            coordinates.append(new QStandardItem(QString::number(seedPoint.z())));
            m_listModel->appendRow(coordinates);
        }

        m_ui->vectorResultComboBox->blockSignals(false);
        m_ui->mappedScalarResultComboBox->blockSignals(false);
    }
    else
    {
        m_ui->numberOfPulsesLineEdit->setText("");
        m_ui->primitiveRadiusLineEdit->setText("");

        QPalette palette = m_ui->selectFixedColorFrame->palette();
        palette.setColor(m_ui->selectFixedColorFrame->backgroundRole(), Qt::darkGray);
        m_ui->selectFixedColorFrame->setPalette(palette);
    }
}


//--------------------------------------------------------------------------------------------------
/// Apply scale factor
//--------------------------------------------------------------------------------------------------
void ParticleTracesWidget::applyScaleFactor()
{
    cee::ug::ParticleTraceGroup* group = m_unstructModel->particleTraceGroup(0);
    Q_ASSERT(group);

    bool ok = false;
    double primitiveRadius = m_ui->primitiveRadiusLineEdit->text().toDouble(&ok);

    if (ok)
    {
        if (m_ui->useRelativeScalingCheckBox->isChecked())
        {
            group->set3DPrimitiveRadius(cee::ug::ParticleTraceGroup::RELATIVE_SCALING, primitiveRadius);
        }
        else
        {
            group->set3DPrimitiveRadius(cee::ug::ParticleTraceGroup::ABSOLUTE_SCALING, primitiveRadius);
        }

        m_unstructModel->updateVisualization(cee::ug::UnstructGridModel::PARTICLE_TRACE);
        m_view->requestRedraw();
    }
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void ParticleTracesWidget::enableDisableItems()
{
    if (!m_unstructModel) 
    {
        setEnabled(false);
    }
    else
    {
        setEnabled(true);

        bool anyTraces = m_unstructModel->particleTraceGroupCount() > 0;

        m_ui->visibleCheckBox->setEnabled(anyTraces);
        m_ui->vectorResultComboBox->setEnabled(anyTraces);
        m_ui->mappedScalarResultComboBox->setEnabled(anyTraces);
        m_ui->visualizationStyleComboBox->setEnabled(anyTraces);
        m_ui->numberOfPulsesLabel->setEnabled(anyTraces);
        m_ui->numberOfPulsesLineEdit->setEnabled(anyTraces);
        m_ui->primitiveRadiusLineEdit->setEnabled(anyTraces);
        m_ui->selectFixedColorFrame->setEnabled(anyTraces);
        m_ui->seedPointsTableView->setEnabled(anyTraces);
        m_ui->removeSeedPointPushButton->setEnabled(anyTraces);
        m_ui->removeAllSeedPointsPushButton->setEnabled(anyTraces);
        m_ui->useRelativeScalingCheckBox->setEnabled(anyTraces);
    }
}

//--------------------------------------------------------------------------------------------------
/// Update widgets to show state of selected seed point, if any.
//--------------------------------------------------------------------------------------------------
void ParticleTracesWidget::applySelection(const QModelIndex& index)
{
    Q_UNUSED(index);
    updatePanel();
}


//--------------------------------------------------------------------------------------------------
/// Show/hide particle traces
//--------------------------------------------------------------------------------------------------
void ParticleTracesWidget::setParticleTracesVisible(bool show)
{
    cee::ug::ParticleTraceGroup* group = m_unstructModel->particleTraceGroup(0);
    Q_ASSERT(group);

    group->setVisible(show);

    m_unstructModel->updateVisualization();
    m_view->requestRedraw();
}


//--------------------------------------------------------------------------------------------------
/// Select vector result
//--------------------------------------------------------------------------------------------------
void ParticleTracesWidget::setVectorResult(int index)
{
    cee::ug::ParticleTraceGroup* group = m_unstructModel->particleTraceGroup(0);
    Q_ASSERT(group);

    int resultId = m_ui->vectorResultComboBox->itemData(index).toInt();
    group->setVectorResultId(resultId);

    m_unstructModel->updateVisualization();
    m_view->requestRedraw();
}


//--------------------------------------------------------------------------------------------------
/// Select mapped scalar result
//--------------------------------------------------------------------------------------------------
void ParticleTracesWidget::setMappedScalarResult(int index)
{
    cee::ug::ParticleTraceGroup* group = m_unstructModel->particleTraceGroup(0);
    Q_ASSERT(group);

    int resultId = m_ui->mappedScalarResultComboBox->itemData(index).toInt();
    group->setMapScalarResultId(resultId);

    m_unstructModel->updateVisualization();
    m_view->requestRedraw();
}


//--------------------------------------------------------------------------------------------------
/// Select visualization style
//--------------------------------------------------------------------------------------------------
void ParticleTracesWidget::setVisualizationStyle(int index)
{
    cee::ug::ParticleTraceGroup* group = m_unstructModel->particleTraceGroup(0);
    Q_ASSERT(group);

    int visualizationStyle = m_ui->visualizationStyleComboBox->itemData(index).toInt();
    group->setVisualizationStyle(static_cast<cee::ug::ParticleTraceGroup::VisualizationStyle>(visualizationStyle));

    m_unstructModel->updateVisualization(cee::ug::UnstructGridModel::PARTICLE_TRACE);
    m_view->requestRedraw();

    updatePanel();
}


//--------------------------------------------------------------------------------------------------
/// Set number of pulses. Applies to comets and points.
//--------------------------------------------------------------------------------------------------
void ParticleTracesWidget::applyNumberOfPulses()
{
    cee::ug::ParticleTraceGroup* group = m_unstructModel->particleTraceGroup(0);
    Q_ASSERT(group);

    bool ok = false;
    double numberOfPulses = m_ui->numberOfPulsesLineEdit->text().toDouble(&ok);

    if (ok)
    {
        group->setPointSamplingNumberOfPulses(numberOfPulses);

        m_unstructModel->updateVisualization(cee::ug::UnstructGridModel::PARTICLE_TRACE);
        m_view->requestRedraw();
    }
}


//--------------------------------------------------------------------------------------------------
/// Enable/disable relative scaling
//--------------------------------------------------------------------------------------------------
void ParticleTracesWidget::enableRelativeScaling(bool enable)
{
    cee::ug::ParticleTraceGroup* group = m_unstructModel->particleTraceGroup(0);
    Q_ASSERT(group);

    if (enable)
    {
        group->set3DPrimitiveRadius(cee::ug::ParticleTraceGroup::RELATIVE_SCALING, 1.0);
    }
    else
    {
        double defaultScale = m_unstructModel->boundingBox().extent().length() * 0.006;
        group->set3DPrimitiveRadius(cee::ug::ParticleTraceGroup::ABSOLUTE_SCALING, defaultScale);
    }

    m_unstructModel->updateVisualization(cee::ug::UnstructGridModel::PARTICLE_TRACE);
    m_view->requestRedraw();

    updatePanel();
}


//--------------------------------------------------------------------------------------------------
/// Set radius of primitives
//--------------------------------------------------------------------------------------------------
void ParticleTracesWidget::applyPrimitiveRadius()
{
    applyScaleFactor();
}


//--------------------------------------------------------------------------------------------------
/// Set radius of primitives
//--------------------------------------------------------------------------------------------------
void ParticleTracesWidget::selectFixedColor()
{
    if (m_unstructModel->particleTraceGroupCount() > 0) 
    {
        cee::ug::ParticleTraceGroup* group = m_unstructModel->particleTraceGroup(0);
        Q_ASSERT(group);

        QColor newColor = QColorDialog::getColor(cee::qt::UtilsGui::toQColor(group->color()), this);
        if (newColor.isValid())
        {
            group->setColor(cee::qt::UtilsGui::toColor3f(newColor));

            QPalette palette = m_ui->selectFixedColorFrame->palette();
            palette.setColor(m_ui->selectFixedColorFrame->backgroundRole(), newColor);
            m_ui->selectFixedColorFrame->setPalette(palette);

            m_unstructModel->updateVisualization(cee::ug::UnstructGridModel::PARTICLE_TRACE);
            m_view->requestRedraw();
        }
    }
}


//--------------------------------------------------------------------------------------------------
/// Add seed point
//--------------------------------------------------------------------------------------------------
void ParticleTracesWidget::addSeedPoint()
{
    addSeedPoint(cee::Vec3d(0.0, 0.0, 0.0));
}


//--------------------------------------------------------------------------------------------------
/// Remove selected particle trace
//--------------------------------------------------------------------------------------------------
void ParticleTracesWidget::removeSeedPoint()
{
    QModelIndex modelIndex = m_ui->seedPointsTableView->currentIndex();
    if (!modelIndex.isValid()) return;

    cee::ug::ParticleTraceGroup* group = m_unstructModel->particleTraceGroup(0);
    Q_ASSERT(group);

    std::vector<cee::Vec3d> seedPoints = group->seedPoints();
    seedPoints.erase(seedPoints.begin() + modelIndex.row());
    group->setSeedPoints(seedPoints);

    m_listModel->removeRow(modelIndex.row());

    m_unstructModel->updateVisualization();
    m_view->requestRedraw();

    updatePanel();
}


//--------------------------------------------------------------------------------------------------
/// Remove all particle traces
//--------------------------------------------------------------------------------------------------
void ParticleTracesWidget::removeAllSeedPoints()
{
    cee::ug::ParticleTraceGroup* group = m_unstructModel->particleTraceGroup(0);
    Q_ASSERT(group);

    group->setSeedPoints(std::vector<cee::Vec3d>());

    m_listModel->setRowCount(0);

    m_unstructModel->updateVisualization();
    m_view->requestRedraw();

    updatePanel();
}


//--------------------------------------------------------------------------------------------------
/// Edit seed coordinates
//--------------------------------------------------------------------------------------------------
void ParticleTracesWidget::setSeedPoint(QStandardItem* item)
{
    cee::ug::ParticleTraceGroup* group = m_unstructModel->particleTraceGroup(0);
    Q_ASSERT(group);

    int seedIndex = m_listModel->indexFromItem(item).row();
    if (seedIndex == -1) return;

    bool xOk, yOk, zOk;
    double x = m_listModel->item(seedIndex, 0)->text().toDouble(&xOk);
    double y = m_listModel->item(seedIndex, 1)->text().toDouble(&yOk);
    double z = m_listModel->item(seedIndex, 2)->text().toDouble(&zOk);

    if (xOk && yOk && zOk)
    {
        std::vector<cee::Vec3d> seedPoints = group->seedPoints();
        seedPoints[seedIndex].set(x, y, z);
        group->setSeedPoints(seedPoints);

        m_unstructModel->updateVisualization();
        m_view->requestRedraw();
    }
}
