//##################################################################################################
//
//  Ceetron Desktop Components
//  QtDemoApp
//  
//  --------------------------------------------------------------------------------------------
//  Copyright (C) 2014, Ceetron AS
//  This is UNPUBLISHED PROPRIETARY SOURCE CODE of Ceetron AS. The contents of this file may 
//  not be disclosed to third parties, copied or duplicated in any form, in whole or in part,
//  without the prior written permission of Ceetron AS.
//##################################################################################################

#include "PartsWidget.h"

#include "ui_PartsWidget.h"

#include "MouseClickEventFilter.h"

#include "CeeVisualization/View.h"

#include "CeeUnstructGrid/DataSource.h"
#include "CeeUnstructGrid/DataSourceDirectory.h"
#include "CeeUnstructGrid/PartInfo.h"
#include "CeeUnstructGrid/PartSettings.h"
#include "CeeUnstructGrid/UnstructGridModel.h"

#include "ceeqtUtilsCore.h"
#include "ceeqtUtilsGui.h"

#include <QColorDialog>
#include <QListView>
#include <QModelIndex>
#include <QMouseEvent>
#include <QStandardItemModel>


//==================================================================================================
//
// List all parts and provide part partSettings
//
//==================================================================================================

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
PartsWidget::PartsWidget(QWidget* parent, Qt::WindowFlags f)
:   QWidget(parent, f),
    m_ui(new Ui::PartsWidget),
    m_listModel(new QStandardItemModel(this)),
    m_mouseClickEventFilter(new MouseClickEventFilter)
{
    m_ui->setupUi(this);
    m_ui->partsListView->setModel(m_listModel);
    m_ui->selectFixedColorFrame->installEventFilter(m_mouseClickEventFilter);

    m_ui->drawStyleComboBox->addItem(tr("Surface"),              cee::ug::PartSettings::SURFACE);
    m_ui->drawStyleComboBox->addItem(tr("Surface mesh"),         cee::ug::PartSettings::SURFACE_MESH);
    m_ui->drawStyleComboBox->addItem(tr("Surface outline mesh"), cee::ug::PartSettings::SURFACE_OUTLINE_MESH);
    m_ui->drawStyleComboBox->addItem(tr("Lines"),                cee::ug::PartSettings::LINES);
    m_ui->drawStyleComboBox->addItem(tr("Points"),               cee::ug::PartSettings::POINTS);
    m_ui->drawStyleComboBox->addItem(tr("Outline"),              cee::ug::PartSettings::OUTLINE);
    m_ui->drawStyleComboBox->addItem(tr("Hidden lines"),         cee::ug::PartSettings::HIDDEN_LINES_REMOVED);
    m_ui->drawStyleComboBox->addItem(tr("None"),                 cee::ug::PartSettings::NONE);

    connect(m_ui->partsListView->selectionModel(), 
        SIGNAL(currentRowChanged(const QModelIndex&, const QModelIndex&)), SLOT(applySelection(const QModelIndex&, const QModelIndex&)));
    connect(m_ui->visibleCheckBox,       SIGNAL(clicked(bool)),               SLOT(setPartVisible(bool)));
    connect(m_ui->opacitySlider,         SIGNAL(valueChanged(int)),           SLOT(setOpacity(int)));
    connect(m_ui->drawStyleComboBox,     SIGNAL(currentIndexChanged(int)),    SLOT(setDrawStyle(int)));
    connect(m_mouseClickEventFilter,     SIGNAL(clicked()),                   SLOT(selectFixedColor()));
    connect(m_ui->smoothShadingCheckBox, SIGNAL(clicked(bool)),               SLOT(enableSmoothShading(bool)));
    connect(m_ui->showFringesCheckBox,   SIGNAL(clicked(bool)),               SLOT(showFringes(bool)));
    connect(m_ui->showVectorsCheckBox,   SIGNAL(clicked(bool)),               SLOT(showVectors(bool)));

    // Make sure the group box title is disabled if the group box itself is disabled
    QPalette palette;
    palette.setColor(QPalette::Disabled, QPalette::WindowText, QApplication::palette().color(QPalette::Disabled, QPalette::WindowText));
    m_ui->settingsGroupBox->setPalette(palette);

    updatePanel();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
PartsWidget::~PartsWidget()
{
    delete m_listModel;

    m_ui->selectFixedColorFrame->removeEventFilter(m_mouseClickEventFilter);
    delete m_mouseClickEventFilter;
}


//--------------------------------------------------------------------------------------------------
/// Set up widget with current meta data and selection from model and view
//--------------------------------------------------------------------------------------------------
void PartsWidget::setup(cee::vis::View* view, cee::ug::UnstructGridModel* model)
{
    m_unstructModel = model;
    m_view  = view;

    const cee::ug::DataSourceDirectory* directory = m_unstructModel->dataSource()->directory();
    Q_ASSERT(directory);

    // Get part meta data for the models data source
    {
        m_listModel->clear();

        size_t geometryCount = directory->geometryCountPerState();
        for (size_t geometryIndex = 0; geometryIndex < geometryCount; ++geometryIndex)
        {
            std::vector<cee::ug::PartInfo> partInfos = directory->partInfos(geometryIndex);
            std::vector<cee::ug::PartInfo>::const_iterator it;
            for (it = partInfos.begin(); it != partInfos.end(); ++it)
            {
                QStandardItem* item = new QStandardItem(cee::qt::UtilsCore::toQString(it->name()));
                item->setData(static_cast<qulonglong>(geometryIndex), Qt::UserRole + 1);
                item->setData(it->id(),      Qt::UserRole + 2);

                m_listModel->appendRow(item);
            }
        }
    }

    updatePanel();
}


//--------------------------------------------------------------------------------------------------
/// A part was selected from outside the panel. (For instance by picking.)
/// Select part in list box and update GUI.
//--------------------------------------------------------------------------------------------------
void PartsWidget::selectPart(size_t geometryIndex, int partId)
{
    QModelIndex index;
    if (geometryIndex != cee::UNDEFINED_SIZE_T && partId != -1)
    {
        int partCount = m_listModel->rowCount();
        for (int i = 0; i < partCount; ++i)
        {
            QModelIndex tmpPartIndex = m_listModel->index(i, 0);
            size_t tmpGeometryIndex = tmpPartIndex.data(Qt::UserRole + 1).value<size_t>();
            int    tmpPartId        = tmpPartIndex.data(Qt::UserRole + 2).toInt();

            if (tmpGeometryIndex == geometryIndex && tmpPartId == partId)
            {
                index = tmpPartIndex;
                break;
            }
        }
    }

    m_ui->partsListView->setCurrentIndex(index);
}


//--------------------------------------------------------------------------------------------------
/// Update panel settings.
/// Clear if no part is available/selected.
//--------------------------------------------------------------------------------------------------
void PartsWidget::updatePanel()
{
    if (m_unstructModel)
    {
        setEnabled(true);

        if (m_ui->partsListView->currentIndex().isValid())
        {
            size_t geometryIndex = m_ui->partsListView->currentIndex().data(Qt::UserRole + 1).value<size_t>();
            int partId           = m_ui->partsListView->currentIndex().data(Qt::UserRole + 2).toInt();

            const cee::ug::PartSettings* partSettings = m_unstructModel->partSettings(geometryIndex, partId);

            m_ui->visibleCheckBox->setChecked(partSettings->visible());

            m_ui->opacitySlider->blockSignals(true);
            m_ui->opacitySlider->setValue(static_cast<int>(partSettings->opacity() * 10.0f)); // Opacity is from 0.0 to 1.0, slider values are from 0 to 10, hence multiply with 10
            m_ui->opacitySlider->blockSignals(false);

            m_ui->drawStyleComboBox->blockSignals(true);
            m_ui->drawStyleComboBox->setCurrentIndex(m_ui->drawStyleComboBox->findData(partSettings->drawStyle()));
            m_ui->drawStyleComboBox->blockSignals(false);

            QPalette palette = m_ui->selectFixedColorFrame->palette();
            palette.setColor(m_ui->selectFixedColorFrame->backgroundRole(), cee::qt::UtilsGui::toQColor(partSettings->color()));
            m_ui->selectFixedColorFrame->setPalette(palette);

            m_ui->smoothShadingCheckBox->setChecked(partSettings->smoothShading());
            m_ui->showFringesCheckBox->setChecked(partSettings->fringesVisible());
            m_ui->showVectorsCheckBox->setChecked(partSettings->vectorsVisible());

            m_ui->settingsGroupBox->setEnabled(true);
        }
        else
        {
            m_ui->visibleCheckBox->setChecked(false);

            QPalette palette = m_ui->selectFixedColorFrame->palette();
            palette.setColor(m_ui->selectFixedColorFrame->backgroundRole(), Qt::darkGray);
            m_ui->selectFixedColorFrame->setPalette(palette);

            m_ui->smoothShadingCheckBox->setChecked(false);
            m_ui->showFringesCheckBox->setChecked(false);
            m_ui->showVectorsCheckBox->setChecked(false);

            m_ui->settingsGroupBox->setEnabled(false);
        }
    }
    else
    {
        setEnabled(false);

        m_ui->visibleCheckBox->setChecked(false);

        QPalette palette = m_ui->selectFixedColorFrame->palette();
        palette.setColor(m_ui->selectFixedColorFrame->backgroundRole(), Qt::darkGray);
        m_ui->selectFixedColorFrame->setPalette(palette);

        m_ui->smoothShadingCheckBox->setChecked(false);
        m_ui->showFringesCheckBox->setChecked(false);
        m_ui->showVectorsCheckBox->setChecked(false);
    }
}


//--------------------------------------------------------------------------------------------------
/// Update widgets to show state of selected part, if any.
/// Emits a signal to tell other that selection has changed and update visualization is needed.
//--------------------------------------------------------------------------------------------------
void PartsWidget::applySelection(const QModelIndex& current, const QModelIndex& previous)
{
    if (previous.isValid())
    {
        size_t geometryIndex = previous.data(Qt::UserRole + 1).value<size_t>();
        int partId           = previous.data(Qt::UserRole + 2).toInt();

        cee::ug::PartSettings* partSettings = m_unstructModel->partSettings(geometryIndex, partId);
        partSettings->setHighlight(false);
    }

    if (current.isValid())
    {
        size_t geometryIndex = current.data(Qt::UserRole + 1).value<size_t>();
        int partId           = current.data(Qt::UserRole + 2).toInt();

        cee::ug::PartSettings* partSettings = m_unstructModel->partSettings(geometryIndex, partId);
        partSettings->setHighlight(true);

        emit selected();
    }

    updatePanel();
}


//--------------------------------------------------------------------------------------------------
/// Show/hide part
//--------------------------------------------------------------------------------------------------
void PartsWidget::setPartVisible(bool show)
{
    if (!m_ui->partsListView->currentIndex().isValid()) return;

    size_t geometryIndex = m_ui->partsListView->currentIndex().data(Qt::UserRole + 1).value<size_t>();
    int partId           = m_ui->partsListView->currentIndex().data(Qt::UserRole + 2).toInt();

    cee::ug::PartSettings* partSettings = m_unstructModel->partSettings(geometryIndex, partId);
    partSettings->setVisible(show);

    m_unstructModel->updateVisualization(cee::ug::UnstructGridModel::VISIBILITY);
    m_view->requestRedraw();
}


//--------------------------------------------------------------------------------------------------
/// Set opacity for the selected part
//--------------------------------------------------------------------------------------------------
void PartsWidget::setOpacity(int value)
{
    if (!m_ui->partsListView->currentIndex().isValid()) return;

    size_t geometryIndex = m_ui->partsListView->currentIndex().data(Qt::UserRole + 1).value<size_t>();
    int partId           = m_ui->partsListView->currentIndex().data(Qt::UserRole + 2).toInt();

    cee::ug::PartSettings* partSettings = m_unstructModel->partSettings(geometryIndex, partId);
    partSettings->setOpacity(static_cast<float>(value) / 10.0f);

    m_unstructModel->updateVisualization(cee::ug::UnstructGridModel::COLOR_AND_OPACITY);
    m_view->requestRedraw();
}


//--------------------------------------------------------------------------------------------------
/// Select draw style for the selected part
//--------------------------------------------------------------------------------------------------
void PartsWidget::setDrawStyle(int index)
{
    if (!m_ui->partsListView->currentIndex().isValid()) return;

    size_t geometryIndex = m_ui->partsListView->currentIndex().data(Qt::UserRole + 1).value<size_t>();
    int partId           = m_ui->partsListView->currentIndex().data(Qt::UserRole + 2).toInt();

    cee::ug::PartSettings* partSettings = m_unstructModel->partSettings(geometryIndex, partId);
    cee::ug::PartSettings::DrawStyle drawStyle = static_cast<cee::ug::PartSettings::DrawStyle>(m_ui->drawStyleComboBox->itemData(index).toInt());
    partSettings->setDrawStyle(drawStyle);

    m_unstructModel->updateVisualization();
    m_view->requestRedraw();
}


//--------------------------------------------------------------------------------------------------
/// Set color for the selected part
//--------------------------------------------------------------------------------------------------
void PartsWidget::selectFixedColor()
{
    if (!m_ui->partsListView->currentIndex().isValid()) return;

    size_t geometryIndex = m_ui->partsListView->currentIndex().data(Qt::UserRole + 1).value<size_t>();
    int partId           = m_ui->partsListView->currentIndex().data(Qt::UserRole + 2).toInt();

    cee::ug::PartSettings* partSettings = m_unstructModel->partSettings(geometryIndex, partId);

    QColor color = cee::qt::UtilsGui::toQColor(partSettings->color());
  
    QColor newColor = QColorDialog::getColor(color, this);
    if (!newColor.isValid()) return;
    if (color == newColor) return;

    partSettings->setColor(cee::qt::UtilsGui::toColor3f(newColor));

    QPalette palette = m_ui->selectFixedColorFrame->palette();
    palette.setColor(m_ui->selectFixedColorFrame->backgroundRole(), newColor);
    m_ui->selectFixedColorFrame->setPalette(palette);

    m_unstructModel->updateVisualization(cee::ug::UnstructGridModel::COLOR_AND_OPACITY);
    m_view->requestRedraw();
}


//--------------------------------------------------------------------------------------------------
/// Use smooth shading
//--------------------------------------------------------------------------------------------------
void PartsWidget::enableSmoothShading(bool enable)
{
    if (!m_ui->partsListView->currentIndex().isValid()) return;

    size_t geometryIndex = m_ui->partsListView->currentIndex().data(Qt::UserRole + 1).value<size_t>();
    int partId           = m_ui->partsListView->currentIndex().data(Qt::UserRole + 2).toInt();

    cee::ug::PartSettings* partSettings = m_unstructModel->partSettings(geometryIndex, partId);
    partSettings->setSmoothShading(enable);

    m_unstructModel->updateVisualization();
    m_view->requestRedraw();
}


//--------------------------------------------------------------------------------------------------
/// Show active scalar on selected part
//--------------------------------------------------------------------------------------------------
void PartsWidget::showFringes(bool show)
{
    if (!m_ui->partsListView->currentIndex().isValid()) return;

    size_t geometryIndex = m_ui->partsListView->currentIndex().data(Qt::UserRole + 1).value<size_t>();
    int partId           = m_ui->partsListView->currentIndex().data(Qt::UserRole + 2).toInt();

    cee::ug::PartSettings* partSettings = m_unstructModel->partSettings(geometryIndex, partId);
    partSettings->setFringesVisible(show);

    m_unstructModel->updateVisualization();
    m_view->requestRedraw();
}


//--------------------------------------------------------------------------------------------------
/// Show active vector on selected part
//--------------------------------------------------------------------------------------------------
void PartsWidget::showVectors(bool show)
{
    if (!m_ui->partsListView->currentIndex().isValid()) return;

    size_t geometryIndex = m_ui->partsListView->currentIndex().data(Qt::UserRole + 1).value<size_t>();
    int partId           = m_ui->partsListView->currentIndex().data(Qt::UserRole + 2).toInt();

    cee::ug::PartSettings* partSettings = m_unstructModel->partSettings(geometryIndex, partId);
    partSettings->setVectorsVisible(show);

    m_unstructModel->updateVisualization();
    m_view->requestRedraw();
}
