//##################################################################################################
//
//  Ceetron Desktop Components
//  QtDemoApp
//  
//  --------------------------------------------------------------------------------------------
//  Copyright (C) 2016, Ceetron AS
//  This is UNPUBLISHED PROPRIETARY SOURCE CODE of Ceetron AS. The contents of this file may 
//  not be disclosed to third parties, copied or duplicated in any form, in whole or in part,
//  without the prior written permission of Ceetron AS.
//##################################################################################################

#include "PlotWidget.h"

#include "ui_PlotWidget.h"

#include "ceeqtUtilsCore.h"
#include "ceeqtUtilsGui.h"

#include "CeeCore/PropertySetCollection.h"
#include "CeeCore/ImageResources.h"

#include "CeeVisualization/View.h"
#include "CeeVisualization/Overlay.h"

#include "CeeUnstructGrid/UnstructGridModel.h"
#include "CeeUnstructGrid/PropertyApplierPlot2d.h"
#include "CeeUnstructGrid/DataSourceVTFx.h"

#include "CeePlot2d/OverlayPlot.h"
#include "CeePlot2d/Curve.h"
#include "CeePlot2d/AxisSettings.h"

#include <QColorDialog>
#include <QTableWidget>
#include <QDialogButtonBox>

//==================================================================================================
//
// 
//
//==================================================================================================

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
PlotWidget::PlotWidget(QWidget* parent, Qt::WindowFlags f)
:   QWidget(parent, f),
    m_ui(new Ui::PlotWidget),
    m_unstructModel(NULL),
    m_view(NULL),
    m_overlayPlot(NULL),
    m_mouseClickEventFilter(new MouseClickEventFilter)
{
    m_ui->setupUi(this);
    m_ui->selectFixedColorFrame->installEventFilter(m_mouseClickEventFilter);

    connect(m_ui->showPlotCheckBox,         SIGNAL(clicked(bool)),              SLOT(showPlot(bool)));
    connect(m_ui->titleLineEdit,            SIGNAL(textEdited(const QString&)), SLOT(setPlotTitle(const QString&)));
    connect(m_ui->drawLegendCheckBox,       SIGNAL(clicked(bool)),              SLOT(showLegend(bool)));
    connect(m_ui->autoRangeCheckBox,        SIGNAL(clicked(bool)),              SLOT(enableAutoRange(bool)));
    connect(m_ui->xMinLineEdit,             SIGNAL(editingFinished()),          SLOT(applyRange()));
    connect(m_ui->xMaxLineEdit,             SIGNAL(editingFinished()),          SLOT(applyRange()));
    connect(m_ui->yMinLineEdit,             SIGNAL(editingFinished()),          SLOT(applyRange()));
    connect(m_ui->yMaxLineEdit,             SIGNAL(editingFinished()),          SLOT(applyRange()));
    connect(m_ui->curveListWidget,          SIGNAL(itemSelectionChanged()),     SLOT(updateCurveSettings()));
    connect(m_ui->addPushButton,            SIGNAL(clicked()),                  SLOT(addCurve()));
    connect(m_ui->editPushButton,           SIGNAL(clicked()),                  SLOT(editCurve()));
    connect(m_ui->deletePushButton,         SIGNAL(clicked()),                  SLOT(removeCurve()));
    connect(m_ui->curveTitleLineEdit,       SIGNAL(textEdited(const QString&)), SLOT(setCurveTitle(const QString&)));
    connect(m_mouseClickEventFilter,        SIGNAL(clicked()),                  SLOT(selectCurveColor()));

    // Make sure the group box title is disabled if the group box itself is disabled
    QPalette palette;
    palette.setColor(QPalette::Disabled, QPalette::WindowText, QApplication::palette().color(QPalette::Disabled, QPalette::WindowText));
    m_ui->curveGroupBox->setPalette(palette);

    updatePanel();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
PlotWidget::~PlotWidget()
{
    m_ui->selectFixedColorFrame->removeEventFilter(m_mouseClickEventFilter);
    delete m_mouseClickEventFilter;
}


//--------------------------------------------------------------------------------------------------
/// Set up widget with current meta data and selection from model and view
//--------------------------------------------------------------------------------------------------
void PlotWidget::setup(cee::vis::View* view, cee::ug::UnstructGridModel* model)
{
    m_unstructModel = model;
    m_view = view;

    if (m_overlayPlot == NULL)
    {
        // Create an empty plot
        cee::PtrRef<cee::vis::Font> normalFont = cee::vis::Font::createNormalFont();
        m_overlayPlot = new cee::plt::OverlayPlot(normalFont.get());
        m_overlayPlot->setSize(400, 300);

        m_view->overlay().addItem(m_overlayPlot.get(), cee::vis::OverlayItem::TOP_RIGHT, cee::vis::OverlayItem::VERTICAL);
    }
    m_overlayPlot->setVisible(false);

    bool plotFoundOnFile = createPlotFromProperties();
    if (!plotFoundOnFile)
    {
        createDummyPlotData();
    }

    updatePanel();

    m_view->requestRedraw();
}


//--------------------------------------------------------------------------------------------------
/// Update panel settings
//--------------------------------------------------------------------------------------------------
void PlotWidget::updatePanel()
{
    if (m_overlayPlot)
    {
        m_ui->showPlotCheckBox->blockSignals(true);
        m_ui->showPlotCheckBox->setChecked(m_overlayPlot->visible());
        m_ui->showPlotCheckBox->blockSignals(false);

        m_ui->titleLineEdit->blockSignals(true);
        m_ui->titleLineEdit->setText(cee::qt::UtilsCore::toQString(m_overlayPlot->title()));
        m_ui->titleLineEdit->blockSignals(false);

        m_ui->drawLegendCheckBox->blockSignals(true);
        m_ui->drawLegendCheckBox->setChecked(m_overlayPlot->drawCurveLegend());
        m_ui->drawLegendCheckBox->blockSignals(false);

        bool useAutoRange = m_overlayPlot->axisSettings().useAutoRange();
        {
            m_ui->autoRangeCheckBox->blockSignals(true);
            m_ui->autoRangeCheckBox->setChecked(useAutoRange);
            m_ui->autoRangeCheckBox->blockSignals(false);

            m_ui->xMinLineEdit->setEnabled(!useAutoRange);
            m_ui->xMaxLineEdit->setEnabled(!useAutoRange);
            m_ui->yMinLineEdit->setEnabled(!useAutoRange);
            m_ui->yMaxLineEdit->setEnabled(!useAutoRange);

            m_ui->xMinLineEdit->setText(QString::number(m_overlayPlot->axisSettings().xAxisRangeMinimum()));
            m_ui->xMaxLineEdit->setText(QString::number(m_overlayPlot->axisSettings().xAxisRangeMaximum()));
            m_ui->yMinLineEdit->setText(QString::number(m_overlayPlot->axisSettings().yAxisRangeMinimum()));
            m_ui->yMaxLineEdit->setText(QString::number(m_overlayPlot->axisSettings().yAxisRangeMaximum()));
        }

        updateCurveList();
    }
}


//--------------------------------------------------------------------------------------------------
/// Load plot from file (if file contains plot properties)
//--------------------------------------------------------------------------------------------------
bool PlotWidget::createPlotFromProperties()
{
    cee::PtrRef<cee::PropertySetCollection> propertyCollection = new cee::PropertySetCollection;
    cee::PtrRef<cee::ImageResources> imgResources = new cee::ImageResources;
    cee::ug::DataSourceVTFx* source = dynamic_cast<cee::ug::DataSourceVTFx*>(m_unstructModel->dataSource());
    if (source)
    {
        source->caseProperties(propertyCollection.get(), imgResources.get());
        cee::ug::PropertyApplierPlot2d propPlotApplier(propertyCollection.get(), source);
        if (propPlotApplier.plotCount() > 0)
        {
            m_view->overlay().removeItem(m_overlayPlot.get());
            cee::PtrRef<cee::plt::OverlayPlot> plot = propPlotApplier.createOverlayPlot(0, cee::vis::Font::createNormalFont().get()); // Only using first plot
            m_view->overlay().addItem(plot.get(), cee::vis::OverlayItem::TOP_RIGHT, cee::vis::OverlayItem::HORIZONTAL);
            m_overlayPlot = plot;

            return true;
        }
    }

    return false;
}


//-------------------------------------------------------------------------------- ------------------
/// Create dummy plot data
//--------------------------------------------------------------------------------------------------
void PlotWidget::createDummyPlotData()
{
    m_overlayPlot->removeAllCurves();
    m_overlayPlot->setTitle("Dummy plot");
    m_overlayPlot->axisSettings().setXAxisTitle("X Axis");
    m_overlayPlot->axisSettings().setYAxisTitle("Y Axis");
    m_overlayPlot->axisSettings().setUseAutoRange(true);
    m_overlayPlot->setBackgroundOpacity(0.5f);

    // Add some dummy data to show in the plot
    const double xValuesRaw[] = { 1, 2, 3, 4, 5 };
    const double yValuesRaw[] = { 1, 2, 3, 2, 3 };
    std::vector<double> xValues(xValuesRaw, xValuesRaw + sizeof(xValuesRaw)/sizeof(double));
    std::vector<double> yValues(yValuesRaw, yValuesRaw + sizeof(yValuesRaw)/sizeof(double));
    m_overlayPlot->addCurve("Dummy values", xValues, yValues);
}


//-------------------------------------------------------------------------------- ------------------
/// Toggle plot visibility
//--------------------------------------------------------------------------------------------------
void PlotWidget::showPlot(bool show)
{
    m_overlayPlot->setVisible(show);
    m_view->requestRedraw();
}


//-------------------------------------------------------------------------------- ------------------
/// Set plot title
//--------------------------------------------------------------------------------------------------
void PlotWidget::setPlotTitle(const QString& title)
{
    m_overlayPlot->setTitle(cee::qt::UtilsCore::toStr(title));
    m_view->requestRedraw();
}


//-------------------------------------------------------------------------------- ------------------
/// Toggle curve legend visibility
//--------------------------------------------------------------------------------------------------
void PlotWidget::showLegend(bool show)
{
    m_overlayPlot->setDrawCurveLegend(show);
    m_view->requestRedraw();
}


//-------------------------------------------------------------------------------- ------------------
/// Toggle auto range on axis
//--------------------------------------------------------------------------------------------------
void PlotWidget::enableAutoRange(bool use)
{
    m_overlayPlot->axisSettings().setUseAutoRange(use);

    updatePanel();

    m_view->requestRedraw();
}


//-------------------------------------------------------------------------------- ------------------
/// Set min/max axis values
//--------------------------------------------------------------------------------------------------
void PlotWidget::applyRange()
{
    double minX = m_ui->xMinLineEdit->text().toDouble();
    double maxX = m_ui->xMaxLineEdit->text().toDouble();
    double minY = m_ui->yMinLineEdit->text().toDouble();
    double maxY = m_ui->yMaxLineEdit->text().toDouble();
    m_overlayPlot->axisSettings().setXAxisRange(minX, maxX);
    m_overlayPlot->axisSettings().setYAxisRange(minY, maxY);

    m_view->requestRedraw();
}


//--------------------------------------------------------------------------------------------------
/// Update curve list
//--------------------------------------------------------------------------------------------------
void PlotWidget::updateCurveList()
{
    QListWidget* listWidget = m_ui->curveListWidget;

    listWidget->blockSignals(true);
    listWidget->clear();

    for (size_t i = 0; i < m_overlayPlot->curveCount(); i++)
    {
        listWidget->addItem(cee::qt::UtilsCore::toQString(m_overlayPlot->curve(i)->title()));
    }

    listWidget->blockSignals(false);

    updateCurveSettings();
}


//--------------------------------------------------------------------------------------------------
/// Update curve settings for selected curve
//--------------------------------------------------------------------------------------------------
void PlotWidget::updateCurveSettings()
{
    int curveIndex = m_ui->curveListWidget->currentRow();

    m_ui->curveGroupBox->setEnabled(curveIndex != -1);

    if (curveIndex != -1)
    {
        cee::plt::Curve* curve = m_overlayPlot->curve(curveIndex);

        m_ui->curveTitleLineEdit->setText(cee::qt::UtilsCore::toQString(curve->title()));

        QPalette palette = m_ui->selectFixedColorFrame->palette();
        palette.setColor(m_ui->selectFixedColorFrame->backgroundRole(), cee::qt::UtilsGui::toQColor(curve->color()));
        m_ui->selectFixedColorFrame->setPalette(palette);
    }
}


//--------------------------------------------------------------------------------------------------
/// Set curve title for selected curve
//--------------------------------------------------------------------------------------------------
void PlotWidget::setCurveTitle(const QString& title)
{
    int curveIndex = m_ui->curveListWidget->currentRow();
    if (curveIndex == -1) return;

    cee::plt::Curve* curve = m_overlayPlot->curve(curveIndex);
    curve->setTitle(cee::qt::UtilsCore::toStr(title));

    m_ui->curveListWidget->currentItem()->setText(title);

    m_view->requestRedraw();
}


//--------------------------------------------------------------------------------------------------
/// Set curve color for selected curve
//--------------------------------------------------------------------------------------------------
void PlotWidget::selectCurveColor()
{
    int curveIndex = m_ui->curveListWidget->currentRow();
    if (curveIndex == -1) return;

    cee::plt::Curve* curve = m_overlayPlot->curve(curveIndex);
    QColor color = cee::qt::UtilsGui::toQColor(curve->color());
    QColor newColor = QColorDialog::getColor(color, this);

    curve->setColor(cee::qt::UtilsGui::toColor3f(newColor));

    QPalette palette = m_ui->selectFixedColorFrame->palette();
    palette.setColor(m_ui->selectFixedColorFrame->backgroundRole(), newColor);
    m_ui->selectFixedColorFrame->setPalette(palette);

    m_view->requestRedraw();
}


//--------------------------------------------------------------------------------------------------
/// Add a curve
//--------------------------------------------------------------------------------------------------
void PlotWidget::addCurve()
{
    QDialog dialog(this);
    dialog.setLayout(new QVBoxLayout());
    QTableWidget tableWidget(&dialog);
    dialog.layout()->addWidget(&tableWidget);

    tableWidget.setColumnCount(2);
    tableWidget.setRowCount(10);

    QDialogButtonBox buttonBox(QDialogButtonBox::Ok | QDialogButtonBox::Cancel);
    dialog.layout()->addWidget(&buttonBox);
    connect(&buttonBox, SIGNAL(accepted()), &dialog, SLOT(accept()));
    connect(&buttonBox, SIGNAL(rejected()), &dialog, SLOT(reject()));

    if (dialog.exec() == QDialog::Accepted)
    {
        std::vector<double> xValues;
        std::vector<double> yValues;
        for (int i = 0; i < tableWidget.rowCount(); i++)
        {
            QTableWidgetItem* itemX = tableWidget.item(i, 0);
            QTableWidgetItem* itemY = tableWidget.item(i, 1);
            if (itemX && itemY)
            { 
                bool ok = false;
                double x = itemX->text().toDouble(&ok);
                double y = itemY->text().toDouble(&ok);
                if (ok)
                {
                    xValues.push_back(x);
                    yValues.push_back(y);
                }
            }
        }

        m_overlayPlot->addCurve("New", xValues, yValues);

        updatePanel();
        m_ui->curveListWidget->setCurrentRow(static_cast<int>(m_overlayPlot->curveCount()) - 1);
        m_view->requestRedraw();
    }
}


//--------------------------------------------------------------------------------------------------
/// Edit a curve
//--------------------------------------------------------------------------------------------------
void PlotWidget::editCurve()
{
    int currentRow = m_ui->curveListWidget->currentRow();
    if (currentRow >= 0)
    {
        QDialog dialog(this);
        dialog.setLayout(new QVBoxLayout());
        QTableWidget tableWidget(&dialog);
        tableWidget.setColumnCount(2);
        dialog.layout()->addWidget(&tableWidget);

        cee::plt::Curve* curve = m_overlayPlot->curve(static_cast<size_t>(currentRow));
        tableWidget.setRowCount(static_cast<int>(curve->valueCount()));

        QDialogButtonBox buttonBox(QDialogButtonBox::Ok | QDialogButtonBox::Cancel);
        dialog.layout()->addWidget(&buttonBox);
        connect(&buttonBox, SIGNAL(accepted()), &dialog, SLOT(accept()));
        connect(&buttonBox, SIGNAL(rejected()), &dialog, SLOT(reject()));

        for (size_t j = 0; j < curve->valueCount(); j++)
        {
            cee::Vec2d vec = curve->value(j);
            tableWidget.setItem(static_cast<int>(j), 0, new QTableWidgetItem(QString::number(vec.x())));
            tableWidget.setItem(static_cast<int>(j), 1, new QTableWidgetItem(QString::number(vec.y())));
        }

        if (dialog.exec() == QDialog::Accepted)
        {
            std::vector<double> xValues;
            std::vector<double> yValues;
            for (int i = 0; i < tableWidget.rowCount(); i++)
            {
                QTableWidgetItem* itemX = tableWidget.item(i, 0);
                QTableWidgetItem* itemY = tableWidget.item(i, 1);
                if (itemX && itemY)
                {
                    bool ok = false;
                    double x = itemX->text().toDouble(&ok);
                    double y = itemY->text().toDouble(&ok);
                    if (ok)
                    {
                        xValues.push_back(x);
                        yValues.push_back(y);
                    }
                }
            }
            curve->setValues(xValues, yValues);

            m_overlayPlot->axisSettings().setUseAutoRange(m_overlayPlot->axisSettings().useAutoRange());
            m_view->requestRedraw();
        }
    }
}


//--------------------------------------------------------------------------------------------------
/// Remove a curve
//--------------------------------------------------------------------------------------------------
void PlotWidget::removeCurve()
{
    int currentCurveIndex = m_ui->curveListWidget->currentRow();
    if (currentCurveIndex != -1)
    {
        m_overlayPlot->removeCurve(currentCurveIndex);

        updateCurveList();
        m_view->requestRedraw();
    }
}


