//##################################################################################################
//
//  Ceetron Desktop Components
//  QtDemoApp
//  
//  --------------------------------------------------------------------------------------------
//  Copyright (C) 2016, Ceetron AS
//  This is UNPUBLISHED PROPRIETARY SOURCE CODE of Ceetron AS. The contents of this file may 
//  not be disclosed to third parties, copied or duplicated in any form, in whole or in part,
//  without the prior written permission of Ceetron AS.
//##################################################################################################

#include "ReportWidget.h"

#include "ui_ReportWidget.h"

#include "ReportCaptureDialog.h"
#include "SendToCloudDialogs.h"

#include "ceeqtUtilsCore.h"
#include "ceeqtUtilsGui.h"

#include "CeeReport/Snapshot.h"
#include "CeeReport/ReportCreatorWord.h"
#include "CeeReport/ReportCreatorPowerPoint.h"
#include "CeeReport/ReportCreatorHtml.h"
#include "CeeReport/FieldValuesGenerator.h"

#include "CeeCore/PropertySetCollection.h"
#include "CeeCore/ImageResources.h"

#include "CeeUnstructGrid/VTFxFileBrowser.h"

#include "CeeExport/ExportVTFx.h"
#include "CeeExport/PropertyBuilderVTFx.h"

#include <QFileDialog>
#include <QDesktopServices>
#include <QIcon>
#include <QUrl>
#include <QSslSocket>
#include <QMessageBox>
#include <QSettings>
#include <QProgressDialog>
#include <QTemporaryFile>
#include <QHttpMultiPart>
#include <QNetworkAccessManager>
#include <QNetworkReply>

//==================================================================================================
//
// 
//
//==================================================================================================

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
ReportWidget::ReportWidget(QWidget* parent, Qt::WindowFlags f)
:   QWidget(parent, f),
    m_ui(new Ui::ReportWidget),
    m_repository(new cee::rep::Repository)
{
    m_ui->setupUi(this);

    m_ui->situationListWidget->setIconSize(QSize(48, 64));

    connect(m_ui->capturePushButton,        SIGNAL(clicked()),              SLOT(captureSnapshotFromView()));
    connect(m_ui->deletePushButton,         SIGNAL(clicked()),              SLOT(deleteSnapshot()));
    connect(m_ui->createReportPushButton,   SIGNAL(clicked()),              SLOT(createReport()));
    connect(m_ui->loadPushButton,           SIGNAL(clicked()),              SLOT(loadRepository()));
    connect(m_ui->savePushButton,           SIGNAL(clicked()),              SLOT(saveRepository()));
    connect(m_ui->uploadPushButton,         SIGNAL(clicked()),              SLOT(sendModelsToCloud()));
    connect(m_ui->situationListWidget,      SIGNAL(itemSelectionChanged()), SLOT(updateFromCurrent()));
    connect(m_ui->cloudRadioButton,         SIGNAL(toggled(bool)), m_ui->uploadPushButton, SLOT(setEnabled(bool)));


    updatePanel();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
ReportWidget::~ReportWidget()
{
}


//--------------------------------------------------------------------------------------------------
/// Set up widget with current meta data and selection from model and view
//--------------------------------------------------------------------------------------------------
void ReportWidget::setup(cee::vis::View* view, cee::ug::UnstructGridModel* model)
{
    m_unstructModel = model;
    m_view  = view;

    updatePanel();
}


//--------------------------------------------------------------------------------------------------
/// Update panel settings.
//--------------------------------------------------------------------------------------------------
void ReportWidget::updatePanel()
{
    int snapshotIndex = m_ui->situationListWidget->currentRow();
    if (snapshotIndex == -1)
    {
        m_ui->titleLineEdit->setText("");
        m_ui->descriptionTextEdit->setPlainText("");

        return;
    }

    const cee::rep::Snapshot* snapshot = m_repository->snapshot(static_cast<size_t>(snapshotIndex));
    if (!snapshot)
    {
        // Error
        return;
    }

    QString snapshotType;
    switch (snapshot->type())
    {
        case cee::rep::Snapshot::OBJECT_IMAGE: snapshotType = "Image";        break;
        case cee::rep::Snapshot::OBJECT_VTFX:  snapshotType = "VTFx Model";   break;
        case cee::rep::Snapshot::OBJECT_TABLE: snapshotType = "Tabular data"; break;
    }
    m_ui->typeLineEdit->setText(snapshotType);

    m_ui->descriptionTextEdit->setPlainText(cee::qt::UtilsCore::toQString(snapshot->description()));
    m_ui->titleLineEdit->setText(cee::qt::UtilsCore::toQString(snapshot->title()));
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void ReportWidget::updateRepositoryList()
{
    QListWidget* listWidget = m_ui->situationListWidget;

    listWidget->blockSignals(true);
    listWidget->clear();

    for (size_t i = 0; i < m_repository->snapshotCount(); i++)
    {
        const cee::rep::Snapshot* snapshot = m_repository->snapshot(i);
        QImage snapshotImage;
        QString snapshotType;
        switch (snapshot->type())
        {
            case cee::rep::Snapshot::OBJECT_IMAGE:
            {
                const cee::Image* image = snapshot->image();
                if (image)
                {
                    snapshotImage = cee::qt::UtilsGui::toQImage(*image);
                }
                else
                {
                    snapshotImage = QImage(":/reporter_logo_72x72.png");
                }
                snapshotType = "image";
                break;
            }
            case cee::rep::Snapshot::OBJECT_VTFX:
            {
                cee::ug::VTFxFileBrowser browser;
                browser.initialize(snapshot->modelVTFx());
                cee::Image* image = browser.caseSnapshot(0);
                if (image)
                {
                    snapshotImage = cee::qt::UtilsGui::toQImage(*image);
                }
                else
                {
                    snapshotImage = QImage(":/reporter_logo_72x72.png");
                }
                snapshotType = "vtfx";
                break;
            }
            case cee::rep::Snapshot::OBJECT_TABLE:
            {
                snapshotImage = QImage(":/table_64x48.png"); 
                snapshotType = "table";
                break;
            }
        }
        QListWidgetItem* pItem = new QListWidgetItem(QIcon(QPixmap::fromImage(snapshotImage)), QString("%1. snapshot\n(%2)").arg(i + 1).arg(snapshotType), listWidget);
        listWidget->addItem(pItem);

    }
    listWidget->blockSignals(false);
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void ReportWidget::captureSnapshotFromView()
{
    ReportCaptureDialog dialog(this, windowFlags());

    if (dialog.exec() == QDialog::Accepted)
    {
        QApplication::setOverrideCursor(Qt::WaitCursor);

        cee::PtrRef<cee::rep::Snapshot> snapshot = new cee::rep::Snapshot(cee::rep::Snapshot::OBJECT_VTFX);
        cee::exp::ExportVTFx exporter(cee::exp::ExportVTFx::DISPLAY_MODEL_ONLY, m_unstructModel.get());
        exporter.setVendorNameAndApplication("Ceetron", "CDC Demo App Qt/C++");

        cee::PtrRef<cee::Image> snapshotImage = new cee::Image();
        if (m_view->renderToImage(1600, 1200, snapshotImage.get()))
        {
            exporter.setSnapshotImage(snapshotImage.get());
        }

        cee::PtrRef<cee::PropertySetCollection> propColl = new cee::PropertySetCollection();
        cee::PtrRef<cee::ImageResources> resources = new cee::ImageResources();
        cee::exp::PropertyBuilderVTFx propBuilder(propColl.get(), resources.get());
        propBuilder.addFromModel(*m_unstructModel.get());
        propBuilder.addFromView(*m_view);
        propBuilder.addDefaultFramesPerSecond(1.0);
        exporter.addProperties(*propColl);
        exporter.addResources(*resources);

        cee::PtrRef<cee::ug::VTFxMemoryFile> memoryFile = new cee::ug::VTFxMemoryFile();
        if (exporter.saveCase(memoryFile.get()))
        {
            snapshot->setModelVTFx(memoryFile.get());
            snapshot->setTitle(cee::qt::UtilsCore::toStr(m_ui->titleLineEdit->text()));
            snapshot->setDescription(cee::qt::UtilsCore::toStr(m_ui->descriptionTextEdit->toPlainText()));
            snapshot->setFieldValues(cee::rep::FieldValuesGenerator::getFieldValuesFromModel(m_unstructModel.get()));

            m_repository->addSnapshot(snapshot.get());

            updateRepositoryList();

            m_ui->situationListWidget->setCurrentRow(static_cast<int>(m_repository->snapshotCount()) - 1);
        }

        QApplication::restoreOverrideCursor();
    }
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void ReportWidget::updateFromCurrent()
{
    if (m_ui->situationListWidget->selectedItems().size() == 0)
    {
        return;
    }

    updatePanel();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void ReportWidget::createReport()
{
    QString filter = "All supported formats (*.docx *.pptx *.html *.htm);;Word documents (*.docx);;PowerPoint presentations (*.pptx);;HTML documents (*.html *.htm)";
    QString templateName = QFileDialog::getOpenFileName(this, tr("Open Template"), "", filter);
    if (!templateName.isEmpty())
    {
        QApplication::setOverrideCursor(Qt::WaitCursor);

        bool sendToCloud = m_ui->cloudRadioButton->isChecked();
        QString suffix = QFileInfo(templateName).suffix().toLower();
        if (suffix == "docx")
        {
            cee::rep::ReportCreatorWord generator(m_repository.get(), cee::qt::UtilsCore::toStr(templateName));
            generator.setUse3dPlugin(!sendToCloud);
            QString reportName = QFileDialog::getSaveFileName(this, tr("Save Word Report"), QFileInfo(templateName).baseName() + "_report", "Word documents(*.docx)");
            if (generator.generate(cee::qt::UtilsCore::toStr(reportName)))
            {
                showReport(cee::qt::UtilsCore::toStr(reportName));
            }
        }
        else if (suffix == "pptx")
        {
            cee::rep::ReportCreatorPowerPoint generator(m_repository.get(), cee::qt::UtilsCore::toStr(templateName));
            generator.setUse3dPlugin(!sendToCloud);
            QString reportName = QFileDialog::getSaveFileName(this, tr("Save PowerPoint Report"), QFileInfo(templateName).baseName() + "_report", "PowerPoint presentations(*.pptx)");
            if (generator.generate(cee::qt::UtilsCore::toStr(reportName)))
            {
                showReport(cee::qt::UtilsCore::toStr(reportName));
            }
        }
        else
        {
            cee::rep::ReportCreatorHtml generator(m_repository.get(), cee::qt::UtilsCore::toStr(templateName));
            generator.setUse3dPlugin(!sendToCloud);
            QString reportName = QFileDialog::getSaveFileName(this, tr("Save HTML Report"), QFileInfo(templateName).baseName() + "_report", "HTML documents(*.html *.htm)");
            if (generator.generate(cee::qt::UtilsCore::toStr(reportName)))
            {
                showReport(cee::qt::UtilsCore::toStr(reportName));
            }
        }
        QApplication::restoreOverrideCursor();

    }
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void ReportWidget::showReport(cee::Str reportName) const
{
    QFileInfo fileInfo(cee::qt::UtilsCore::toQString(reportName));
    QString file = "file:///" + fileInfo.absoluteFilePath();

    QUrl url = QUrl(file, QUrl::TolerantMode);
    QDesktopServices::openUrl(url);
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void ReportWidget::deleteSnapshot()
{
    size_t snapshotIndex = m_ui->situationListWidget->currentRow();

    m_repository->removeSnapshot(snapshotIndex);

    updateRepositoryList();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void ReportWidget::loadRepository()
{
    QString filter = "Repository files (*.repx)";
    QString fileName = QFileDialog::getOpenFileName(this, tr("Open Repository File"), "", filter);
    if (!fileName.isEmpty())
    {
        if (QFileInfo(fileName).suffix().toLower() == "repx")
        {
            m_repository = new cee::rep::Repository();
            if (m_repository->loadRepository(cee::qt::UtilsCore::toStr(fileName)))
            {
                updateRepositoryList();
            }
        }
    }
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void ReportWidget::saveRepository()
{
    QString filter = "Repository files (*.repx)";
    QString fileName = QFileDialog::getSaveFileName(this, tr("Save Repository File"), "", filter);
    if (!fileName.isEmpty())
    {
        m_repository->saveRepository(cee::qt::UtilsCore::toStr(fileName));
    }
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void ReportWidget::sendModelsToCloud()
{
    // Start with first snapshot in repository
    m_snapshotIndexToUploadToCloud = 0;

#ifndef QT_NO_OPENSSL
    // Check that we have SSL/HTTPS support through Qt (normally provided via OpenSSL as a dynamic library)
    if (!QSslSocket::supportsSsl())
    {
        QMessageBox::critical(this, "Error Sending to Cloud", "Your system and/or Qt version does not support secure connections (HTTPS).\nAre the OpenSSL dynamic libraries missing?");
        return;
    }
#else
    // There is no compile time SSL support in this Qt version
    // Either download another Qt version or recompile Qt with OpenSSl support
    QMessageBox::critical(this, "Error Sending to Cloud", "Your Qt version was not built with OpenSSL support.");
    return;
#endif

    // Note!
    // In a real application the upload ID should be kept secret. 
    // Storing it to registry/settings file in the following fashion should be avoided.
    QSettings settings;
    QString uploadId = settings.value("CloudUploadId").toString();
    if (uploadId.isEmpty())
    {
        EnterUploadIDDialog resDialog(this);
        resDialog.resize(600, 270);

        if (resDialog.exec() == QDialog::Accepted && resDialog.uploadId().length() > 1)
        {
            uploadId = resDialog.uploadId();
            settings.setValue("CloudUploadId", uploadId);
        }
    }

    if (uploadId.isEmpty())
    {
        return;
    }

    m_progressDialog = new QProgressDialog("Preparing file for upload...", "", 0, 100);
    m_progressDialog->setCancelButton(NULL);
    m_progressDialog->setMinimumDuration(0);
    m_progressDialog->show();
    m_progressDialog->setValue(1);

    qApp->processEvents(QEventLoop::ExcludeUserInputEvents);

    uploadNextSnapshot();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void ReportWidget::uploadNextSnapshot()
{
    if (m_repository->snapshotCount() <= m_snapshotIndexToUploadToCloud)
    {
        // All uploaded
        return;
    }

    cee::rep::Snapshot* snapshot = m_repository->snapshot(m_snapshotIndexToUploadToCloud);
    Q_ASSERT(snapshot);

    while (snapshot->type() != cee::rep::Snapshot::OBJECT_VTFX)
    {
        m_snapshotIndexToUploadToCloud++;

        if (m_repository->snapshotCount() <= m_snapshotIndexToUploadToCloud)
        {
            // All uploaded
            return;
        }

        snapshot = m_repository->snapshot(m_snapshotIndexToUploadToCloud);
    }

    sendToCloud(snapshot);
}


//--------------------------------------------------------------------------------------------------
/// Send the model to Ceetron cloud
//--------------------------------------------------------------------------------------------------
void ReportWidget::sendToCloud(const cee::rep::Snapshot* snapshot)
{
    if (!snapshot) return;
    if (!snapshot->modelVTFx()) return;

    QTemporaryFile tmpFile(QDir::tempPath() + "/uploadFile.XXXXXX");
    tmpFile.open();
    tmpFile.setAutoRemove(false);
    QString tempFilename = tmpFile.fileName();
    tempFilename += ".vtfx";

    snapshot->modelVTFx()->save(cee::qt::UtilsCore::toStr(tempFilename));

    // Send to cloud!
    QHttpMultiPart *multiPart = new QHttpMultiPart(QHttpMultiPart::FormDataType);
    multiPart->setBoundary("boundary_Ceetron89906a4aJn1kvlAG52SFG2AD");

    QHttpPart vtfxPart;
    vtfxPart.setHeader(QNetworkRequest::ContentTypeHeader, QVariant("application/octet-stream"));
    vtfxPart.setHeader(QNetworkRequest::ContentDispositionHeader, QVariant("form-data; name=\"vtfx\" filename=\"upload.vtfx\""));

    QFile *file = new QFile(tempFilename);
    if (!file->open(QIODevice::ReadOnly))
    {
        QMessageBox::critical(this, "Upload error", QString("Error opening temporary file %1").arg(tempFilename), QMessageBox::Ok);
        return;
    }


    vtfxPart.setBodyDevice(file);

    file->setParent(multiPart); // we cannot delete the file now, so delete it with the multiPart
    multiPart->append(vtfxPart);

    QString uploadApp = QUrl::toPercentEncoding("CDC Demo App Qt");

    QFileInfo fi(cee::qt::UtilsCore::toQString(snapshot->title())); 
    QString caseName = QUrl::toPercentEncoding(fi.fileName());

    QSettings settings;
    QString uploadId = settings.value("CloudUploadId").toString();
    QString urlString = QString("https://cloud.ceetron.com/api/v1/models?uploadId=%1&uploadApp=%2&name=%3").arg(uploadId).arg(uploadApp).arg(caseName);

    QUrl url(urlString);
    QNetworkRequest request(url);

    QNetworkAccessManager* manager = new QNetworkAccessManager(this);
    QNetworkReply* reply = manager->post(request, multiPart);
    multiPart->setParent(reply);

    connect(reply, SIGNAL(finished()), SLOT(cloudUploadFinished()));
    connect(reply, SIGNAL(uploadProgress(qint64, qint64)), SLOT(updateUploadProgress(qint64, qint64)));

    m_progressDialog->setLabelText("Uploading file to Ceetron Cloud...");
}


//--------------------------------------------------------------------------------------------------
/// Handle the end of the model upload to the cloud
//--------------------------------------------------------------------------------------------------
void ReportWidget::cloudUploadFinished()
{
    bool error = false;
    QString htmlDlgMsg;

    QNetworkReply* reply = (QNetworkReply*)sender();
    const int httpStatusCode = reply->attribute(QNetworkRequest::HttpStatusCodeAttribute).toInt();
    if (httpStatusCode > 0)
    {
        QString responseJson = reply->readAll();

        if (httpStatusCode == 200)
        {
            const QString ceetronCloudUrl = "https://cloud.ceetron.com/";
            const QString modelKey = parseTopLevelJsonValue(responseJson, "modelKey");
            QString viewerUrl = ceetronCloudUrl + "v/" + modelKey;

            cee::rep::Snapshot* snapshot = m_repository->snapshot(m_snapshotIndexToUploadToCloud);
            snapshot->setCloudViewerUrl(cee::qt::UtilsCore::toStr(viewerUrl));
        }
        else
        {
            const QString failReason = parseTopLevelJsonValue(responseJson, "message");
            const QString apiErrorCode = parseTopLevelJsonValue(responseJson, "apiErrorCode");
            htmlDlgMsg = QString("<h1>Upload Failed</h1><p>%1 <br> httpStatusCode:%2 <br> apiErrorCode:%3</p>").arg(failReason).arg(httpStatusCode).arg(apiErrorCode);

            error = true;
        }
    }
    else
    {
        htmlDlgMsg = QString("<h1>Network Error During Upload</h1><p>%1</p>").arg(reply->errorString());
        error = true;
    }

    reply->deleteLater();

    if (error)
    {
        UploadResultDialog resDialog(this, htmlDlgMsg, "Upload to CeetronCloud");
        resDialog.resize(650, 250);
        resDialog.exec();
    }
    else
    {
        bool moreSnapshotsLeft = false;
        {
            m_snapshotIndexToUploadToCloud++;

            while (m_snapshotIndexToUploadToCloud < m_repository->snapshotCount())
            {
                if (m_repository->snapshot(m_snapshotIndexToUploadToCloud)->type() == cee::rep::Snapshot::OBJECT_VTFX)
                {
                    moreSnapshotsLeft = true;
                    break;
                }
                m_snapshotIndexToUploadToCloud++;
            }
        }

        if (moreSnapshotsLeft)
        {
            uploadNextSnapshot();
        }
        else
        {
            htmlDlgMsg = "<h1>Upload Successful</h1>The links to the models will be added to the images in the generated report.</p>";
            UploadResultDialog resDialog(this, htmlDlgMsg, "Upload to CeetronCloud");
            resDialog.resize(650, 250);
            resDialog.exec();
        }
    }
}


//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void ReportWidget::updateUploadProgress(qint64 bytesSent, qint64 bytesTotal)
{
    if (m_progressDialog)
    {
        double progressPct = bytesTotal > 0 ? 100.0*static_cast<double>(bytesSent) / static_cast<double>(bytesTotal) : 100.0;
        m_progressDialog->setValue(static_cast<int>(progressPct));
    }
}


//--------------------------------------------------------------------------------------------------
/// Static helper function to do crude parsing of top level JSON values
/// Should be replaced with QJsonDocument for Qt5
//--------------------------------------------------------------------------------------------------
QString ReportWidget::parseTopLevelJsonValue(const QString json, const QString key)
{
    QString valString = "";

    const QString quotedKey = QString("\"%1\"").arg(key);
    const int keyIdx = json.indexOf(quotedKey);
    if (keyIdx >= 0)
    {
        int startQuote = json.indexOf("\"", keyIdx + quotedKey.length());
        int endQuote = json.indexOf("\"", startQuote + 1);

        if (startQuote >= 0 && endQuote > (startQuote + 1))
        {
            valString = json.mid(startQuote + 1, endQuote - startQuote - 1);
        }
    }

    return valString;
}