//##################################################################################################
//
//  Ceetron Desktop Components
//  QtDemoApp
//  
//  --------------------------------------------------------------------------------------------
//  Copyright (C) 2014, Ceetron AS
//  This is UNPUBLISHED PROPRIETARY SOURCE CODE of Ceetron AS. The contents of this file may 
//  not be disclosed to third parties, copied or duplicated in any form, in whole or in part,
//  without the prior written permission of Ceetron AS.
//##################################################################################################

#include "VectorResultsWidget.h"

#include "ui_VectorResultsWidget.h"

#include "MouseClickEventFilter.h"

#include "CeeVisualization/View.h"

#include "CeeUnstructGrid/ColorMapper.h"
#include "CeeUnstructGrid/DataSource.h"
#include "CeeUnstructGrid/DataSourceDirectory.h"
#include "CeeUnstructGrid/ModelSpec.h"
#include "CeeUnstructGrid/ResultInfo.h"
#include "CeeUnstructGrid/UnstructGridModel.h"
#include "CeeUnstructGrid/VectorSettings.h"

#include "ceeqtUtilsCore.h"
#include "ceeqtUtilsGui.h"

#include <QColorDialog>
#include <QLineEdit>
#include <QListView>
#include <QModelIndex>
#include <QStandardItemModel>


//==================================================================================================
//
// Manages vector settings
//
//==================================================================================================

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
VectorResultsWidget::VectorResultsWidget(QWidget* parent, Qt::WindowFlags f)
:   QWidget(parent, f),
    m_ui(new Ui::VectorResultsWidget),
    m_listModel(new QStandardItemModel(this)),
    m_mouseClickEventFilter(new MouseClickEventFilter)
{
    m_ui->setupUi(this);
    m_ui->vectorResultsListView->setModel(m_listModel);
    m_ui->selectFixedColorFrame->installEventFilter(m_mouseClickEventFilter);

    m_ui->colorModeComboBox->addItem(tr("Single"),           cee::ug::VectorSettings::SINGLE);
    m_ui->colorModeComboBox->addItem(tr("Color by fringes"), cee::ug::VectorSettings::COLOR_BY_FRINGES);

    connect(m_ui->vectorResultsListView,      SIGNAL(clicked(const QModelIndex&)), SLOT(applySelection(const QModelIndex&)));
    connect(m_ui->useRelativeScalingCheckBox, SIGNAL(clicked(bool)),               SLOT(enableRelativeScaling(bool)));
    connect(m_ui->scaleFactorLineEdit,        SIGNAL(editingFinished()),           SLOT(applyScaleFactor()));
    connect(m_ui->colorModeComboBox,          SIGNAL(currentIndexChanged(int)),    SLOT(setColorMode(int)));
    connect(m_mouseClickEventFilter,          SIGNAL(clicked()),                   SLOT(selectFixedColor()));

    // Make sure the group box title is disabled if the group box itself is disabled
    QPalette palette;
    palette.setColor(QPalette::Disabled, QPalette::WindowText, QApplication::palette().color(QPalette::Disabled, QPalette::WindowText));
    m_ui->settingsGroupBox->setPalette(palette);

    updatePanel();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
VectorResultsWidget::~VectorResultsWidget()
{
    delete m_listModel;

    m_ui->selectFixedColorFrame->removeEventFilter(m_mouseClickEventFilter);
    delete m_mouseClickEventFilter;
}


//--------------------------------------------------------------------------------------------------
/// Set up widget with current meta data and selection from model and view
//--------------------------------------------------------------------------------------------------
void VectorResultsWidget::setup(cee::vis::View* view, cee::ug::UnstructGridModel* model)
{
    m_unstructModel = model;
    m_view  = view;

    const cee::ug::DataSourceDirectory* directory = m_unstructModel->dataSource()->directory();
    Q_ASSERT(directory);

    // Get vector results meta data for the models data source
    {
        m_listModel->clear();

        std::vector<cee::ug::ResultInfo> vectorInfos = directory->vectorResultInfos();
        std::vector<cee::ug::ResultInfo>::const_iterator it;
        for (it = vectorInfos.begin(); it != vectorInfos.end(); ++it)
        {
            QStandardItem* item = new QStandardItem(cee::qt::UtilsCore::toQString(it->name()));
            item->setData(it->id(), Qt::UserRole + 1);

            m_listModel->appendRow(item);
        }
    }

    // Select vector result used in model
    {
        int resultId = -1;
        if (m_unstructModel->modelSpec().vectorResultIds().size() > 0)
        {
            resultId = m_unstructModel->modelSpec().vectorResultIds()[0];
        }

        int rowCount = m_listModel->rowCount();
        for (int i = 0; i < rowCount; ++i)
        {
            QStandardItem* item = m_listModel->item(i);
            int itemResultId = item->data(Qt::UserRole + 1).toInt();
            if (itemResultId == resultId)
            {
                m_ui->vectorResultsListView->setCurrentIndex(m_listModel->indexFromItem(item));
                break;
            }
        }
    }

    updatePanel();
}


//--------------------------------------------------------------------------------------------------
/// Update panel settings.
/// Clear if no vector result is available/selected.
//--------------------------------------------------------------------------------------------------
void VectorResultsWidget::updatePanel()
{
    if (m_unstructModel)
    {
        setEnabled(true);

        QModelIndex vectorResultIndex = m_ui->vectorResultsListView->currentIndex();
        if (vectorResultIndex.isValid())
        {
            int resultId = vectorResultIndex.data(Qt::UserRole + 1).toInt();
            cee::ug::VectorSettings* settings = m_unstructModel->vectorSettings(resultId);

            m_ui->useRelativeScalingCheckBox->setChecked(settings->scaleMode() == cee::ug::VectorSettings::RELATIVE_SCALING);
            m_ui->scaleFactorLineEdit->setText(QString::number(settings->scaleFactor()));

            m_ui->colorModeComboBox->blockSignals(true);
            m_ui->colorModeComboBox->setCurrentIndex(m_ui->colorModeComboBox->findData(settings->colorMode()));
            m_ui->colorModeComboBox->blockSignals(false);

            QPalette palette = m_ui->selectFixedColorFrame->palette();
            palette.setColor(m_ui->selectFixedColorFrame->backgroundRole(), cee::qt::UtilsGui::toQColor(settings->singleVectorColor()));
            m_ui->selectFixedColorFrame->setPalette(palette);

            m_ui->settingsGroupBox->setEnabled(true);
        }
        else
        {
            m_ui->useRelativeScalingCheckBox->setChecked(false);
            m_ui->scaleFactorLineEdit->setText("");

            QPalette palette = m_ui->selectFixedColorFrame->palette();
            palette.setColor(m_ui->selectFixedColorFrame->backgroundRole(), Qt::darkGray);
            m_ui->selectFixedColorFrame->setPalette(palette);

            m_ui->settingsGroupBox->setEnabled(false);
        }
    }
    else
    {
        setEnabled(false);

        m_ui->useRelativeScalingCheckBox->setChecked(false);
        m_ui->scaleFactorLineEdit->setText("");

        QPalette palette = m_ui->selectFixedColorFrame->palette();
        palette.setColor(m_ui->selectFixedColorFrame->backgroundRole(), Qt::darkGray);
        m_ui->selectFixedColorFrame->setPalette(palette);
    }
}


//--------------------------------------------------------------------------------------------------
/// Update widgets to show state of selected vector result
//--------------------------------------------------------------------------------------------------
void VectorResultsWidget::applySelection(const QModelIndex& index)
{
    Q_UNUSED(index);
    updatePanel();
}


//--------------------------------------------------------------------------------------------------
/// Enable/disable relative scaling
//--------------------------------------------------------------------------------------------------
void VectorResultsWidget::enableRelativeScaling(bool enable)
{
    QModelIndex vectorResultIndex = m_ui->vectorResultsListView->currentIndex();
    if (!vectorResultIndex.isValid()) return;

    int resultId = vectorResultIndex.data(Qt::UserRole + 1).toInt();
    cee::ug::VectorSettings* settings = m_unstructModel->vectorSettings(resultId);

    // Set default scale factor
    if (enable)
    {
        settings->setScaleFactor(cee::ug::VectorSettings::RELATIVE_SCALING, 0.2);
    }
    else
    {
        double defaultScale = m_unstructModel->boundingBox().extent().length() * 0.2;
        settings->setScaleFactor(cee::ug::VectorSettings::ABSOLUTE_SCALING, defaultScale);
    }

    m_unstructModel->updateVisualization();
    m_view->requestRedraw();

    updatePanel();
}


//--------------------------------------------------------------------------------------------------
/// Apply vector scale factor
//--------------------------------------------------------------------------------------------------
void VectorResultsWidget::applyScaleFactor()
{
    QModelIndex vectorResultIndex = m_ui->vectorResultsListView->currentIndex();
    if (!vectorResultIndex.isValid()) return;

    int resultId = vectorResultIndex.data(Qt::UserRole + 1).toInt();
    cee::ug::VectorSettings* settings = m_unstructModel->vectorSettings(resultId);

    bool ok = false;
    double scaleFactor = m_ui->scaleFactorLineEdit->text().toDouble(&ok);
    if (ok)
    {
        if (m_ui->useRelativeScalingCheckBox->isChecked())
        {
            settings->setScaleFactor(cee::ug::VectorSettings::RELATIVE_SCALING, scaleFactor);
        }
        else
        {
            settings->setScaleFactor(cee::ug::VectorSettings::ABSOLUTE_SCALING, scaleFactor);
        }

        m_unstructModel->updateVisualization();
        m_view->requestRedraw();
    }
}


//--------------------------------------------------------------------------------------------------
/// Change vector coloring mode
//--------------------------------------------------------------------------------------------------
void VectorResultsWidget::setColorMode(int index)
{
    QModelIndex vectorResultIndex = m_ui->vectorResultsListView->currentIndex();
    if (!vectorResultIndex.isValid()) return;

    int resultId = vectorResultIndex.data(Qt::UserRole + 1).toInt();
    cee::ug::VectorSettings* settings = m_unstructModel->vectorSettings(resultId);

    int colorMode = m_ui->colorModeComboBox->itemData(index).toInt();
    settings->setColorMode(static_cast<cee::ug::VectorSettings::ColorMode>(colorMode));
    
    m_unstructModel->updateVisualization();
    m_view->requestRedraw();
}


//--------------------------------------------------------------------------------------------------
/// Set vector color
//--------------------------------------------------------------------------------------------------
void VectorResultsWidget::selectFixedColor()
{
    QModelIndex vectorResultIndex = m_ui->vectorResultsListView->currentIndex();
    if (!vectorResultIndex.isValid()) return;

    int resultId = vectorResultIndex.data(Qt::UserRole + 1).toInt();
    cee::ug::VectorSettings* settings = m_unstructModel->vectorSettings(resultId);

    QColor color = cee::qt::UtilsGui::toQColor(settings->singleVectorColor());

    QColor newColor = QColorDialog::getColor(color, this);
    if (!newColor.isValid()) return;
    if (color == newColor) return;

    settings->setSingleVectorColor(cee::qt::UtilsGui::toColor3f(newColor));

    QPalette palette = m_ui->selectFixedColorFrame->palette();
    palette.setColor(m_ui->selectFixedColorFrame->backgroundRole(), newColor);
    m_ui->selectFixedColorFrame->setPalette(palette);

    m_unstructModel->updateVisualization();
    m_view->requestRedraw();
}
