//##################################################################################################
//
//  Ceetron Desktop Components
//  QtMultiView
//  
//  --------------------------------------------------------------------------------------------
//  Copyright (C) 2011, Ceetron AS
//  This is UNPUBLISHED PROPRIETARY SOURCE CODE of Ceetron AS. The contents of this file may 
//  not be disclosed to third parties, copied or duplicated in any form, in whole or in part,
//  without the prior written permission of Ceetron AS.
//##################################################################################################

#include "QMVMainWindow.h"

#include "CeeCore/Instance.h"

#include "CeeVisualization/VisualizationComponent.h"
#include "CeeVisualization/Camera.h"
#include "CeeVisualization/OverlayAxisCross.h"
#include "CeeVisualization/OverlayTextBox.h"
#include "CeeVisualization/Overlay.h"
#include "CeeVisualization/Font.h"
#include "CeeVisualization/View.h"

#include "CeeGeometry/GeometryModel.h"
#include "CeeGeometry/DataGenerator.h"
#include "CeeGeometry/Part.h"
#include "CeeGeometry/PartSettings.h"
#include "CeeGeometry/EffectHalo.h"
#include "CeeGeometry/EffectColor.h"

#include "CeeUnstructGrid/DataSourceMemory.h"
#include "CeeUnstructGrid/UnstructGridModel.h"
#include "CeeUnstructGrid/ModelSpec.h"
#include "CeeUnstructGrid/PartSettings.h"

#include "ceeqtUtilsCore.h"

#include <QFrame>
#include <QGridLayout>
#include <QAction>
#include <QMenu>
#include <QMenuBar>




//==================================================================================================
//
// 
//
//==================================================================================================

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
QMVMainWindow::QMVMainWindow(cee::vis::OpenGLContextGroup* contextGroup)
:   m_openGLContextGroup(contextGroup)
{
    QFrame* mainFrame = new QFrame;
    QGridLayout* frameLayout = new QGridLayout;
    mainFrame->setLayout(frameLayout);
    setCentralWidget(mainFrame);

    QMenu* menu = menuBar()->addMenu("Tests");
    menu->addAction("Create 1 Viewer", this, SLOT(slotCreate1Viewer()));
    menu->addAction("Create 2 Viewers", this, SLOT(slotCreate2Viewers()));
    menu->addAction("Create 3 Viewers", this, SLOT(slotCreate3Viewers()));
    menu->addSeparator();
    menu->addAction("Create 2 Viewers, Recycle Views", this, SLOT(slotCreate2ViewersRecycleViews()));
    menu->addAction("Create 3 Viewers, Recycle Views", this, SLOT(slotCreate3ViewersRecycleViews()));
    menu->addSeparator();
    menu->addAction("Populate All Views With Same Model", this, SLOT(slotPopulateViewsWithSameModel()));
    menu->addAction("Populate Views With Individual Models", this, SLOT(slotPopulateViewsWithIndividualModels()));
    menu->addSeparator();
    menu->addAction("Clear Contents Of All Views", this, SLOT(slotClearAllViews()));
    menu->addAction("Delete All Views", this, SLOT(slotDeleteAllViews()));
    menu->addAction("Delete All Viewers", this, SLOT(slotDeleteAllViewers()));

    createViewers(3, NEW_VIEWS);
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QMVMainWindow::slotCreate1Viewer()
{
    createViewers(1, NEW_VIEWS);
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QMVMainWindow::slotCreate2Viewers()
{
    createViewers(2, NEW_VIEWS);
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QMVMainWindow::slotCreate3Viewers()
{
    createViewers(3, NEW_VIEWS);
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QMVMainWindow::slotCreate3ViewersRecycleViews()
{
    createViewers(3, RECYCLE_VIEWS);
    redrawAllViewers();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QMVMainWindow::slotCreate2ViewersRecycleViews()
{
    createViewers(2, RECYCLE_VIEWS);
    redrawAllViewers();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QMVMainWindow::slotPopulateViewsWithSameModel()
{
    clearViewContents();
    populateViewsWithSingleModel();
    redrawAllViewers();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QMVMainWindow::slotPopulateViewsWithIndividualModels()
{
    clearViewContents();
    populateViewsWithMixedModels();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QMVMainWindow::slotClearAllViews()
{
    clearViewContents();
    redrawAllViewers();
}



//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QMVMainWindow::slotDeleteAllViews()
{
    for (size_t i = 0; i < MAX_NUM_VIEWERS; i++)
    {
        cee::qt::ViewerOpenGLWidgetDemo* viewer = m_viewers[i];
        if (viewer)
        {
            viewer->setView(NULL);
        }
    }

    redrawAllViewers();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QMVMainWindow::slotDeleteAllViewers()
{
    deleteViewers();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QMVMainWindow::createViewers(size_t numViewers, ViewCreation viewCreation)
{
    Q_ASSERT(numViewers > 0 && numViewers <= MAX_NUM_VIEWERS);

    std::vector<cee::PtrRef<cee::vis::View> > currentViews;
    if (viewCreation == RECYCLE_VIEWS)
    {
        // No viewers should now be present
        for (size_t i = 0; i < viewCount(); i++)
        {
            currentViews.push_back(view(i));
        }
    }

    // Delete any existing viewers
    deleteViewers();

    // No viewers should now be present
    for (size_t i = 0; i < MAX_NUM_VIEWERS; i++)
    {
        Q_ASSERT(m_viewers[i].isNull());
    }


    QWidget* parentWidget = centralWidget();
    QGridLayout* layout = dynamic_cast<QGridLayout*>(parentWidget->layout());
    Q_ASSERT(layout);

    cee::qt::ViewerOpenGLWidgetDemo* firstViewer = NULL;
    if (numViewers > 0)
    {
        m_viewers[0] = new cee::qt::ViewerOpenGLWidgetDemo(m_openGLContextGroup.get(), parentWidget);
        int columnSpan = (numViewers == 3) ? 2 : 1;
        layout->addWidget(m_viewers[0], 0, 0, 1, columnSpan);
        firstViewer = m_viewers[0];
        firstViewer->setFocus();
    }

    if (numViewers > 1)
    {
        m_viewers[1] = new cee::qt::ViewerOpenGLWidgetDemo(m_openGLContextGroup.get(), parentWidget);
        layout->addWidget(m_viewers[1], 1, 0);
    }

    if (numViewers > 2)
    {
        m_viewers[2] = new cee::qt::ViewerOpenGLWidgetDemo(m_openGLContextGroup.get(), parentWidget);
        layout->addWidget(m_viewers[2], 1, 1);
    }

    for (size_t i = 0; i < numViewers; i++)
    {
        cee::qt::ViewerOpenGLWidgetDemo* viewer = m_viewers[i];
        if (viewer)
        {
            if (viewCreation == NEW_VIEWS)
            {
                cee::PtrRef<cee::vis::View> view = new cee::vis::View;
                viewer->setView(view.get());

                view->overlay().addItem(new cee::vis::OverlayAxisCross(&view->camera(), cee::vis::Font::createNormalFont().get()), cee::vis::OverlayItem::BOTTOM_LEFT, cee::vis::OverlayItem::HORIZONTAL);

                cee::vis::OverlayTextBox* textBox = new cee::vis::OverlayTextBox(cee::vis::Font::createNormalFont().get());
                textBox->setText(cee::qt::UtilsCore::toStr(QString("View %1").arg(i + 1)));
                textBox->setSizeToFitText();
                view->overlay().addItem(textBox, cee::vis::OverlayItem::TOP_LEFT, cee::vis::OverlayItem::HORIZONTAL);
            }
            else
            {
                if (i < currentViews.size())
                {
                    cee::PtrRef<cee::vis::View> view = currentViews[i];
                    viewer->setView(view.get());
                }
            }
        }
    }
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QMVMainWindow::deleteViewers()
{
    QWidget* parentWidget = centralWidget();
    QLayout* layout = parentWidget->layout();

    for (size_t i = 0; i < MAX_NUM_VIEWERS; i++)
    {
        cee::qt::ViewerOpenGLWidgetDemo* viewer = m_viewers[i];
        if (viewer)
        {
            layout->removeWidget(viewer);
            viewer->setView(NULL);
            delete viewer;
            m_viewers[i] = NULL;
        }
    }
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QMVMainWindow::populateViewsWithSingleModel()
{
    size_t numViews = viewCount();

    cee::PtrRef<cee::geo::GeometryModel> theModel = new cee::geo::GeometryModel;

    if (numViews > 0)
    {
        cee::vis::View* v = view(0);
        v->addModel(theModel.get());
    }

    {
        cee::geo::Part* part = new cee::geo::Part(cee::geo::DataGenerator::createBox(cee::Vec3d(0, 0, 0), cee::Vec3d(1, 0.75f, 0.5)).get());
        part->settings().addEffect(new cee::geo::EffectColor(cee::Color3f(0.00f, 0.53f, 0.58f)));
        theModel->addPart(part);
    }
    {
        cee::geo::Part* part = new cee::geo::Part(cee::geo::DataGenerator::createBox(cee::Vec3d(1.1, 0, 0), cee::Vec3d(1, 0.75f, 0.5)).get());
        part->settings().addEffect(new cee::geo::EffectColor(cee::Color3f(0.83f, 0.62f, 0.38f)));
        theModel->addPart(part);
    }
    {
        cee::geo::Part* part = new cee::geo::Part(cee::geo::DataGenerator::createBox(cee::Vec3d(0, 1.1, 0), cee::Vec3d(1, 0.75f, 0.5)).get());
        part->settings().addEffect(new cee::geo::EffectColor(cee::Color3f(0.85f, 0.32f, 0.11f)));
        theModel->addPart(part);
    }

    {
        cee::geo::Part* part = new cee::geo::Part(cee::geo::DataGenerator::createSphere(cee::Vec3d(1.1, 1.1, 0), 0.5, 10).get());
        part->settings().addEffect(new cee::geo::EffectColor(cee::Color3f(0.83f, 0.58f, 0.54f)));
        part->settings().addEffect(new cee::geo::EffectHalo);
        theModel->addPart(part);
    }

    if (numViews > 0)
    {
        cee::vis::View* v = view(0);
        cee::BoundingBox bb = v->boundingBox();
        v->camera().fitView(bb, cee::Vec3d(0, 0, -1), cee::Vec3d(0, 1, 0));
    }

    if (numViews > 1)
    {
        cee::vis::View* v = view(1);
        v->addModel(theModel.get());
        cee::BoundingBox bb = v->boundingBox();
        v->camera().fitView(bb, cee::Vec3d(-1, 0, 0), cee::Vec3d(0, 0, 1));
    }

    if (numViews > 2)
    {
        cee::vis::View* v = view(2);
        v->addModel(theModel.get());
        cee::BoundingBox bb = v->boundingBox();
        v->camera().fitView(bb, cee::Vec3d(0, -1, 0), cee::Vec3d(1, 0, 0));
    }
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QMVMainWindow::populateViewsWithMixedModels()
{
    const size_t numViews = viewCount();
    for (size_t i = 0; i < numViews; i++)
    {
        cee::vis::View* v = view(i);

        size_t m = i % 3;
        if      (m == 0) populateView_sphereAndBoxesHalo(v);
        else if (m == 1) populateView_unstructGridModel(v);
        else if (m == 2) populateView_cylinder(v);
    }

    redrawAllViewers();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QMVMainWindow::populateView_sphereAndBoxesHalo(cee::vis::View* view)
{
    cee::PtrRef<cee::geo::GeometryModel> boxModel = new cee::geo::GeometryModel;
    view->addModel(boxModel.get());
    {
        cee::geo::Part* part = new cee::geo::Part(cee::geo::DataGenerator::createBox(cee::Vec3d(0, 0, 0), cee::Vec3d(1, 0.75f, 0.5)).get());
        part->settings().addEffect(new cee::geo::EffectColor(cee::Color3f(0.8f, 0.2f, 0.2f)));
        boxModel->addPart(part);
    }
    {
        cee::geo::Part* part = new cee::geo::Part(cee::geo::DataGenerator::createBox(cee::Vec3d(1.1, 0, 0), cee::Vec3d(1, 0.75f, 0.5)).get());
        part->settings().addEffect(new cee::geo::EffectColor(cee::Color3f(0.2f, 0.7f, 0.2f)));
        boxModel->addPart(part);
    }
    {
        cee::geo::Part* part = new cee::geo::Part(cee::geo::DataGenerator::createBox(cee::Vec3d(0, 1.1, 0), cee::Vec3d(1, 0.75f, 0.5)).get());
        part->settings().addEffect(new cee::geo::EffectColor(cee::Color3f(0.2f, 0.2f, 0.7f)));
        boxModel->addPart(part);
    }

    cee::PtrRef<cee::geo::GeometryModel> sphereModel = new cee::geo::GeometryModel;
    view->addModel(sphereModel.get());
    {
        cee::geo::Part* part = new cee::geo::Part(cee::geo::DataGenerator::createSphere(cee::Vec3d(1.1, 1.1, 0), 0.5, 10).get());
        part->settings().addEffect(new cee::geo::EffectColor(cee::Color3f(0.7f, 0.7f, 0.0f)));
        part->settings().addEffect(new cee::geo::EffectHalo);
        sphereModel->addPart(part);
    }

    cee::BoundingBox bb = view->boundingBox();
    view->camera().fitView(bb, cee::Vec3d(0, 0, -1), cee::Vec3d(0, 1, 0));
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QMVMainWindow::populateView_unstructGridModel(cee::vis::View* view)
{
    cee::PtrRef<cee::ug::DataSourceMemory> dataSource = new cee::ug::DataSourceMemory(1, 1);
    {
        cee::PtrRef<cee::ug::DataNodes> nodes = new cee::ug::DataNodes(false);
        nodes->resize(7);
        nodes->setNode(0, cee::Vec3d(0, 0, 0));
        nodes->setNode(1, cee::Vec3d(2, 0, 0));
        nodes->setNode(2, cee::Vec3d(2, 2, 0));
        nodes->setNode(3, cee::Vec3d(0, 2, 0));
        nodes->setNode(4, cee::Vec3d(0, 3, 0));
        nodes->setNode(5, cee::Vec3d(2, 3, 0));
        nodes->setNode(6, cee::Vec3d(1, 5, 0));

        cee::PtrRef<cee::ug::DataElements> elements = new cee::ug::DataElements(false, 0);
        const unsigned int elemConnQuad[] = {0, 1, 2, 3};
        const unsigned int elemConnTri[] = {4, 5, 6};
        elements->addElement(cee::ug::Element::QUADS, elemConnQuad, 4);
        elements->addElement(cee::ug::Element::TRIANGLES, elemConnTri, 3);

        cee::PtrRef<cee::ug::DataPart> part = new cee::ug::DataPart(1);
        part->setNodes(nodes.get());
        part->setElements(elements.get());

        cee::PtrRef<cee::ug::DataGeometry> geo = new cee::ug::DataGeometry;
        geo->addPart(part.get());

        cee::PtrRef<cee::ug::DataState> state = new cee::ug::DataState(1, 1);
        state->setGeometry(0, geo.get());

        dataSource->addState(state.get());
        dataSource->updateDirectoryFromStates();
    }

    {
        cee::PtrRef<cee::ug::UnstructGridModel> model = new cee::ug::UnstructGridModel;
        view->addModel(model.get());
        model->setDataSource(dataSource.get());
        model->modelSpec().setStateId(1);
        model->partSettings(0, 1)->setColor(cee::Color3f(0.8f, 0.8f, 0));
        model->partSettings(0, 1)->setDrawStyle(cee::ug::PartSettings::SURFACE_MESH);
        model->updateVisualization();
    }

    cee::BoundingBox bb = view->boundingBox();
    view->camera().fitView(bb, cee::Vec3d(0, 0, -1), cee::Vec3d(0, 1, 0));
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QMVMainWindow::populateView_cylinder(cee::vis::View* view)
{
    cee::PtrRef<cee::geo::GeometryModel> model = new cee::geo::GeometryModel;
    view->addModel(model.get());

    cee::geo::Part* part = new cee::geo::Part(cee::geo::DataGenerator::createCylinder(cee::Vec3d(0, 0, 0), 1, 1, cee::Vec3d(0, 1, 0), 2, 20).get());
    part->settings().addEffect(new cee::geo::EffectColor(cee::Color3f(0.8f, 0.2f, 0.2f)));
    model->addPart(part);

    cee::BoundingBox bb = view->boundingBox();
    view->camera().fitView(bb, cee::Vec3d(0, 0, -1), cee::Vec3d(0, 1, 0));
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
size_t QMVMainWindow::viewCount()
{
    for (size_t i = 0; i < MAX_NUM_VIEWERS; i++)
    {
        cee::qt::ViewerOpenGLWidgetDemo* viewer = m_viewers[i];
        cee::vis::View* view = viewer ? viewer->view() : NULL;
        if (!view)
        {
            return i;
        }
    }

    return MAX_NUM_VIEWERS;
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
cee::vis::View* QMVMainWindow::view(size_t index)
{
    Q_ASSERT(index < MAX_NUM_VIEWERS);
    cee::qt::ViewerOpenGLWidgetDemo* viewer = m_viewers[index];
    if (viewer)
    {
        return viewer->view();
    }

    return NULL;
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QMVMainWindow::clearViewContents()
{
    for (size_t i = 0; i < MAX_NUM_VIEWERS; i++)
    {
        cee::qt::ViewerOpenGLWidgetDemo* viewer = m_viewers[i];
        cee::vis::View* view = viewer ? viewer->view() : NULL;
        if (view)
        {
            view->removeAllModels();
        }
    }
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QMVMainWindow::redrawAllViewers()
{
    for (size_t i = 0; i < MAX_NUM_VIEWERS; i++)
    {
        cee::qt::ViewerOpenGLWidgetDemo* viewer = m_viewers[i];
        if (viewer)
        {
            viewer->update();
        }
    }
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QMVMainWindow::closeEvent(QCloseEvent*)
{
    deleteViewers();
}




