//##################################################################################################
//
//  Ceetron Desktop Components
//  QtReport
//  
//  --------------------------------------------------------------------------------------------
//  Copyright (C) 2015, Ceetron AS
//  This is UNPUBLISHED PROPRIETARY SOURCE CODE of Ceetron AS. The contents of this file may 
//  not be disclosed to third parties, copied or duplicated in any form, in whole or in part,
//  without the prior written permission of Ceetron AS.
//##################################################################################################

#include "QRMainWindow.h"

#include "CeeVisualization/View.h"
#include "CeeVisualization/VisualizationComponent.h"
#include "CeeVisualization/Camera.h"
#include "CeeVisualization/CameraInputHandler.h"

#include "CeeCore/PropertySetCollection.h"
#include "CeeCore/ImageResources.h"

#include "CeeUnstructGrid/DataSourceVTFx.h"
#include "CeeUnstructGrid/ModelSpec.h"
#include "CeeUnstructGrid/PropertyApplierVTFx.h"

#include "CeeExport/ExportVTFx.h"
#include "CeeExport/PropertyBuilderVTFx.h"

#include "CeeReport/ReportCreatorWord.h"
#include "CeeReport/ReportCreatorPowerPoint.h"
#include "CeeReport/ReportCreatorHtml.h"
#include "CeeReport/FieldValuesGenerator.h"

#include "ceeqtUtilsCore.h"

#include <QFrame>
#include <QHBoxLayout>
#include <QAction>
#include <QMenu>
#include <QMenuBar>
#include <QFileDialog>
#include <QFileInfo>
#include <QMessageBox>
#include <QDesktopServices>
#include <QApplication>
#include <QUrl>

//==================================================================================================
//
// 
//
//==================================================================================================

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
QRMainWindow::QRMainWindow(cee::vis::OpenGLContextGroup* contextGroup)
{
    QFrame* mainFrame = new QFrame;
    QHBoxLayout* frameLayout = new QHBoxLayout(mainFrame);
    setCentralWidget(mainFrame);

    m_viewer = new cee::qt::ViewerOpenGLWidgetDemo(contextGroup, mainFrame);

    m_viewer->setFocus();
    frameLayout->addWidget(m_viewer);

    QMenu* menu = menuBar()->addMenu("&File");
    menu->addAction("Open VTFx Model...", this, SLOT(slotOpen()));
    menu->addSeparator();
    QMenu* wordMenu = menu->addMenu("Create Word Report");
    {
        wordMenu->addAction("Single Image", this, SLOT(slotCreateImageWordReport()));
        wordMenu->addAction("Single VTFx", this, SLOT(slotCreateVTFxWordReport()));
        wordMenu->addAction("Single Table", this, SLOT(slotCreateTableWordReport()));
        wordMenu->addAction("Single Chart", this, SLOT(slotCreateChartWordReport()));
        wordMenu->addAction("Generic Image", this, SLOT(slotCreateGenericImageWordReport()));
        wordMenu->addAction("Sample", this, SLOT(slotCreateSampleWordReport()));
    }
    QMenu* pptMenu = menu->addMenu("Create PowerPoint Report");
    {
        pptMenu->addAction("Single Image", this, SLOT(slotCreateImagePowerPointReport()));
        pptMenu->addAction("Single VTFx", this, SLOT(slotCreateVTFxPowerPointReport()));
        pptMenu->addAction("Single Table", this, SLOT(slotCreateTablePowerPointReport()));
        pptMenu->addAction("Image + VTFx + Table", this, SLOT(slotCreateSamplePowerPointReport()));
        pptMenu->addAction("Generic image", this, SLOT(slotCreateGenericImagePowerPointReport()));
    }
    QMenu* htmlMenu = menu->addMenu("Create HTML Report");
    {
        htmlMenu->addAction("Single Image", this, SLOT(slotCreateImageHtmlReport()));
        htmlMenu->addAction("Single Table", this, SLOT(slotCreateTableHtmlReport()));
    }
    menu->addSeparator();
    menu->addAction("Create all reports", this, SLOT(slotCreateAllReports()));

    cee::vis::View* gcView = new cee::vis::View;
    m_viewer->setView(gcView);
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
QRMainWindow::~QRMainWindow()
{
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QRMainWindow::slotOpen()
{
    if (!m_viewer) return;

    QString allFiles = "";

    QString filter = "VTFx files (*.vtfx)";
    QString fileName = QFileDialog::getOpenFileName(this, tr("Open Model File"), "", filter);
    if (!fileName.isEmpty())
    {
        if (QFileInfo(fileName).suffix().toLower() == "vtfx")
        {
            openFile(fileName);
        }
    }
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QRMainWindow::slotCreateImageWordReport()
{
    QApplication::setOverrideCursor(Qt::WaitCursor);

    cee::PtrRef<cee::rep::Snapshot> snapshot = createImageSnapshot();
    if (snapshot)
    {
        cee::PtrRef<cee::rep::Repository> repository = new cee::rep::Repository();
        repository->addSnapshot(snapshot.get());

        cee::rep::ReportCreatorWord generator(repository.get(), "../../ReportTemplates/SingleImage.docx");

        cee::Str reportName = "SingleImage.docx";
        if (generator.generate(reportName))
        {
            showReport(reportName);
        }
    }

    QApplication::restoreOverrideCursor();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QRMainWindow::slotCreateVTFxWordReport()
{
    QApplication::setOverrideCursor(Qt::WaitCursor);

    cee::PtrRef<cee::rep::Snapshot> snapshot = createVTFxSnapshot();
    if (snapshot)
    {
        cee::PtrRef<cee::rep::Repository> repository = new cee::rep::Repository();
        repository->addSnapshot(snapshot.get());

        cee::rep::ReportCreatorWord generator(repository.get(), "../../ReportTemplates/SingleVTFx.docx");

        cee::Str reportName = "SingleVTFx.docx";
        if (generator.generate(reportName))
        {
            showReport(reportName);
        }
    }

    QApplication::restoreOverrideCursor();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QRMainWindow::slotCreateTableWordReport()
{
    if (!m_viewer) return;

    QApplication::setOverrideCursor(Qt::WaitCursor);

    cee::PtrRef<cee::rep::Snapshot> snapshot = createScalarTableSnapshot();
    if (snapshot)
    {
        cee::PtrRef<cee::rep::Repository> repository = new cee::rep::Repository();
        repository->addSnapshot(snapshot.get());

        cee::rep::ReportCreatorWord generator(repository.get(), "../../ReportTemplates/SingleTable.docx");

        cee::Str reportName = "SingleTable.docx";
        if (generator.generate(reportName))
        {
            showReport(reportName);
        }
    }

    QApplication::restoreOverrideCursor();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QRMainWindow::slotCreateChartWordReport()
{
    if (!m_viewer) return;

    QApplication::setOverrideCursor(Qt::WaitCursor);

    cee::PtrRef<cee::rep::Snapshot> snapshot = createChartSnapshot();
    if (snapshot)
    {
        cee::PtrRef<cee::rep::Repository> repository = new cee::rep::Repository();
        repository->addSnapshot(snapshot.get());

        cee::rep::ReportCreatorWord generator(repository.get(), "../../ReportTemplates/SingleChart.docx");

        cee::Str reportName = "SingleChart.docx";
        if (generator.generate(reportName))
        {
            showReport(reportName);
        }
    }

    QApplication::restoreOverrideCursor();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QRMainWindow::slotCreateGenericImageWordReport()
{
    QApplication::setOverrideCursor(Qt::WaitCursor);

    // Load repository containing three image snapshots
    cee::PtrRef<cee::rep::Repository> repository = new cee::rep::Repository();
    if (repository->loadRepository("../../ReportTemplates/ExampleRepository.repx"))
    {
        cee::rep::ReportCreatorWord generator(repository.get(), "../../ReportTemplates/GenericImage.docx");
        cee::Str reportName = "GenericImage.docx";
        if (generator.generate(reportName))
        {
            showReport(reportName);
        }
    }

    QApplication::restoreOverrideCursor();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QRMainWindow::slotCreateSampleWordReport()
{
    QApplication::setOverrideCursor(Qt::WaitCursor);

    // Load repository containing three image snapshots
    cee::PtrRef<cee::rep::Repository> repository = new cee::rep::Repository();
    if (repository->loadRepository("../../ReportTemplates/Situations.repx"))
    {
        cee::rep::ReportCreatorWord generator(repository.get(), "../../ReportTemplates/Situations.docx");
        cee::Str reportName = "Situations.docx";
        if (generator.generate(reportName))
        {
            showReport(reportName);
        }
    }

    QApplication::restoreOverrideCursor();
}

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QRMainWindow::slotCreateImagePowerPointReport()
{
    QApplication::setOverrideCursor(Qt::WaitCursor);

    cee::PtrRef<cee::rep::Snapshot> snapshot = createImageSnapshot();
    if (snapshot)
    {
        cee::PtrRef<cee::rep::Repository> repository = new cee::rep::Repository();
        repository->addSnapshot(snapshot.get());

        cee::rep::ReportCreatorPowerPoint generator(repository.get(), "../../ReportTemplates/SingleImage.pptx");

        cee::Str reportName = "SingleImage.pptx";
        if (generator.generate(reportName))
        {
            showReport(reportName);
        }
    }

    QApplication::restoreOverrideCursor();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QRMainWindow::slotCreateVTFxPowerPointReport()
{
    QApplication::setOverrideCursor(Qt::WaitCursor);

    cee::PtrRef<cee::rep::Snapshot> snapshot = createVTFxSnapshot();
    if (snapshot)
    {
        cee::PtrRef<cee::rep::Repository> repository = new cee::rep::Repository();
        repository->addSnapshot(snapshot.get());

        cee::rep::ReportCreatorPowerPoint generator(repository.get(), "../../ReportTemplates/SingleVTFx.pptx");

        cee::Str reportName = "SingleVTFx.pptx";
        if (generator.generate(reportName))
        {
            showReport(reportName);
        }
    }

    QApplication::restoreOverrideCursor();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QRMainWindow::slotCreateTablePowerPointReport()
{
    QApplication::setOverrideCursor(Qt::WaitCursor);

    cee::PtrRef<cee::rep::Snapshot> snapshot = createScalarTableSnapshot();
    if (snapshot)
    {
        cee::PtrRef<cee::rep::Repository> repository = new cee::rep::Repository();
        repository->addSnapshot(snapshot.get());

        cee::rep::ReportCreatorPowerPoint generator(repository.get(), "../../ReportTemplates/SingleTable.pptx");

        cee::Str reportName = "SingleTable.pptx";
        if (generator.generate(reportName))
        {
            showReport(reportName);
        }
    }

    QApplication::restoreOverrideCursor();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QRMainWindow::slotCreateSamplePowerPointReport()
{
    QApplication::setOverrideCursor(Qt::WaitCursor);

    cee::PtrRef<cee::rep::Snapshot> snapshot1 = createImageSnapshot();
    cee::PtrRef<cee::rep::Snapshot> snapshot2 = createScalarTableSnapshot();
    cee::PtrRef<cee::rep::Snapshot> snapshot3 = createVTFxSnapshot();
    if (snapshot1 && snapshot2 && snapshot3)
    {
        cee::PtrRef<cee::rep::Repository> repository = new cee::rep::Repository();
        repository->addSnapshot(snapshot1.get());
        repository->addSnapshot(snapshot2.get());
        repository->addSnapshot(snapshot3.get());

        cee::rep::ReportCreatorPowerPoint generator(repository.get(), "../../ReportTemplates/SimpleSample.pptx");
        cee::Str reportName = "SimpleSample.pptx";
        if (generator.generate(reportName))
        {
            showReport(reportName);
        }
    }

    QApplication::restoreOverrideCursor();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QRMainWindow::slotCreateGenericImagePowerPointReport()
{
    QApplication::setOverrideCursor(Qt::WaitCursor);

    // Load repository containing three image snapshots
    cee::PtrRef<cee::rep::Repository> repository = new cee::rep::Repository();
    if (repository->loadRepository("../../ReportTemplates/ExampleRepository.repx"))
    {
        cee::rep::ReportCreatorPowerPoint generator(repository.get(), "../../ReportTemplates/GenericImage.pptx");
        cee::Str reportName = "GenericImage.pptx";
        if (!generator.generate(reportName))
        {
            // Error
            return;
        }
        showReport(reportName);
    }
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QRMainWindow::slotCreateImageHtmlReport()
{
    QApplication::setOverrideCursor(Qt::WaitCursor);

    cee::PtrRef<cee::rep::Snapshot> snapshot = createImageSnapshot();
    if (snapshot)
    {
        cee::PtrRef<cee::rep::Repository> repository = new cee::rep::Repository();
        repository->addSnapshot(snapshot.get());

        cee::rep::ReportCreatorHtml generator(repository.get(), "../../ReportTemplates/SingleImage.html");

        cee::Str reportName = "SingleImage.html";
        if (generator.generate(reportName))
        {
            showReport(reportName);
        }
    }

    QApplication::restoreOverrideCursor();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QRMainWindow::slotCreateTableHtmlReport()
{
    QApplication::setOverrideCursor(Qt::WaitCursor);

    cee::PtrRef<cee::rep::Snapshot> snapshot = createScalarTableSnapshot();
    if (snapshot)
    {
        cee::PtrRef<cee::rep::Repository> repository = new cee::rep::Repository();
        repository->addSnapshot(snapshot.get());

        cee::rep::ReportCreatorHtml generator(repository.get(), "../../ReportTemplates/SingleTable.html");

        cee::Str reportName = "SingleTable.html";
        if (generator.generate(reportName))
        {
            showReport(reportName);
        }
    }

    QApplication::restoreOverrideCursor();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QRMainWindow::slotCreateAllReports()
{
    slotCreateImageWordReport();
    slotCreateVTFxWordReport();
    slotCreateTableWordReport();
    slotCreateImagePowerPointReport();
    slotCreateVTFxPowerPointReport();
    slotCreateTablePowerPointReport();
    slotCreateSamplePowerPointReport();
    slotCreateSampleWordReport();
    slotCreateGenericImageWordReport();
    slotCreateGenericImagePowerPointReport();
    slotCreateImageHtmlReport();
    slotCreateTableHtmlReport();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
cee::PtrRef<cee::rep::Snapshot> QRMainWindow::createImageSnapshot()
{
    if (!m_viewer || !m_unstructModel)
    {
        return NULL;
    }

    cee::PtrRef<cee::Image> image = new cee::Image();
    if (!m_viewer->view()->renderToImage(image.get()))
    {
        // Error
        return NULL;
    }

    cee::PtrRef<cee::rep::Snapshot> snapshot = new cee::rep::Snapshot(cee::rep::Snapshot::OBJECT_IMAGE);
    snapshot->setTitle("My Image Title");
    snapshot->setDescription("This is a description inserted from the Ceetron Report component. Here you can write a descriptive text about your model or data without having to edit the Word/PowerPoint document manually.");
    snapshot->setFieldValues(cee::rep::FieldValuesGenerator::getFieldValuesFromModel(m_unstructModel.get()));
    snapshot->setImage(image.get());

    return snapshot;
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
cee::PtrRef<cee::rep::Snapshot> QRMainWindow::createVTFxSnapshot()
{
    if (!m_viewer || !m_unstructModel)
    {
        return NULL;
    }

    cee::exp::ExportVTFx exporter(cee::exp::ExportVTFx::DISPLAY_MODEL_ONLY, m_unstructModel.get());

    cee::PtrRef<cee::PropertySetCollection> propColl = new cee::PropertySetCollection;
    cee::PtrRef<cee::ImageResources> resources = new cee::ImageResources;
    cee::exp::PropertyBuilderVTFx propBuilder(propColl.get(), resources.get());
    propBuilder.addFromModel(*m_unstructModel.get());
    propBuilder.addFromView(*m_viewer->view());
    propBuilder.addDefaultFramesPerSecond(1.0);
    exporter.addProperties(*propColl);
    exporter.addResources(*resources);

    cee::PtrRef<cee::ug::VTFxMemoryFile> memoryFile = new cee::ug::VTFxMemoryFile();
    if (!exporter.saveCase(memoryFile.get()))
    {
        return NULL;
    }

    cee::PtrRef<cee::rep::Snapshot> snapshot = new cee::rep::Snapshot(cee::rep::Snapshot::OBJECT_VTFX);
    snapshot->setTitle("My VTFx Model Title");
    snapshot->setDescription("This is a description inserted from the Ceetron Report component. Here you can write a descriptive text about your model or data without having to edit the Word/PowerPoint document manually.");
    std::map<cee::Str, cee::Str> fieldValues = cee::rep::FieldValuesGenerator::getFieldValuesFromModel(m_unstructModel.get());
    snapshot->setFieldValues(cee::rep::FieldValuesGenerator::getFieldValuesFromModel(m_unstructModel.get()));
    snapshot->setModelVTFx(memoryFile.get());

    return snapshot;
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
cee::PtrRef<cee::rep::Snapshot> QRMainWindow::createScalarTableSnapshot()
{
    if (!m_unstructModel)
    {
        return NULL;
    }

    const cee::ug::DataSourceDirectory* directory = m_unstructModel->dataSource()->directory();

    cee::PtrRef<cee::rep::Table> table = new cee::rep::Table(5, 2);

    int currentResultId = m_unstructModel->modelSpec().fringesResultId();
    if (currentResultId != -1)
    {
        const cee::ug::ResultInfo scalarInfo = directory->findScalarResultInfo(currentResultId);

        // First row
        {
            table->setValue(0, 1, "Values");
        }
        // Second row
        {
            table->setValue(1, 0, "Scalar name: ");
            table->setValue(1, 1, scalarInfo.name());
        }
        // Third row
        {
            table->setValue(2, 0, "Mapping:");
            switch (scalarInfo.resultMapping())
            {
                case cee::ug::PER_ELEMENT:          table->setValue(2, 1, "Per element"); break;
                case cee::ug::PER_ELEMENT_NODE:     table->setValue(2, 1, "Per element node"); break;
                case cee::ug::PER_ELEMENT_SURFACE:  table->setValue(2, 1, "Per element surface"); break;
                case cee::ug::PER_NODE:             table->setValue(2, 1, "Per element node"); break;
                default:                            table->setValue(2, 1, "Unknown"); break;
            }
        }
        // Fourth row
        {
            double minimum, maximum;
            m_unstructModel->scalarRange(currentResultId, &minimum, &maximum);
            table->setValue(3, 0, "Minimum: ");
            table->setValue(3, 1, cee::Str::number(minimum));
            table->setValue(4, 0, "Maximum: ");
            table->setValue(4, 1, cee::Str::number(maximum));
        }
    }

    cee::PtrRef<cee::rep::Snapshot> snapshot = new cee::rep::Snapshot(cee::rep::Snapshot::OBJECT_TABLE);
    snapshot->setTitle("My Table Title");
    snapshot->setDescription("This is a description inserted from the Ceetron Report component. Here you can write a descriptive text about your model or data without having to edit the Word/PowerPoint document manually.");
    snapshot->setFieldValues(cee::rep::FieldValuesGenerator::getFieldValuesFromModel(m_unstructModel.get()));
    snapshot->setTable(table.get());

    return snapshot;
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
cee::PtrRef<cee::rep::Snapshot> QRMainWindow::createChartSnapshot()
{
    cee::PtrRef<cee::rep::Table> table = new cee::rep::Table(7, 3);
    table->setChartXAxisTitle("Time steps");
    table->setChartYAxisTitle("Variations");

    // Column titles
    table->setValue(0, 0, "Steps"); 
    table->setValue(0, 1, "Heat");
    table->setValue(0, 2, "Height");

    table->setValue(1, 0, "1");
    table->setValue(1, 1, "15");
    table->setValue(1, 2, "0");
    table->setValue(2, 0, "2");
    table->setValue(2, 1, "35");
    table->setValue(2, 2, "2");
    table->setValue(3, 0, "3");
    table->setValue(3, 1, "66");
    table->setValue(3, 2, "4");
    table->setValue(4, 0, "4");
    table->setValue(4, 1, "42");
    table->setValue(4, 2, "5");
    table->setValue(5, 0, "5");
    table->setValue(5, 1, "25");
    table->setValue(5, 2, "7");
    table->setValue(6, 0, "6");
    table->setValue(6, 1, "20");
    table->setValue(6, 2, "10");

    cee::PtrRef<cee::rep::Snapshot> snapshot = new cee::rep::Snapshot(cee::rep::Snapshot::OBJECT_TABLE);
    snapshot->setTitle("My Chart Title");
    snapshot->setDescription("This is a description inserted from the Ceetron Report component. Here you can write a descriptive text about your model or data without having to edit the Word/PowerPoint document manually.");
    snapshot->setFieldValues(cee::rep::FieldValuesGenerator::getFieldValuesFromModel(m_unstructModel.get()));
    snapshot->setTable(table.get());

    return snapshot;
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QRMainWindow::showReport(const cee::Str& reportName) const
{
    QFileInfo fileInfo(cee::qt::UtilsCore::toQString(reportName));
    QString file = "file:///" + fileInfo.absoluteFilePath();

    QUrl url = QUrl(file, QUrl::TolerantMode);
    QDesktopServices::openUrl(url);
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QRMainWindow::openFile(const QString& fileName)
{
    int datasourceId = 1;
    cee::ug::DataSourceVTFx* source = new cee::ug::DataSourceVTFx(datasourceId);
    cee::ug::Error error;
    if (source->openCase(cee::qt::UtilsCore::toStr(fileName), 0, &error))
    {
        m_unstructModel = new cee::ug::UnstructGridModel();
        m_unstructModel->setDataSource(source);

        // Remove all old models from the view. We just want the latest.
        cee::vis::View* gcView = m_viewer->view();
        gcView->removeAllModels();
        gcView->addModel(m_unstructModel.get());

        // Read model properties and view properties from the VTFx data source
        cee::PtrRef<cee::PropertySetCollection> propertyCollection = new cee::PropertySetCollection;
        cee::PtrRef<cee::ImageResources> imgResources = new cee::ImageResources;
        source->caseProperties(propertyCollection.get(), imgResources.get());
        cee::ug::PropertyApplierVTFx propApplier(propertyCollection.get(), imgResources.get());
        propApplier.applyToModel(m_unstructModel.get());

        // Make sure that at least one state is specified. If we have a file without properties
        // we need to specify which state to show
        if (m_unstructModel->modelSpec().stateIds().size() == 0)
        {
            if (source->directory()->stateInfos().size() > 0)
            {
                int firstStateId = source->directory()->stateInfos()[0].id();
                m_unstructModel->modelSpec().setStateId(firstStateId);
            }
            else
            {
                QMessageBox::critical(this, tr("Error!"), tr("No states in file"));
            }
        }

        // Set up the initial visualization of this model
        m_unstructModel->updateVisualization();

        // Default camera setup
        cee::BoundingBox boundingBox = gcView->boundingBox();
        gcView->camera().fitView(boundingBox, cee::Vec3d(0, 1, 0), cee::Vec3d(0, 0, 1));
        gcView->camera().inputHandler()->setRotationPoint(boundingBox.center());

        // Apply properties to view
        // Apply properties after default view setup so we get a default view if camera is not specified
        propApplier.applyToView(gcView);

        gcView->requestRedraw();
    }
}




