//##################################################################################################
//
//  Ceetron Desktop Components
//  QtReportRepositoryManager
//  
//  --------------------------------------------------------------------------------------------
//  Copyright (C) 2016, Ceetron AS
//  This is UNPUBLISHED PROPRIETARY SOURCE CODE of Ceetron AS. The contents of this file may 
//  not be disclosed to third parties, copied or duplicated in any form, in whole or in part,
//  without the prior written permission of Ceetron AS.
//##################################################################################################

#include "QRMAddSituationDialog.h"

#include "ui_QRMAddSituationDialog.h"

#include "CeeVisualization/View.h"
#include "CeeVisualization/Camera.h"
#include "CeeVisualization/CameraInputHandler.h"

#include "CeeCore/PropertySetCollection.h"
#include "CeeCore/ImageResources.h"
#include "CeeCore/Image.h"

#include "CeeUnstructGrid/DataSourceVTFx.h"
#include "CeeUnstructGrid/ModelSpec.h"
#include "CeeUnstructGrid/PropertyApplierVTFx.h"

#include "CeeReport/Snapshot.h"
#include "CeeExport/ExportVTFx.h"
#include "CeeExport/PropertyBuilderVTFx.h"

#include "CeeReport/FieldValuesGenerator.h"

#include "ceeqtUtilsCore.h"
#include "ceeqtUtilsGui.h"
#include "ceeqtViewerOpenGLWidget.h"

#include <QFrame>
#include <QAction>
#include <QMessageBox>
#include <QFileInfo>
#include <QApplication>


//==================================================================================================
//
// 
//
//==================================================================================================

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
QRMAddSituationDialog::QRMAddSituationDialog(cee::vis::OpenGLContextGroup* contextGroup)
: m_ui(new Ui::QRMAddSituationDialog)
{
    m_ui->setupUi(this);

    QHBoxLayout* frameLayout = new QHBoxLayout(m_ui->viewerFrame);
    m_viewer = new cee::qt::ViewerOpenGLWidgetDemo(contextGroup, m_ui->viewerFrame, this);
    m_viewer->setFocus();
    frameLayout->addWidget(m_viewer);

    connect(m_ui->addImagePushButton, SIGNAL(pressed()), this, SLOT(slotAddAsImage()));
    connect(m_ui->addTablePushButton, SIGNAL(pressed()), this, SLOT(slotAddAsTable()));
    connect(m_ui->addVTFxPushButton, SIGNAL(pressed()), this, SLOT(slotAddAsVTFx()));
    connect(m_ui->addChartPushButton, SIGNAL(pressed()), this, SLOT(slotAddAsChart()));

    cee::vis::View* gcView = new cee::vis::View;
    m_viewer->setView(gcView);
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
QRMAddSituationDialog::~QRMAddSituationDialog()
{
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QRMAddSituationDialog::loadModel(const cee::Str& vtfxFileName)
{
    m_modelFileName = vtfxFileName;

    // The model will be loaded in onViewerOpenGLReady() to ensure openGL setup is finished
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QRMAddSituationDialog::onViewerOpenGLReady(cee::qt::ViewerOpenGLWidget* /*viewer*/)
{
    int datasourceId = 1;
    cee::ug::DataSourceVTFx* source = new cee::ug::DataSourceVTFx(datasourceId);
    cee::ug::Error error;
    if (source->openCase(m_modelFileName, 0, &error))
    {
        m_unstructModel = new cee::ug::UnstructGridModel();
        m_unstructModel->setDataSource(source);

        // Remove all old models from the view. We just want the latest.
        cee::vis::View* gcView = m_viewer->view();
        gcView->removeAllModels();
        gcView->addModel(m_unstructModel.get());

        // Read model properties and view properties from the VTFx data source
        cee::PtrRef<cee::PropertySetCollection> propertyCollection = new cee::PropertySetCollection;
        cee::PtrRef<cee::ImageResources> imgResources = new cee::ImageResources;
        source->caseProperties(propertyCollection.get(), imgResources.get());
        cee::ug::PropertyApplierVTFx propApplier(propertyCollection.get(), imgResources.get());
        propApplier.applyToModel(m_unstructModel.get());

        // Make sure that at least one state is specified. If we have a file without properties
        // we need to specify which state to show
        if (m_unstructModel->modelSpec().stateIds().size() == 0)
        {
            if (source->directory()->stateInfos().size() > 0)
            {
                int firstStateId = source->directory()->stateInfos()[0].id();
                m_unstructModel->modelSpec().setStateId(firstStateId);
            }
            else
            {
                QMessageBox::critical(this, tr("Error!"), tr("No states in file"));
                return;
            }
        }

        // Set up the initial visualization of this model
        m_unstructModel->updateVisualization();

        // Default camera setup
        cee::BoundingBox boundingBox = gcView->boundingBox();
        gcView->camera().fitView(boundingBox, cee::Vec3d(0, 1, 0), cee::Vec3d(0, 0, 1));
        gcView->camera().inputHandler()->setRotationPoint(boundingBox.center());

        // Apply properties to view
        // Apply properties after default view setup so we get a default view if camera is not specified
        propApplier.applyToView(gcView);

        gcView->requestRedraw();

        m_ui->titleLineEdit->setText(QFileInfo(cee::qt::UtilsCore::toQString(m_modelFileName)).baseName());
    }
}



//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QRMAddSituationDialog::slotAddAsImage()
{
    QApplication::setOverrideCursor(Qt::WaitCursor);

    cee::PtrRef<cee::Image> image = new cee::Image();
    if (m_viewer->view()->renderToImage(1600, 1200, image.get()))
    {
        cee::PtrRef<cee::rep::Snapshot> snapshot = new cee::rep::Snapshot(cee::rep::Snapshot::OBJECT_IMAGE);
        snapshot->setTitle(cee::qt::UtilsCore::toStr(m_ui->titleLineEdit->text()));
        snapshot->setDescription(cee::qt::UtilsCore::toStr(m_ui->descriptionTextEdit->toPlainText()));
        snapshot->setFieldValues(cee::rep::FieldValuesGenerator::getFieldValuesFromModel(m_unstructModel.get()));
        snapshot->setImage(image.get());

        emit acceptedSnapshot(snapshot.get());

        accept();
    }

    QApplication::restoreOverrideCursor();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QRMAddSituationDialog::slotAddAsVTFx()
{
    QApplication::setOverrideCursor(Qt::WaitCursor);

    cee::exp::ExportVTFx exporter(cee::exp::ExportVTFx::DISPLAY_MODEL_ONLY, m_unstructModel.get());

    cee::PtrRef<cee::Image> snapshot = new cee::Image();
    m_viewer->view()->renderToImage(1280, 960, snapshot.get());
    exporter.setSnapshotImage(snapshot.get());

    cee::PtrRef<cee::PropertySetCollection> propColl = new cee::PropertySetCollection;
    cee::PtrRef<cee::ImageResources> resources = new cee::ImageResources;
    cee::exp::PropertyBuilderVTFx propBuilder(propColl.get(), resources.get());
    propBuilder.addFromModel(*m_unstructModel.get());
    propBuilder.addFromView(*m_viewer->view());
    propBuilder.addDefaultFramesPerSecond(1.0);
    exporter.addProperties(*propColl);
    exporter.addResources(*resources);

    cee::PtrRef<cee::ug::VTFxMemoryFile> memoryFile = new cee::ug::VTFxMemoryFile();
    if (exporter.saveCase(memoryFile.get()))
    {
        cee::PtrRef<cee::rep::Snapshot> snapshot = new cee::rep::Snapshot(cee::rep::Snapshot::OBJECT_VTFX);
        snapshot->setTitle(cee::qt::UtilsCore::toStr(m_ui->titleLineEdit->text()));
        snapshot->setDescription(cee::qt::UtilsCore::toStr(m_ui->descriptionTextEdit->toPlainText()));
        snapshot->setFieldValues(cee::rep::FieldValuesGenerator::getFieldValuesFromModel(m_unstructModel.get()));
        snapshot->setModelVTFx(memoryFile.get());

        emit acceptedSnapshot(snapshot.get());

        accept();
    }

    QApplication::restoreOverrideCursor();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QRMAddSituationDialog::slotAddAsTable()
{
    QApplication::setOverrideCursor(Qt::WaitCursor);

    const cee::ug::DataSourceDirectory* directory = m_unstructModel->dataSource()->directory();

    cee::PtrRef<cee::rep::Table> table = new cee::rep::Table(5, 2);

    int currentResultId = m_unstructModel->modelSpec().fringesResultId();
    if (currentResultId != -1)
    {
        const cee::ug::ResultInfo scalarInfo = directory->findScalarResultInfo(currentResultId);

        // First row
        {
            table->setValue(0, 1, "Values");
        }
        // Second row
        {
            table->setValue(1, 0, "Scalar name: ");
            table->setValue(1, 1, scalarInfo.name());
        }
        // Third row
        {
            table->setValue(2, 0, "Mapping:");
            switch (scalarInfo.resultMapping())
            {
                case cee::ug::PER_ELEMENT:          table->setValue(2, 1, "Per element"); break;
                case cee::ug::PER_ELEMENT_NODE:     table->setValue(2, 1, "Per element node"); break;
                case cee::ug::PER_ELEMENT_SURFACE:  table->setValue(2, 1, "Per element surface"); break;
                case cee::ug::PER_NODE:             table->setValue(2, 1, "Per element node"); break;
                default:                            table->setValue(2, 1, "Unknown"); break;
            }
        }
        // Fourth row
        {
            double minimum, maximum;
            m_unstructModel->scalarRange(currentResultId, &minimum, &maximum);
            table->setValue(3, 0, "Minimum: ");
            table->setValue(3, 1, cee::Str::number(minimum));
            table->setValue(4, 0, "Maximum: ");
            table->setValue(4, 1, cee::Str::number(maximum));
        }
    }

    cee::PtrRef<cee::rep::Snapshot> snapshot = new cee::rep::Snapshot(cee::rep::Snapshot::OBJECT_TABLE);
    snapshot->setTitle(cee::qt::UtilsCore::toStr(m_ui->titleLineEdit->text()));
    snapshot->setDescription(cee::qt::UtilsCore::toStr(m_ui->descriptionTextEdit->toPlainText()));
    snapshot->setFieldValues(cee::rep::FieldValuesGenerator::getFieldValuesFromModel(m_unstructModel.get()));
    snapshot->setTable(table.get());

    emit acceptedSnapshot(snapshot.get());
    accept();

    QApplication::restoreOverrideCursor();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QRMAddSituationDialog::slotAddAsChart()
{
    QApplication::setOverrideCursor(Qt::WaitCursor);

    cee::PtrRef<cee::rep::Table> table = new cee::rep::Table(7, 3);
    table->setChartXAxisTitle("Time steps");
    table->setChartYAxisTitle("Variations");

    // Column titles
    table->setValue(0, 0, "Steps");
    table->setValue(0, 1, "Heat");
    table->setValue(0, 2, "Height");

    table->setValue(1, 0, "1");
    table->setValue(1, 1, "15");
    table->setValue(1, 2, "0");
    table->setValue(2, 0, "2");
    table->setValue(2, 1, "35");
    table->setValue(2, 2, "2");
    table->setValue(3, 0, "3");
    table->setValue(3, 1, "66");
    table->setValue(3, 2, "4");
    table->setValue(4, 0, "4");
    table->setValue(4, 1, "42");
    table->setValue(4, 2, "5");
    table->setValue(5, 0, "5");
    table->setValue(5, 1, "25");
    table->setValue(5, 2, "7");
    table->setValue(6, 0, "6");
    table->setValue(6, 1, "20");
    table->setValue(6, 2, "10");

    cee::PtrRef<cee::rep::Snapshot> snapshot = new cee::rep::Snapshot(cee::rep::Snapshot::OBJECT_TABLE);
    snapshot->setTitle(cee::qt::UtilsCore::toStr(m_ui->titleLineEdit->text()));
    snapshot->setDescription(cee::qt::UtilsCore::toStr(m_ui->descriptionTextEdit->toPlainText()));
    snapshot->setFieldValues(cee::rep::FieldValuesGenerator::getFieldValuesFromModel(m_unstructModel.get()));
    snapshot->setTable(table.get());

    emit acceptedSnapshot(snapshot.get());
    accept();

    QApplication::restoreOverrideCursor();
}


