//##################################################################################################
//
//  Ceetron Desktop Components
//  QtReportRepositoryViewer
//  
//  --------------------------------------------------------------------------------------------
//  Copyright (C) 2015, Ceetron AS
//  This is UNPUBLISHED PROPRIETARY SOURCE CODE of Ceetron AS. The contents of this file may 
//  not be disclosed to third parties, copied or duplicated in any form, in whole or in part,
//  without the prior written permission of Ceetron AS.
//##################################################################################################

#include "QRMMainWindow.h"

#include "ui_QRMMainWindow.h"

#include "QRMAddSituationDialog.h"
#include "QRMSummaryDialog.h"
#include "QRMSendToCloudDialogs.h"

#include "CeeVisualization/View.h"
#include "CeeVisualization/Camera.h"
#include "CeeVisualization/CameraInputHandler.h"

#include "CeeCore/PropertySetCollection.h"
#include "CeeCore/ImageResources.h"

#include "CeeUnstructGrid/DataSourceVTFx.h"
#include "CeeUnstructGrid/ModelSpec.h"
#include "CeeUnstructGrid/PropertyApplierVTFx.h"
#include "CeeUnstructGrid/VTFxFileBrowser.h"

#include "CeeReport/Repository.h"
#include "CeeReport/Snapshot.h"
#include "CeeReport/ReportCreatorWord.h"
#include "CeeReport/ReportCreatorHtml.h"
#include "CeeReport/ReportCreatorPowerPoint.h"

#include "CeeExport/PropertyBuilderVTFx.h"
#include "CeeExport/ExportVTFx.h"

#include "ceeqtUtilsCore.h"
#include "ceeqtUtilsGui.h"

#include <QFrame>
#include <QHBoxLayout>
#include <QAction>
#include <QMenu>
#include <QMenuBar>
#include <QFileDialog>
#include <QFileInfo>
#include <QMessageBox>
#include <QPixmap>
#include <QDesktopServices>
#include <QSslSocket>
#include <QSettings>
#include <QProgressDialog>
#include <QTemporaryFile>
#include <QHttpMultiPart>
#include <QNetworkAccessManager>
#include <QNetworkReply>


//==================================================================================================
//
// 
//
//==================================================================================================

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
QRMMainWindow::QRMMainWindow(cee::vis::OpenGLContextGroup* contextGroup)
:   m_ui(new Ui::QRMMainWindow),
    m_progressDialog(NULL)
{
    m_ui->setupUi(this);

    m_contextGroup = contextGroup;

    QHBoxLayout* frameLayout = new QHBoxLayout(m_ui->viewFrame);
    m_viewer = new cee::qt::ViewerOpenGLWidgetDemo(contextGroup, m_ui->viewFrame);
    m_viewer->setFocus();
    frameLayout->addWidget(m_viewer);

    connect(m_ui->actionOpen_Repository, SIGNAL(triggered()), this, SLOT(slotOpen()));
    connect(m_ui->actionExit, SIGNAL(triggered()), this, SLOT(close()));
    connect(m_ui->actionOpen_Model, SIGNAL(triggered()), this, SLOT(slotOpenModel()));
    connect(m_ui->actionSave_Repository, SIGNAL(triggered()), this, SLOT(slotSave()));
    connect(m_ui->actionClear_Upload_Id, SIGNAL(triggered()), this, SLOT(clearCloudUploadId()));

    connect(m_ui->clearSituationListButton, SIGNAL(pressed()), this, SLOT(slotClearSnapshotList()));
    connect(m_ui->situationListWidget, SIGNAL(currentItemChanged(QListWidgetItem*, QListWidgetItem*)), this, SLOT(slotUpdateFromCurrent()));
    connect(m_ui->cloudPushButton, SIGNAL(pressed()), this, SLOT(slotSendToCloud()));

    connect(m_ui->actionCreate_Word_Report, SIGNAL(triggered()), this, SLOT(slotCreateWordReport()));
    connect(m_ui->actionCreate_PowerPoint_Report, SIGNAL(triggered()), this, SLOT(slotCreatePowerPointReport()));
    connect(m_ui->actionCreate_HTML_Report, SIGNAL(triggered()), this, SLOT(slotCreateHtmlReport()));
    connect(m_ui->actionView_Report_Summary, SIGNAL(triggered()), this, SLOT(slotViewReportSummary()));
    connect(m_ui->actionValidate_Report, SIGNAL(triggered()), this, SLOT(slotViewReportValidation()));

    cee::vis::View* gcView = new cee::vis::View;
    m_viewer->setView(gcView);

    m_repository = new cee::rep::Repository();
    m_ui->repositoryLabel->setText("");
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
QRMMainWindow::~QRMMainWindow()
{
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QRMMainWindow::slotOpen()
{
    QString filter = "Repository files (*.repx)";
    QString fileName = QFileDialog::getOpenFileName(this, tr("Open Repository File"), "", filter);
    if (!fileName.isEmpty())
    {
        if (QFileInfo(fileName).suffix().toLower() == "repx")
        {
            loadRepository(cee::qt::UtilsCore::toStr(fileName));
        }
    }
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QRMMainWindow::slotSave()
{
    QString filter = "Repository files (*.repx)";
    QString fileName = QFileDialog::getSaveFileName(this, tr("Save Repository File"), "", filter);
    if (!fileName.isEmpty())
    {
        if (QFileInfo(fileName).suffix().toLower() == "repx")
        {
            m_repository->saveRepository(cee::qt::UtilsCore::toStr(fileName));
        }
    }

    m_ui->repositoryLabel->setText(fileName);
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QRMMainWindow::slotUpdateFromCurrent()
{
    QRMSituationItem* item = static_cast<QRMSituationItem*>(m_ui->situationListWidget->currentItem());
    size_t snapshotIndex = item->snapshotIndex();

    const cee::rep::Snapshot* snapshot = m_repository->snapshot(snapshotIndex);
    if (!snapshot)
    {
        // Error
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);

    // Update viewer from snapshot
    QString snapshotType;
    switch (snapshot->type())
    {
        case cee::rep::Snapshot::OBJECT_IMAGE: snapshotType = "Image";        break;
        case cee::rep::Snapshot::OBJECT_VTFX:  snapshotType = "VTFx Model";   break;
        case cee::rep::Snapshot::OBJECT_TABLE: snapshotType = "Tabular data"; break;
    }

    // Description
    m_ui->situationTextEdit->clear();
    m_ui->situationTextEdit->setFontUnderline(true);
    m_ui->situationTextEdit->setFontWeight(87);
    m_ui->situationTextEdit->append(QString("Snapshot %1").arg(snapshotIndex + 1));
    m_ui->situationTextEdit->setFontUnderline(false);
    m_ui->situationTextEdit->setFontWeight(50);
    m_ui->situationTextEdit->append("Title: " + cee::qt::UtilsCore::toQString(snapshot->title()));
    m_ui->situationTextEdit->append("Type: " + snapshotType + "\n");
    if (snapshot->type() == cee::rep::Snapshot::OBJECT_TABLE)
    {
        m_ui->situationTextEdit->append("X axis title: " + cee::qt::UtilsCore::toQString(snapshot->table()->chartXAxisTitle()));
        m_ui->situationTextEdit->append("Y axis title: " + cee::qt::UtilsCore::toQString(snapshot->table()->chartYAxisTitle()) + "\n");
    }
    m_ui->situationTextEdit->setFontWeight(87);
    m_ui->situationTextEdit->setFontUnderline(true);
    m_ui->situationTextEdit->append("Description");
    m_ui->situationTextEdit->setFontUnderline(false);
    m_ui->situationTextEdit->setFontWeight(50);
    m_ui->situationTextEdit->append(cee::qt::UtilsCore::toQString(snapshot->description()));

    showSnapshot(snapshotIndex);

    // Only enable "send to cloud" for VTFx snapshots
    m_ui->cloudPushButton->setEnabled(snapshot->type() == cee::rep::Snapshot::OBJECT_VTFX);

    QApplication::restoreOverrideCursor();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QRMMainWindow::loadRepository(const cee::Str& filename)
{
    m_repository = new cee::rep::Repository();
    if (m_repository->loadRepository(filename))
    {
        updateRepositoryList();
    }

    m_ui->repositoryLabel->setText(cee::qt::UtilsCore::toQString(filename));
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QRMMainWindow::updateRepositoryList()
{
    QRMListWidget* listWidget = m_ui->situationListWidget;

    listWidget->blockSignals(true);
    listWidget->clear();

    for (size_t i = 0; i < m_repository->snapshotCount(); i++)
    {
        const cee::rep::Snapshot* snapshot = m_repository->snapshot(i);
        QImage snapshotImage;
        switch (snapshot->type())
        {
            case cee::rep::Snapshot::OBJECT_IMAGE:
            {
                const cee::Image* image = snapshot->image();
                if (image)
                {
                    snapshotImage = cee::qt::UtilsGui::toQImage(*image);
                }
                else
                {
                    snapshotImage = QImage(":/reporter_logo_72x72.png");
                }
                break;
            }
            case cee::rep::Snapshot::OBJECT_VTFX:
            {
                cee::ug::VTFxFileBrowser browser;
                browser.initialize(snapshot->modelVTFx());
                cee::Image* image = browser.caseSnapshot(0);
                if (image)
                {
                    snapshotImage = cee::qt::UtilsGui::toQImage(*browser.caseSnapshot(0));
                }
                else
                {
                    snapshotImage = QImage(":/reporter_logo_72x72.png");
                }
				break;
            }
            case cee::rep::Snapshot::OBJECT_TABLE:
            {
                snapshotImage = QImage(":/table_64x48.png");
                break;
            }
        }
        listWidget->addListSnapshot(i, snapshotImage, QString("%1. ").arg(i + 1) + cee::qt::UtilsCore::toQString(snapshot->title()));
    }
    listWidget->blockSignals(false);
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QRMMainWindow::showSnapshot(size_t snapshotIndex)
{
    const cee::rep::Snapshot* snapshot = m_repository->snapshot(snapshotIndex);
    if (!snapshot)
    {
        // Error
        return;
    }

    switch (snapshot->type())
    {
        case cee::rep::Snapshot::OBJECT_IMAGE: showImageSnapshot(snapshotIndex); break;
        case cee::rep::Snapshot::OBJECT_VTFX:  showVTFxSnapshot(snapshotIndex);  break;
        case cee::rep::Snapshot::OBJECT_TABLE: showTableSnapshot(snapshotIndex); break;
    }
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QRMMainWindow::showVTFxSnapshot(size_t snapshotIndex)
{
    QApplication::setOverrideCursor(Qt::WaitCursor);

    cee::vis::View* gcView = m_viewer->view();
    gcView->removeAllModels();

    const cee::rep::Snapshot* snapshot = m_repository->snapshot(snapshotIndex);
    Q_ASSERT(snapshot);
    Q_ASSERT(snapshot->type() == cee::rep::Snapshot::OBJECT_VTFX);

    const cee::ug::VTFxMemoryFile* memoryFile = snapshot->modelVTFx();
    Q_ASSERT(memoryFile);

    int datasourceId = 1;
    cee::ug::DataSourceVTFx* source = new cee::ug::DataSourceVTFx(datasourceId);
    cee::ug::Error error;
    if (source->openCase(memoryFile, 0, &error))
    {
        if (m_unstructModel.get() == NULL)
        {
            m_unstructModel = new cee::ug::UnstructGridModel();
        }
        m_unstructModel->setDataSource(source);

        // Remove all old models from the view. We just want the latest.
        gcView->removeAllModels();
        gcView->addModel(m_unstructModel.get());

        // Read model properties and view properties from the VTFx data source
        cee::PtrRef<cee::PropertySetCollection> propertyCollection = new cee::PropertySetCollection;
        cee::PtrRef<cee::ImageResources> imgResources = new cee::ImageResources;
        source->caseProperties(propertyCollection.get(), imgResources.get());
        cee::ug::PropertyApplierVTFx propApplier(propertyCollection.get(), imgResources.get());
        propApplier.applyToModel(m_unstructModel.get());

        // Make sure that at least one state is specified. If we have a file without properties
        // we need to specify which state to show
        if (m_unstructModel->modelSpec().stateIds().size() == 0)
        {
            if (source->directory()->stateInfos().size() > 0)
            {
                int firstStateId = source->directory()->stateInfos()[0].id();
                m_unstructModel->modelSpec().setStateId(firstStateId);
            }
            else
            {
                QMessageBox::critical(this, tr("Error!"), tr("No states in file"));
            }
        }

        // Set up the initial visualization of this model
        m_unstructModel->updateVisualization();

        // Default camera setup
        cee::BoundingBox boundingBox = gcView->boundingBox();
        gcView->camera().fitView(boundingBox, cee::Vec3d(0, 1, 0), cee::Vec3d(0, 0, 1));
        gcView->camera().inputHandler()->setRotationPoint(boundingBox.center());

        // Apply properties to view
        // Apply properties after default view setup so we get a default view if camera is not specified
        propApplier.applyToView(gcView);

        gcView->requestRedraw();
    }
    else
    {
        QMessageBox::critical(this, tr("Error!"), tr("Failed reading VTFx from repository"));
    }

    m_ui->stackedWidget->setCurrentWidget(m_ui->vtfxPage);

    QApplication::restoreOverrideCursor();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QRMMainWindow::showImageSnapshot(size_t snapshotIndex)
{
    const cee::rep::Snapshot* snapshot = m_repository->snapshot(snapshotIndex);
    Q_ASSERT(snapshot);
    Q_ASSERT(snapshot->type() == cee::rep::Snapshot::OBJECT_IMAGE);

    if (snapshot->image())
    {
        QImage situationImage = cee::qt::UtilsGui::toQImage(*snapshot->image());
        m_ui->imageLabel->setPixmap(QPixmap::fromImage(situationImage));

        resizeImage();
    }

    m_ui->stackedWidget->setCurrentWidget(m_ui->imagePage);
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QRMMainWindow::resizeImage()
{
#if QT_VERSION >= 0x060000
	if (!m_ui->imageLabel->pixmap().isNull())
	{
		QSize newSize = m_ui->imageLabel->pixmap().size();
		newSize.scale(m_ui->imagePage->size(), Qt::KeepAspectRatio);

		m_ui->imageLabel->resize(newSize);
	}
#else
	if (m_ui->imageLabel->pixmap() != 0)
    {
        QSize newSize = m_ui->imageLabel->pixmap()->size();
        newSize.scale(m_ui->imagePage->size(), Qt::KeepAspectRatio);

        m_ui->imageLabel->resize(newSize);
    }
#endif
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
bool QRMMainWindow::showTableSnapshot(size_t snapshotIndex)
{
    const cee::rep::Snapshot* snapshot = m_repository->snapshot(snapshotIndex);
    Q_ASSERT(snapshot);
    Q_ASSERT(snapshot->type() == cee::rep::Snapshot::OBJECT_TABLE);

    m_ui->tableWidget->clear();

    const cee::rep::Table* table = snapshot->table();
    
    m_ui->tableWidget->setRowCount(static_cast<int>(table->rowCount()));
    m_ui->tableWidget->setColumnCount(static_cast<int>(table->columnCount()));

    for (size_t r = 0; r < table->rowCount(); r++)
    {
        for (size_t c = 0; c < table->columnCount(); c++)
        {
            m_ui->tableWidget->setItem(static_cast<int>(r), static_cast<int>(c), new QTableWidgetItem(cee::qt::UtilsCore::toQString(table->value(r, c))));
        }
    }

    m_ui->stackedWidget->setCurrentWidget(m_ui->tablePage);

    return true;
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QRMMainWindow::resizeEvent(QResizeEvent* event)
{
    QMainWindow::resizeEvent(event);

    resizeImage();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QRMMainWindow::slotOpenModel()
{
    QString filter = "VTFx files (*.vtfx)";
    QString fileName = QFileDialog::getOpenFileName(this, tr("Open Model File"), "", filter);
    if (!fileName.isEmpty())
    {
        if (QFileInfo(fileName).suffix().toLower() == "vtfx")
        {
            QRMAddSituationDialog dialog(m_contextGroup);
            dialog.loadModel(cee::qt::UtilsCore::toStr(fileName));
            connect(&dialog, SIGNAL(acceptedSnapshot(cee::rep::Snapshot*)), this, SLOT(slotAddSnapshot(cee::rep::Snapshot*)));

            dialog.exec();
        }
    }
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QRMMainWindow::slotAddSnapshot(cee::rep::Snapshot* snapshot)
{
    m_repository->addSnapshot(snapshot);
    updateRepositoryList();

    m_ui->situationListWidget->setCurrentRow(static_cast<int>(m_repository->snapshotCount()) - 1);

    connect(m_ui->situationListWidget, SIGNAL(currentItemChanged(QListWidgetItem*, QListWidgetItem*)), this, SLOT(slotUpdateFromCurrent()));
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QRMMainWindow::slotCreateWordReport()
{
    QString filter = "Word documents (*.docx)";
    QString templateName = QFileDialog::getOpenFileName(this, tr("Open Word Template"), "", filter);
    if (!templateName.isEmpty())
    {
        if (QFileInfo(templateName).suffix().toLower() == "docx")
        {
            QApplication::setOverrideCursor(Qt::WaitCursor);
            cee::rep::ReportCreatorWord generator(m_repository.get(), cee::qt::UtilsCore::toStr(templateName));
            QString reportName = QFileDialog::getSaveFileName(this, tr("Save Word Report"), QFileInfo(templateName).baseName() + "_report", filter);
            if (generator.generate(cee::qt::UtilsCore::toStr(reportName)))
            {
                showReport(cee::qt::UtilsCore::toStr(reportName));
            }
            QApplication::restoreOverrideCursor();
        }
    }
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QRMMainWindow::slotCreatePowerPointReport()
{
    QString filter = "PowerPoint presentation (*.pptx)";
    QString templateName = QFileDialog::getOpenFileName(this, tr("Open PowerPoint Template"), "", filter);
    if (!templateName.isEmpty())
    {
        if (QFileInfo(templateName).suffix().toLower() == "pptx")
        {
            QApplication::setOverrideCursor(Qt::WaitCursor);
            cee::rep::ReportCreatorPowerPoint generator(m_repository.get(), cee::qt::UtilsCore::toStr(templateName));
            QString reportName = QFileDialog::getSaveFileName(this, tr("Save PowerPoint Report"), QFileInfo(templateName).baseName() + "_report", filter);
            if (generator.generate(cee::qt::UtilsCore::toStr(reportName)))
            {
                showReport(cee::qt::UtilsCore::toStr(reportName));
            }
            QApplication::restoreOverrideCursor();
        }
    }
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QRMMainWindow::slotCreateHtmlReport()
{
    QString filter = "HTML document (*.html *.htm)";
    QString templateName = QFileDialog::getOpenFileName(this, tr("Open HTML Template"), "", filter);
    if (!templateName.isEmpty())
    {
        QString suffix = QFileInfo(templateName).suffix().toLower();
        if (suffix == "html" || suffix == "htm")
        {
            QApplication::setOverrideCursor(Qt::WaitCursor);
            cee::rep::ReportCreatorHtml generator(m_repository.get(), cee::qt::UtilsCore::toStr(templateName));
            QString reportName = QFileDialog::getSaveFileName(this, tr("Save HTML Report"), QFileInfo(templateName).baseName() + "_report", filter);
            if (generator.generate(cee::qt::UtilsCore::toStr(reportName)))
            {
                showReport(cee::qt::UtilsCore::toStr(reportName));
            }
            QApplication::restoreOverrideCursor();
        }
    }
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QRMMainWindow::showReport(cee::Str reportName) const
{
    QFileInfo fileInfo(cee::qt::UtilsCore::toQString(reportName));
    QString file = "file:///" + fileInfo.absoluteFilePath();

    QUrl url = QUrl(file, QUrl::TolerantMode);
    QDesktopServices::openUrl(url);
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QRMMainWindow::slotClearSnapshotList()
{
    m_repository->removeAllSnapshots();

    updateRepositoryList();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QRMMainWindow::slotViewReportSummary()
{
    QString filter = "Word document (*.docx);;PowerPoint template (*.pptx);;HTML document (*.html *.htm)";
    QString templateName = QFileDialog::getOpenFileName(this, tr("Open Template"), "", filter);
    if (!templateName.isEmpty())
    {
        cee::Str summary;
        QString suffix = QFileInfo(templateName).suffix().toLower();
        if (suffix == "docx")
        {
            cee::rep::ReportCreatorWord::summarize(cee::qt::UtilsCore::toStr(templateName), &summary);
            QRMSummaryDialog summaryDialog(cee::qt::UtilsCore::toStr(templateName), summary);
            summaryDialog.exec();
        }
        else if (suffix == "pptx")
        {
            cee::rep::ReportCreatorPowerPoint::summarize(cee::qt::UtilsCore::toStr(templateName), &summary);
            QRMSummaryDialog summaryDialog(cee::qt::UtilsCore::toStr(templateName), summary);
            summaryDialog.exec();
        }
        else if (suffix == "htm" || suffix == "html")
        {
            cee::rep::ReportCreatorHtml::summarize(cee::qt::UtilsCore::toStr(templateName), &summary);
            QRMSummaryDialog summaryDialog(cee::qt::UtilsCore::toStr(templateName), summary);
            summaryDialog.exec();
        }
    }
}



//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void QRMMainWindow::slotViewReportValidation()
{
    QString filter = "Word document (*.docx);;PowerPoint template (*.pptx);;HTML document (*.html *.htm)";
    QString templateName = QFileDialog::getOpenFileName(this, tr("Open Template"), "", filter);
    if (!templateName.isEmpty())
    {
        cee::Str summary;
        QString suffix = QFileInfo(templateName).suffix().toLower();
        if (suffix == "docx")
        {
            cee::rep::ReportCreatorWord creator(m_repository.get(), cee::qt::UtilsCore::toStr(templateName));
            creator.validate(&summary);
            QRMSummaryDialog summaryDialog(cee::qt::UtilsCore::toStr(templateName), summary);
            summaryDialog.exec();
        }
        else if (suffix == "pptx")
        {
            cee::rep::ReportCreatorPowerPoint creator(m_repository.get(), cee::qt::UtilsCore::toStr(templateName));
            creator.validate(&summary);
            QRMSummaryDialog summaryDialog(cee::qt::UtilsCore::toStr(templateName), summary);
            summaryDialog.exec();
        }
        else if (suffix == "htm" || suffix == "html")
        {
            cee::rep::ReportCreatorHtml creator(m_repository.get(), cee::qt::UtilsCore::toStr(templateName));
            creator.validate(&summary);
            QRMSummaryDialog summaryDialog(cee::qt::UtilsCore::toStr(templateName), summary);
            summaryDialog.exec();
        }
    }
}


//--------------------------------------------------------------------------------------------------
/// Send the model to Ceetron cloud
//--------------------------------------------------------------------------------------------------
void QRMMainWindow::sendToCloud(const cee::rep::Snapshot* snapshot)
{
    if (!snapshot)
    {
        return;
    }

#ifndef QT_NO_OPENSSL
    // Check that we have SSL/HTTPS support through Qt (normally provided via OpenSSL as a dynamic library)
    if (!QSslSocket::supportsSsl())
    {
        QMessageBox::critical(this, "Error Sending to Cloud", "Your system and/or Qt version does not support secure connections (HTTPS).\nAre the OpenSSL dynamic libraries missing?");
        return;
    }
#else
    // There is no compile time SSL support in this Qt version
    // Either download another Qt version or recompile Qt with OpenSSl support
    QMessageBox::critical(this, "Error Sending to Cloud", "Your Qt version was not built with OpenSSL support.");
    return;
#endif

    // Note!
    // In a real application the upload ID should be kept secret. 
    // Storing it to registry/settings file in the following fashion should be avoided.
    QSettings settings;
    QString uploadId = settings.value("CloudUploadId").toString();
    if (uploadId.isEmpty())
    {
        EnterUploadIDDialog resDialog(this);
        resDialog.resize(600, 270);

        if (resDialog.exec() == QDialog::Accepted && resDialog.uploadId().length() > 1)
        {
            uploadId = resDialog.uploadId();
            settings.setValue("CloudUploadId", uploadId);
        }
    }

    if (uploadId.isEmpty())
    {
        return;
    }

    m_progressDialog = new QProgressDialog("Preparing file for upload...", "", 0, 100);
    m_progressDialog->setCancelButton(NULL);
    m_progressDialog->setMinimumDuration(0);
    m_progressDialog->show();
    m_progressDialog->setValue(1);

    qApp->processEvents(QEventLoop::ExcludeUserInputEvents);

    QTemporaryFile tmpFile(QDir::tempPath() + "/uploadFile.XXXXXX");
    tmpFile.open();
    tmpFile.setAutoRemove(false);
    QString tempFilename = tmpFile.fileName();
    tempFilename += ".vtfx";

    snapshot->modelVTFx()->save(cee::qt::UtilsCore::toStr(tempFilename));

    // Send to cloud!
    QHttpMultiPart *multiPart = new QHttpMultiPart(QHttpMultiPart::FormDataType);
    multiPart->setBoundary("boundary_Ceetron89906a4aJn1kvlAG52SFG2AD");

    QHttpPart vtfxPart;
    vtfxPart.setHeader(QNetworkRequest::ContentTypeHeader, QVariant("application/octet-stream"));
    vtfxPart.setHeader(QNetworkRequest::ContentDispositionHeader, QVariant("form-data; name=\"vtfx\" filename=\"upload.vtfx\""));

    QFile *file = new QFile(tempFilename);
    if (!file->open(QIODevice::ReadOnly))
    {
        QMessageBox::critical(this, "Upload error", QString("Error opening temporary file %1").arg(tempFilename), QMessageBox::Ok);
        return;
    }


    vtfxPart.setBodyDevice(file);

    file->setParent(multiPart); // we cannot delete the file now, so delete it with the multiPart
    multiPart->append(vtfxPart);

    QString uploadApp = QUrl::toPercentEncoding("CDC ReportManager App Qt");

    QFileInfo fi(cee::qt::UtilsCore::toQString(snapshot->name()));
    QString caseName = QUrl::toPercentEncoding(fi.fileName());

    QString urlString = QString("https://cloud.ceetron.com/api/v1/models?uploadId=%1&uploadApp=%2&name=%3").arg(uploadId).arg(uploadApp).arg(caseName); 

    QUrl url(urlString);
    QNetworkRequest request(url);

    QNetworkAccessManager* manager = new QNetworkAccessManager(this);
    QNetworkReply* reply = manager->post(request, multiPart);
    multiPart->setParent(reply);

    connect(reply, SIGNAL(finished()), SLOT(cloudUploadFinished()));
    connect(reply, SIGNAL(uploadProgress(qint64, qint64)), SLOT(updateUploadProgress(qint64, qint64)));

    m_progressDialog->setLabelText("Uploading file to Ceetron Cloud...");
}


//--------------------------------------------------------------------------------------------------
/// Handle the end of the model upload to the cloud
//--------------------------------------------------------------------------------------------------
void QRMMainWindow::cloudUploadFinished()
{
    delete m_progressDialog;
    m_progressDialog = NULL;

    QString htmlDlgMsg;

    QNetworkReply* reply = (QNetworkReply*)sender();
    const int httpStatusCode = reply->attribute(QNetworkRequest::HttpStatusCodeAttribute).toInt();
    if (httpStatusCode > 0)
    {
        QString responseJson = reply->readAll();

        if (httpStatusCode == 200)
        {
            const QString ceetronCloudUrl = "https://cloud.ceetron.com/";
            const QString modelKey = parseTopLevelJsonValue(responseJson, "modelKey");
            const QString viewerUrl = ceetronCloudUrl + "v/" + modelKey;
            htmlDlgMsg = "<h1>Upload Successful</h1><p>Click <a href=\"" + viewerUrl + "\">here<a> to see your uploaded model.<p>The link to the model is:</p><pre>" + viewerUrl + "</pre>" + "This link will be used when creating HTML reports";

            // Store uploaded model url in the situation
            QRMSituationItem* item = static_cast<QRMSituationItem*>(m_ui->situationListWidget->currentItem());
            size_t situationIndex = item->snapshotIndex();

            cee::rep::Snapshot* snapshot = m_repository->snapshot(situationIndex);
            if (snapshot)
            {
                snapshot->setCloudViewerUrl(cee::qt::UtilsCore::toStr(viewerUrl));
            }
        }
        else
        {
            const QString failReason = parseTopLevelJsonValue(responseJson, "message");
            const QString apiErrorCode = parseTopLevelJsonValue(responseJson, "apiErrorCode");
            htmlDlgMsg = QString("<h1>Upload Failed</h1><p>%1 <br> httpStatusCode:%2 <br> apiErrorCode:%3</p>").arg(failReason).arg(httpStatusCode).arg(apiErrorCode);
        }
    }
    else
    {
        htmlDlgMsg = QString("<h1>Network Error During Upload</h1><p>%1</p>").arg(reply->errorString());
    }

    UploadResultDialog resDialog(this, htmlDlgMsg, "Upload to CeetronCloud");
    resDialog.resize(650, 250);
    resDialog.exec();

    reply->deleteLater();

}


//--------------------------------------------------------------------------------------------------
/// Static helper function to do crude parsing of top level JSON values
/// Should be replaced with QJsonDocument for Qt5
//--------------------------------------------------------------------------------------------------
QString QRMMainWindow::parseTopLevelJsonValue(const QString json, const QString key)
{
    QString valString = "";

    const QString quotedKey = QString("\"%1\"").arg(key);
    const int keyIdx = json.indexOf(quotedKey);
    if (keyIdx >= 0)
    {
        int startQuote = json.indexOf("\"", keyIdx + quotedKey.length());
        int endQuote = json.indexOf("\"", startQuote + 1);

        if (startQuote >= 0 && endQuote > (startQuote + 1))
        {
            valString = json.mid(startQuote + 1, endQuote - startQuote - 1);
        }
    }

    return valString;
}


//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void QRMMainWindow::updateUploadProgress(qint64 bytesSent, qint64 bytesTotal)
{
    if (m_progressDialog)
    {
        double progressPct = bytesTotal > 0 ? 100.0*static_cast<double>(bytesSent) / static_cast<double>(bytesTotal) : 100.0;
        m_progressDialog->setValue(static_cast<int>(progressPct));
    }
}


//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void QRMMainWindow::clearCloudUploadId()
{
    QSettings settings;
    settings.setValue("CloudUploadId", "");
}


//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void QRMMainWindow::slotSendToCloud()
{
    // Store uploaded model url in the snapshot
    QRMSituationItem* item = static_cast<QRMSituationItem*>(m_ui->situationListWidget->currentItem());
    size_t situationIndex = item->snapshotIndex();

    cee::rep::Snapshot* snapshot = m_repository->snapshot(situationIndex);
    if (snapshot)
    {
        // Upload snapshot to cloud and save url for use in reports
        sendToCloud(snapshot);
    }
}
