//##################################################################################################
//
//  Ceetron Desktop Components
//  QtSendToCloud
//  
//  --------------------------------------------------------------------------------------------
//  Copyright (C) 2015, Ceetron AS
//  This is UNPUBLISHED PROPRIETARY SOURCE CODE of Ceetron AS. The contents of this file may 
//  not be disclosed to third parties, copied or duplicated in any form, in whole or in part,
//  without the prior written permission of Ceetron AS.
//##################################################################################################


#include <QCoreApplication>
#include <QObject>
#include <QtCore>
#include <QHttpMultiPart>
#include <QNetworkAccessManager>
#include <QNetworkReply>
#include <QSslSocket>
#include <QStringList>
#include <QFile>

#include "QSCCloudTracker.h"

#include <locale.h>
#include <iostream>

//--------------------------------------------------------------------------------------------------
/// SendToCloud app
/// 
/// Usage:
/// sendToCloud -uploadId MY_UPLOAD_ID -uploadApp \"My cool app\" -caseName \"Nice model\" filename.vtfx
//--------------------------------------------------------------------------------------------------
int main(int argc, char *argv[])
{
    // On Linux, Qt will use the system locale, force number formatting settings back to "C" locale
    setlocale(LC_NUMERIC,"C");

    QCoreApplication app(argc, argv);

#if (QT_VERSION >= 0x050000 && defined(QT_NO_SSL)) || (QT_VERSION < 0x050000 && defined(QT_NO_OPENSSL))
    // There is no compile time SSL support in this Qt version
    // Either download another Qt version or recompile Qt with OpenSSl support
    std::cerr << "Your Qt version was not built with OpenSSL support." << std::endl;
    return EXIT_FAILURE;
#else
    // Check that we have SSL/HTTPS support through Qt (normally provided via OpenSSL as a dynamic library)
    if (!QSslSocket::supportsSsl())
    {
        std::cerr << "Your system and/or Qt version does not support secure connections (HTTPS).\nAre the OpenSSL dynamic libraries missing?" << std::endl;
        return EXIT_FAILURE;
    }
#endif

    QString cloudURL = "https://cloud.ceetron.com";
    QString uploadId = "";
    QString uploadApp = "";
    QString caseName = "";
    QString filename = "";
    QString parentKey = "";
    QString caseId = "";
    int uploadIdIndex = -1;
    int uploadAppIndex = -1;
    int caseNameIndex = -1;
    int cloudUrlIndex = -1;
    int parentKeyIndex = -1;
    int caseIdIndex = -1;

    // Parse command line parameters
    QStringList arguments = app.arguments();

    if (arguments.size() == 1)
    {
        std::cout << "sendToCloud v. 1.2-0" << std::endl << "---------------------" << std::endl << std::endl;
        std::cout << "Note: Default upload URL: " << cloudURL.toStdString() << std::endl << std::endl;
        std::cout << "Usage:" << std::endl << std::endl;
        std::cout << "> sendToCloud -url CLOUD_HOST_NAME -uploadId MY_UPLOAD_ID -uploadApp \"My cool app\" -caseName \"Nice model\" -parentKey _sHGmPcHTL-mgytYUBUhHg filename.vtfx" << std::endl << std::endl;
        std::cout << "     -url       The URL to the instance of Ceetron Cloud (Private) to send the model to"  << std::endl;
        std::cout << "     -uploadId  The upload ID of the account that the model will be added to. This can be found in the home page of the user on the Ceetron Cloud (private) instance." << std::endl;
        std::cout << "     -caseName  The name of the case that will be displayed in the portal" << std::endl;
        std::cout << "     -uploadApp The source of the upload. Information only."  << std::endl;
        std::cout << "     -parentKey (Optional) The key of the folder the model will be put into. The key can be found in the portal." << std::endl;
        std::cout << "     -caseId    (Optional) The ID of the case in the VTFx file to generate a view for."  << std::endl  << std::endl;
        return EXIT_FAILURE;
    }

    for (int i = 1; i < arguments.size(); ++i) 
    {
        if (uploadIdIndex == i) 
        {
            uploadId = arguments[i];
            uploadIdIndex = -1;
        } 
        else if (uploadAppIndex == i) 
        {
            uploadApp = arguments[i];
            uploadAppIndex = -1;
        }
        else if (caseNameIndex == i) 
        {
            caseName = arguments[i];
            caseNameIndex = -1;
        }
        else if (cloudUrlIndex == i) 
        {
            cloudURL = arguments[i];
            cloudUrlIndex = -1;
        }
        else if (parentKeyIndex == i)
        {
            parentKey = arguments[i];
            parentKeyIndex = -1;
        }
        else if (caseIdIndex == i)
        {
            caseId = arguments[i];
            caseIdIndex = -1;
        }
        else if (arguments[i].toLower() == "-uploadid")
        {
            uploadIdIndex = i + 1;
        }
        else if (arguments[i].toLower() == "-uploadapp")
        {
            uploadAppIndex = i + 1;
        }
        else if (arguments[i].toLower() == "-casename")
        {
            caseNameIndex = i + 1;
        }
        else if (arguments[i].toLower() == "-url")
        {
            cloudUrlIndex = i + 1;
        }
        else if (arguments[i].toLower() == "-parentkey")
        {
            parentKeyIndex = i + 1;
        }
        else if (arguments[i].toLower() == "-caseid")
        {
            caseIdIndex = i + 1;
        }
        else
        {
            filename = arguments[i];
        }
    }

    // Check that all the needed info has been provided
    if (filename.isEmpty()) 
    {
        std::cout << "No VTFx file name provided" << std::endl;
        return EXIT_FAILURE;
    }

    if (uploadId.isEmpty()) 
    {
        std::cout << "No Upload ID provided (-uploadId)" << std::endl;
        return EXIT_FAILURE;
    }

    if (uploadApp.isEmpty()) 
    {
        std::cout << "No Upload App provided (-uploadApp)" << std::endl;
        return EXIT_FAILURE;
    }

    if (caseName.isEmpty()) 
    {
        std::cout << "No case name provided (-caseName)" << std::endl;
        return EXIT_FAILURE;
    }

    // We're ready to SendToCloud!
    QHttpMultiPart *multiPart = new QHttpMultiPart(QHttpMultiPart::FormDataType);
    multiPart->setBoundary("boundary_Ceetron89906a4aJn1kvlAG52SFG2AD");

    QHttpPart vtfxPart;
    vtfxPart.setHeader(QNetworkRequest::ContentTypeHeader, QVariant("application/octet-stream"));

    if (filename.toLower().indexOf(".vtfx") >= 0) 
    {
        vtfxPart.setHeader(QNetworkRequest::ContentDispositionHeader, QVariant("form-data; name=\"vtfx\" filename=\"" + filename + "\""));
    } 
    else 
    {
        vtfxPart.setHeader(QNetworkRequest::ContentDispositionHeader, QVariant("form-data; name=\"vtf\" filename=\"" + filename + "\""));
    }

    QFile *file = new QFile(filename);
    if (!file->open(QIODevice::ReadOnly))
    {
        std::cout << "Error opening file: " << filename.toStdString() << std::endl;
        return EXIT_FAILURE;
    }

    vtfxPart.setBodyDevice(file);

    file->setParent(multiPart); 
    multiPart->append(vtfxPart);

    QString urlString = QString(cloudURL + "/api/v1/models?uploadId=%1&uploadApp=%2&name=%3").arg(QString(QUrl::toPercentEncoding(uploadId))).arg(QString(QUrl::toPercentEncoding(uploadApp))).arg(QString(QUrl::toPercentEncoding(caseName)));
    if (parentKey.length() > 0)
    {
        urlString += QString("&parentKey=%1").arg(parentKey);
    }
    if (caseId.length() > 0)
    {
        urlString += QString("&caseId=%1").arg(caseId);
    }

    std::cout << "#INFO: Using URL: " << urlString.toStdString() << std::endl;

    QUrl url(urlString);
    QNetworkRequest request(url);

    QNetworkAccessManager* manager = new QNetworkAccessManager(NULL);
    QNetworkReply* reply = manager->post(request, multiPart);
    multiPart->setParent(reply);

    CloudUploadTracker* uploadTracker = new CloudUploadTracker(&app, cloudURL);
    QObject::connect(reply, SIGNAL(finished()), uploadTracker, SLOT(cloudUploadFinished()));
    QObject::connect(reply, SIGNAL(uploadProgress(qint64, qint64)), uploadTracker, SLOT(updateUploadProgress(qint64, qint64)));
    QObject::connect(uploadTracker, SIGNAL(finished()), &app, SLOT(quit()));

    std::cout << "#INFO: Uploading file to Ceetron Cloud..." << std::endl;

    int result = app.exec();
    std::cout << "#INFO: Done." << std::endl;

    return result;
}
