//##################################################################################################
//
//  Ceetron Desktop Components
//  WxMinimal
//  
//  --------------------------------------------------------------------------------------------
//  Copyright (C) 2011, Ceetron AS
//  This is UNPUBLISHED PROPRIETARY SOURCE CODE of Ceetron AS. The contents of this file may 
//  not be disclosed to third parties, copied or duplicated in any form, in whole or in part,
//  without the prior written permission of Ceetron AS.
//##################################################################################################

#include "WMMainFrame.h"

#include "wx/wx.h"

#include "CeeCore/Version.h"

#include "CeeVisualization/VisualizationComponent.h"
#include "CeeVisualization/Camera.h"

#include "CeeGeometry/GeometryModel.h"
#include "CeeGeometry/DataGenerator.h"
#include "CeeGeometry/Part.h"
#include "CeeGeometry/PartSettings.h"
#include "CeeGeometry/EffectColor.h"
#include "CeeGeometry/EffectHalo.h"



//==================================================================================================
//
// 
//
//==================================================================================================

enum
{
    WM_CreateDefaultScene = 1,
    WM_ClearScene
};

BEGIN_EVENT_TABLE(WMMainFrame, wxFrame)
    EVT_SHOW(WMMainFrame::onShow)
    EVT_CLOSE(WMMainFrame::onClose)

    EVT_MENU(WM_CreateDefaultScene, WMMainFrame::onCreateDefaultScene)
    EVT_MENU(WM_ClearScene, WMMainFrame::onClearScene)
    EVT_MENU(wxID_EXIT, WMMainFrame::onExit)
    EVT_MENU(wxID_ABOUT, WMMainFrame::onAbout)
END_EVENT_TABLE()

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
WMMainFrame::WMMainFrame()
:   wxFrame(NULL, wxID_ANY, L"WxMinimal"),
    m_contextGroup(cee::vis::VisualizationComponent::createOpenGLContextGroup())
{
    wxMenu* testMenu = new wxMenu;
    testMenu->Append(WM_CreateDefaultScene, L"Default Scene");
    testMenu->Append(WM_ClearScene, L"Clear Scene");
    testMenu->AppendSeparator();
    testMenu->Append(wxID_EXIT);

    wxMenu* helpMenu = new wxMenu;
    helpMenu->Append(wxID_ABOUT, L"&About");

    wxMenuBar* menuBar = new wxMenuBar;
    menuBar->Append(testMenu, L"&Test");
    menuBar->Append(helpMenu, L"&Help");
    SetMenuBar(menuBar);


    // According to docs, passing NULL for attributes should work, but for now it seems we need to
    // pass an explicit list in order to get a double buffered context on Linux
    int i = 0;
    int attribList[10];
    attribList[i++] = WX_GL_RGBA;
    attribList[i++] = WX_GL_DOUBLEBUFFER;
    attribList[i++] = WX_GL_DEPTH_SIZE;     attribList[i++] = 1;
    attribList[i++] = 0;

    // Need to check that creation will be OK before actually doing it, 
    // otherwise we get into trouble while constructing the canvas below.
    // In production code this scenario must be handled and/or reported to the user
    wxASSERT(wxGLCanvas::IsDisplaySupported(attribList));

    // Create the viewer here, but defer initialization of viewer 
    // until after we have been shown (see onShow()))
    m_viewer = new cee::wx::ViewerCanvasDemo(m_contextGroup.get(), this, attribList);
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void WMMainFrame::onShow(wxShowEvent&)
{
    // Should be shown by now, so try and initialize the viewer before setting up the default scene
    wxASSERT(m_viewer);
    m_viewer->initializeViewer();

    setupDefaultScene();
}



//--------------------------------------------------------------------------------------------------
/// Intercepts default handling of the close event to do cleanup of the viewer
//--------------------------------------------------------------------------------------------------
void WMMainFrame::onClose(wxCloseEvent&)
{
    // Shut down the viewer before destroying windows
    m_viewer->shutdownViewer();

    Destroy();
}



//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void WMMainFrame::onCreateDefaultScene(wxCommandEvent&)
{
    setupDefaultScene();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void WMMainFrame::onClearScene(wxCommandEvent&)
{
    m_viewer->setView(NULL);
    m_viewer->Refresh();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void WMMainFrame::onExit(wxCommandEvent&)
{
    // Just call close and let onClose() event handler do cleanup
    Close(true);
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void WMMainFrame::onAbout(wxCommandEvent&)
{
    wxString verInfo;
    verInfo << L"wxWidgets ver.:  " << wxString::FromAscii(wxVERSION_NUM_DOT_STRING) << L"\n";
    verInfo << L"Ceetron Desktop Components ver.:  " << wxString::FromAscii(CEE_VERSION_STRING) << L"\n";
    wxMessageBox(verInfo, L"About WxMultiView", wxOK | wxICON_INFORMATION, this);
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void WMMainFrame::setupDefaultScene()
{
    wxASSERT(m_viewer);

    m_viewer->setView(NULL);
    cee::vis::View* gcView = new cee::vis::View;
    m_viewer->setView(gcView);

    cee::PtrRef<cee::geo::GeometryModel> boxModel = new cee::geo::GeometryModel;
    gcView->addModel(boxModel.get());
    {
        cee::geo::Part* part = new cee::geo::Part(cee::geo::DataGenerator::createBox(cee::Vec3d(0, 0, 0), cee::Vec3d(1, 0.75f, 0.5)).get());
        part->settings().addEffect(new cee::geo::EffectColor(cee::Color3f(0.8f, 0.2f, 0.2f)));
        part->settings().addEffect(new cee::geo::EffectHalo);
        boxModel->addPart(part);
    }
    {
        cee::geo::Part* part = new cee::geo::Part(cee::geo::DataGenerator::createBox(cee::Vec3d(1.1, 0, 0), cee::Vec3d(1, 0.75f, 0.5)).get());
        part->settings().addEffect(new cee::geo::EffectColor(cee::Color3f(0.2f, 0.7f, 0.2f)));
        boxModel->addPart(part);
    }
    {
        cee::geo::Part* part = new cee::geo::Part(cee::geo::DataGenerator::createBox(cee::Vec3d(0, 1.1, 0), cee::Vec3d(1, 0.75f, 0.5)).get());
        part->settings().addEffect(new cee::geo::EffectColor(cee::Color3f(0.2f, 0.2f, 0.7f)));
        boxModel->addPart(part);
    }

    cee::PtrRef<cee::geo::GeometryModel> sphereModel = new cee::geo::GeometryModel;
    gcView->addModel(sphereModel.get());
    {
        cee::geo::Part* part = new cee::geo::Part(cee::geo::DataGenerator::createSphere(cee::Vec3d(1.1, 1.1, 0), 0.5, 10).get());
        part->settings().addEffect(new cee::geo::EffectColor(cee::Color3f(0.7f, 0.7f, 0.0f)));
        sphereModel->addPart(part);
    }

    cee::BoundingBox bb = gcView->boundingBox();
    gcView->camera().fitView(bb, cee::Vec3d(0, 0, -1), cee::Vec3d(0, 1, 0));
    gcView->requestRedraw();
}



