//##################################################################################################
//
//  Ceetron Desktop Components
//  WxMultiView
//  
//  --------------------------------------------------------------------------------------------
//  Copyright (C) 2011, Ceetron AS
//  This is UNPUBLISHED PROPRIETARY SOURCE CODE of Ceetron AS. The contents of this file may 
//  not be disclosed to third parties, copied or duplicated in any form, in whole or in part,
//  without the prior written permission of Ceetron AS.
//##################################################################################################

#include "WMVMainFrame.h"

#include "wx/wx.h"

#include "CeeCore/Version.h"

#include "CeeVisualization/VisualizationComponent.h"
#include "CeeVisualization/Camera.h"
#include "CeeVisualization/Font.h"
#include "CeeVisualization/Overlay.h"
#include "CeeVisualization/OverlayTextBox.h"

#include "CeeGeometry/GeometryModel.h"
#include "CeeGeometry/DataGenerator.h"
#include "CeeGeometry/Part.h"
#include "CeeGeometry/PartSettings.h"
#include "CeeGeometry/EffectColor.h"
#include "CeeGeometry/EffectHalo.h"

#include "CeeUnstructGrid/DataSourceMemory.h"
#include "CeeUnstructGrid/UnstructGridModel.h"
#include "CeeUnstructGrid/ModelSpec.h"
#include "CeeUnstructGrid/PartSettings.h"



//==================================================================================================
//
// 
//
//==================================================================================================

enum
{
    WMV_Create1Viewer = 1,
    WMV_Create2Viewers,
    WMV_Create3Viewers,
    WMV_CreateCylinderInFirstViewer,
    WMV_DeleteAllViews,
    WMV_DeleteAllViewers
};

BEGIN_EVENT_TABLE(WMVMainFrame, wxFrame)
    EVT_SHOW(WMVMainFrame::onShow)
    EVT_CLOSE(WMVMainFrame::onClose)

    EVT_MENU_RANGE(WMV_Create1Viewer, WMV_Create3Viewers, WMVMainFrame::onCreateNViewers)
    EVT_MENU(WMV_CreateCylinderInFirstViewer, WMVMainFrame::onCreateCylinderInFirstViewer)
    EVT_MENU(WMV_DeleteAllViews, WMVMainFrame::onDeleteAllViews)
    EVT_MENU(WMV_DeleteAllViewers, WMVMainFrame::onDeleteAllViewers)
    EVT_MENU(wxID_EXIT, WMVMainFrame::onExit)
    EVT_MENU(wxID_ABOUT, WMVMainFrame::onAbout)
END_EVENT_TABLE()

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
WMVMainFrame::WMVMainFrame()
:   wxFrame(NULL, wxID_ANY, L"WxMultiView"),
    m_contextGroup(cee::vis::VisualizationComponent::createOpenGLContextGroup())
{
    wxMenu* testMenu = new wxMenu;
    testMenu->Append(WMV_Create1Viewer,  L"Create &1 Viewer\tCtrl+1");
    testMenu->Append(WMV_Create2Viewers, L"Create &2 Viewers\tCtrl+2");
    testMenu->Append(WMV_Create3Viewers, L"Create &3 Viewers\tCtrl+3");
    testMenu->AppendSeparator();
    testMenu->Append(WMV_CreateCylinderInFirstViewer, L"Create Cylinder In First Viewer");
    testMenu->Append(WMV_DeleteAllViews, L"Delete All Views");
    testMenu->Append(WMV_DeleteAllViewers, L"&Delete All Viewers\tCtrl+D");
    testMenu->AppendSeparator();
    testMenu->Append(wxID_EXIT);

    wxMenu* helpMenu = new wxMenu;
    helpMenu->Append(wxID_ABOUT, L"&About");

    wxMenuBar* menuBar = new wxMenuBar;
    menuBar->Append(testMenu, L"&Test");
    menuBar->Append(helpMenu, L"&Help");
    SetMenuBar(menuBar);

    // For debugging
    SetBackgroundColour(*wxGREEN);

    wxBoxSizer* sizer = new wxBoxSizer(wxHORIZONTAL);
    SetSizer(sizer);
    Layout();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void WMVMainFrame::onShow(wxShowEvent&)
{
    createViewers(2);
}


//--------------------------------------------------------------------------------------------------
/// Intercepts default handling of the close event.
/// 
/// Clean up the viewers and their resources here before the actual windows get destroyed
//--------------------------------------------------------------------------------------------------
void WMVMainFrame::onClose(wxCloseEvent&)
{
    deleteViewers();
    
    Destroy();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void WMVMainFrame::onCreateNViewers(wxCommandEvent& event)
{
    int numViews = event.GetId();

    createViewers(numViews);
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void WMVMainFrame::onCreateCylinderInFirstViewer(wxCommandEvent&)
{
    cee::wx::ViewerCanvas* aViewer = m_viewers.size() > 0 ? m_viewers[0] : NULL;
    wxASSERT(aViewer);

    cee::vis::View* view = aViewer->view();
    if (view)
    {
        view->removeAllModels();
    }
    else
    {
        view = new cee::vis::View;
    }

    populateView_cylinder(view);
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void WMVMainFrame::onDeleteAllViews(wxCommandEvent&)
{
    const size_t numViewers = m_viewers.size();
    for (size_t i = 0; i < numViewers; i++)
    {
        m_viewers[i]->setView(NULL);
        m_viewers[i]->Refresh();
    }
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void WMVMainFrame::onDeleteAllViewers(wxCommandEvent&)
{
    deleteViewers();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void WMVMainFrame::onExit(wxCommandEvent&)
{
    // Just call close and let onClose() event handler do cleanup
    Close(true);
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void WMVMainFrame::onAbout(wxCommandEvent&)
{
    wxString verInfo;
    verInfo << L"wxWidgets ver.:  " << wxString::FromAscii(wxVERSION_NUM_DOT_STRING) << L"\n";
    verInfo << L"Ceetron Desktop Components ver.:  " << wxString::FromAscii(CEE_VERSION_STRING) << L"\n";
    wxMessageBox(verInfo, L"About WxMultiView", wxOK | wxICON_INFORMATION, this);
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void WMVMainFrame::deleteViewers()
{
    // Experience show that on Linux/wxGTK, when we're running on physical hardware, the system is sensistive
    // to timing issues when shutting down and deleting OpenGL resources.
    // To counter this, we try and do all cleanup and freeing of resources before deleting the actual windows.

    // Loop will shut down all the viewers, thereby releasing OpenGL resources
    const size_t numViewers = m_viewers.size();
    for (size_t i = 0; i < numViewers; i++)
    {
        m_viewers[i]->shutdownViewer();
    }

    wxSizer* theSizer = GetSizer();
    while (theSizer->GetChildren().GetCount() > 0)
    {
        theSizer->Detach(0);
    }

    while (!m_viewers.empty())
    {
        m_viewers.back()->Destroy();
        m_viewers.pop_back();
    }

    wxASSERT(theSizer->GetChildren().GetCount() == 0);
    wxASSERT(m_viewers.empty());
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void WMVMainFrame::createViewers(size_t numViewers)
{
    wxASSERT(numViewers > 0);

    deleteViewers();

    int i = 0;
    int attribList[10];
    attribList[i++] = WX_GL_RGBA;
    attribList[i++] = WX_GL_DOUBLEBUFFER;
    attribList[i++] = WX_GL_DEPTH_SIZE;     attribList[i++] = 1;
    attribList[i++] = 0;
    wxASSERT(wxGLCanvas::IsDisplaySupported(attribList));

    wxSizer* theSizer = GetSizer();
    cee::PtrRef<cee::vis::Font> font =  cee::vis::Font::createNormalFont();

    cee::wx::ViewerCanvasDemo* firstViewer = NULL;
    for (size_t i = 0; i < numViewers; i++)
    {
        cee::wx::ViewerCanvasDemo* viewer;
        if (!firstViewer) 
        {
            viewer = new cee::wx::ViewerCanvasDemo(m_contextGroup.get(), this, attribList);
            firstViewer = viewer;
        }
        else
        {
            viewer = new cee::wx::ViewerCanvasDemo(m_contextGroup.get(), this, attribList, firstViewer);
        }

        theSizer->Add(viewer, wxSizerFlags(1).Expand().Border(wxALL, 5));
        viewer->initializeViewer();

        cee::vis::View* view = new cee::vis::View;
        viewer->setView(view);

        {
            cee::vis::OverlayTextBox* viewDescription = new cee::vis::OverlayTextBox(font.get());
            viewDescription->setText(cee::Str("View %1").arg(static_cast<int>(i + 1)));
            viewDescription->setSizeToFitText();
            view->overlay().addItem(viewDescription, cee::vis::OverlayItem::TOP_LEFT, cee::vis::OverlayItem::VERTICAL);
        }

        size_t m = i % 3;
        if      (m == 0) populateView_sphereAndBoxesHalo(view);
        else if (m == 1) populateView_unstructGridModel(view);
        else if (m == 2) populateView_cylinder(view);
        
        m_viewers.push_back(viewer);
    }

    theSizer->Layout();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void WMVMainFrame::populateView_sphereAndBoxesHalo(cee::vis::View* view)
{
    cee::PtrRef<cee::geo::GeometryModel> boxModel = new cee::geo::GeometryModel;
    view->addModel(boxModel.get());
    {
        cee::geo::Part* part = new cee::geo::Part(cee::geo::DataGenerator::createBox(cee::Vec3d(0, 0, 0), cee::Vec3d(1, 0.75f, 0.5)).get());
        part->settings().addEffect(new cee::geo::EffectColor(cee::Color3f(0.8f, 0.2f, 0.2f)));
        boxModel->addPart(part);
    }
    {
        cee::geo::Part* part = new cee::geo::Part(cee::geo::DataGenerator::createBox(cee::Vec3d(1.1, 0, 0), cee::Vec3d(1, 0.75f, 0.5)).get());
        part->settings().addEffect(new cee::geo::EffectColor(cee::Color3f(0.2f, 0.7f, 0.2f)));
        boxModel->addPart(part);
    }
    {
        cee::geo::Part* part = new cee::geo::Part(cee::geo::DataGenerator::createBox(cee::Vec3d(0, 1.1, 0), cee::Vec3d(1, 0.75f, 0.5)).get());
        part->settings().addEffect(new cee::geo::EffectColor(cee::Color3f(0.2f, 0.2f, 0.7f)));
        boxModel->addPart(part);
    }

    cee::PtrRef<cee::geo::GeometryModel> sphereModel = new cee::geo::GeometryModel;
    view->addModel(sphereModel.get());
    {
        cee::geo::Part* part = new cee::geo::Part(cee::geo::DataGenerator::createSphere(cee::Vec3d(1.1, 1.1, 0), 0.5, 10).get());
        part->settings().addEffect(new cee::geo::EffectColor(cee::Color3f(0.7f, 0.7f, 0.0f)));
        part->settings().addEffect(new cee::geo::EffectHalo);
        sphereModel->addPart(part);
    }

    cee::BoundingBox bb = view->boundingBox();
    view->camera().fitView(bb, cee::Vec3d(0, 0, -1), cee::Vec3d(0, 1, 0));
    view->requestRedraw();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void WMVMainFrame::populateView_unstructGridModel(cee::vis::View* view)
{
    cee::PtrRef<cee::ug::DataSourceMemory> dataSource = new cee::ug::DataSourceMemory(1, 1);
    {
        cee::PtrRef<cee::ug::DataNodes> nodes = new cee::ug::DataNodes(false);
        nodes->resize(7);
        nodes->setNode(0, cee::Vec3d(0, 0, 0));
        nodes->setNode(1, cee::Vec3d(2, 0, 0));
        nodes->setNode(2, cee::Vec3d(2, 2, 0));
        nodes->setNode(3, cee::Vec3d(0, 2, 0));
        nodes->setNode(4, cee::Vec3d(0, 3, 0));
        nodes->setNode(5, cee::Vec3d(2, 3, 0));
        nodes->setNode(6, cee::Vec3d(1, 5, 0));

        cee::PtrRef<cee::ug::DataElements> elements = new cee::ug::DataElements(false, 0);
        const unsigned int elemConnQuad[] = {0, 1, 2, 3};
        const unsigned int elemConnTri[] = {4, 5, 6};
        elements->addElement(cee::ug::Element::QUADS, elemConnQuad, 4);
        elements->addElement(cee::ug::Element::TRIANGLES, elemConnTri, 3);

        cee::PtrRef<cee::ug::DataPart> part = new cee::ug::DataPart(1);
        part->setNodes(nodes.get());
        part->setElements(elements.get());

        cee::PtrRef<cee::ug::DataGeometry> geo = new cee::ug::DataGeometry;
        geo->addPart(part.get());

        cee::PtrRef<cee::ug::DataState> state = new cee::ug::DataState(1, 1);
        state->setGeometry(0, geo.get());

        dataSource->addState(state.get());
        dataSource->updateDirectoryFromStates();
    }

    {
        cee::PtrRef<cee::ug::UnstructGridModel> model = new cee::ug::UnstructGridModel;
        view->addModel(model.get());
        model->setDataSource(dataSource.get());
        model->modelSpec().setStateId(1);
        model->partSettings(0, 1)->setColor(cee::Color3f(0.8f, 0.8f, 0));
        model->partSettings(0, 1)->setDrawStyle(cee::ug::PartSettings::SURFACE_MESH);
        model->updateVisualization();
    }

    cee::BoundingBox bb = view->boundingBox();
    view->camera().fitView(bb, cee::Vec3d(0, 0, -1), cee::Vec3d(0, 1, 0));
    view->requestRedraw();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void WMVMainFrame::populateView_cylinder(cee::vis::View* view)
{
    cee::PtrRef<cee::geo::GeometryModel> model = new cee::geo::GeometryModel;
    view->addModel(model.get());
    
    cee::geo::Part* part = new cee::geo::Part(cee::geo::DataGenerator::createCylinder(cee::Vec3d(0, 0, 0), 1, 1, cee::Vec3d(0, 1, 0), 2, 20).get());
    part->settings().addEffect(new cee::geo::EffectColor(cee::Color3f(0.8f, 0.2f, 0.2f)));
    model->addPart(part);

    cee::BoundingBox bb = view->boundingBox();
    view->camera().fitView(bb, cee::Vec3d(0, 0, -1), cee::Vec3d(0, 1, 0));
    view->requestRedraw();
}


