//##################################################################################################
//
//  Ceetron Desktop Components
//  Platform integration for Qt
//  
//  --------------------------------------------------------------------------------------------
//  Copyright (C) 2018, Ceetron AS
//  This is UNPUBLISHED PROPRIETARY SOURCE CODE of Ceetron AS. The contents of this file may 
//  not be disclosed to third parties, copied or duplicated in any form, in whole or in part,
//  without the prior written permission of Ceetron AS.
//##################################################################################################

#include "ceeqtOpenGLFunctionsBackend_qt5.h"

#include "CeeCore/CoreComponent.h"
#include "CeeCore/LogManager.h"

#include <QtGlobal>

#include <QSet>

#if QT_VERSION < 0x050600
#error This Qt Backend requires Qt 5.6 or later
#endif

#include <QOpenGLContext>
#include <QOpenGLFunctions>
#include <QOpenGLExtraFunctions>
#include <QOpenGLFunctions_2_1>
#include <QOpenGLFunctions_3_2_Core>

#if QT_VERSION >= 0x060000
#include <QOpenGLVersionFunctionsFactory>
#endif

// Get rid of unused warning -> error
#if defined(__clang__)
#pragma clang diagnostic ignored "-Wunused-parameter"
#endif

#ifdef WIN32
#pragma warning (disable: 4100)
#endif

#if defined(__GNUC__) && (CVF_GCC_VER >= 40200)
_Pragma(GCC diagnostic ignored "-Wunused-parameter")
#endif

namespace cee {
namespace qt {


    
//==================================================================================================
//
// 
//
//==================================================================================================

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
OpenGLFunctionsBackend_qt5::OpenGLFunctionsBackend_qt5()
 :  m_qtFuncs(NULL),
    m_qtExtraFuncs(NULL),
    m_extraFuncFlags(0),
    m_qtFuncs_21(NULL),
    m_qtFuncs_32(NULL),
    m_loggerName("cee.comp.qt")
{
    m_logMgr = cee::CoreComponent::logManager();
}
    
//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
OpenGLFunctionsBackend_qt5::~OpenGLFunctionsBackend_qt5()
{
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
bool OpenGLFunctionsBackend_qt5::initializeBackend(cee::vis::OpenGLBackendInfo* backendInfo)
{
    QOpenGLContext* currQtContext = QOpenGLContext::currentContext();
    Q_ASSERT(currQtContext);

    const QSurfaceFormat surfFormat = currQtContext->format();
    const QSurfaceFormat::RenderableType renderableType = surfFormat.renderableType();
    const QSurfaceFormat::OpenGLContextProfile profile = surfFormat.profile();
    const QOpenGLContext::OpenGLModuleType openGLModuleType = QOpenGLContext::openGLModuleType();
    const int majorVersion = surfFormat.majorVersion();
    const int minorVersion = surfFormat.minorVersion();

    cee::Str sRenderableType = "UNKNOWN";
    if      (renderableType == QSurfaceFormat::DefaultRenderableType) sRenderableType = "DefaultRenderableType";
    else if (renderableType == QSurfaceFormat::OpenGL)                sRenderableType = "OpenGL";
    else if (renderableType == QSurfaceFormat::OpenGLES)              sRenderableType = "OpenGLES";
    else if (renderableType == QSurfaceFormat::OpenVG)                sRenderableType = "OpenVG";

    cee::Str sProfile = "UNKNOWN";
    if      (profile == QSurfaceFormat::NoProfile)            sProfile = "NoProfile";
    else if (profile == QSurfaceFormat::CoreProfile)          sProfile = "CoreProfile";
    else if (profile == QSurfaceFormat::CompatibilityProfile) sProfile = "CompatibilityProfile";

    cee::Str sOpenGLModuleType = "UNKNOWN";
    if      (openGLModuleType == QOpenGLContext::LibGL)     sOpenGLModuleType = "LibGL";
    else if (openGLModuleType == QOpenGLContext::LibGLES)   sOpenGLModuleType = "LibGLES";

    // Write debug info about what we discover to log
    m_logMgr->logDebug(m_loggerName, "Using OpenGLFunctionsBackend_qt5");
    m_logMgr->logDebug(m_loggerName, cee::Str("  qt.OpenGL version %1.%2").arg(majorVersion).arg(minorVersion));
    m_logMgr->logDebug(m_loggerName, "  qt.RenderableType: " + sRenderableType);
    m_logMgr->logDebug(m_loggerName, "  qt.Profile: " + sProfile);
    m_logMgr->logDebug(m_loggerName, "  qt.OpenGLModuleType: " + sOpenGLModuleType);

    // Try and log the file name of the OpenGL DLL in use
#ifdef WIN32
    cee::Str sOpenGLModuleFileName = "UNKNOWN";
#if QT_VERSION >= 0x060000
	HMODULE hOpenGLModule = QNativeInterface::QWGLContext::openGLModuleHandle();
#elif QT_VERSION >= 0x050300
	HMODULE hOpenGLModule = static_cast<HMODULE>(QOpenGLContext::openGLModuleHandle());
#endif
	if (hOpenGLModule)
    {
        wchar_t szFileNameBuffer[MAX_PATH];
        if (::GetModuleFileNameW(hOpenGLModule, szFileNameBuffer, sizeof(szFileNameBuffer)))
        {
            sOpenGLModuleFileName = cee::Str(szFileNameBuffer);
        }
    }
    m_logMgr->logDebug(m_loggerName, "  qt.OpenGLModuleFileName: " + sOpenGLModuleFileName);
#endif


    const bool isDesktop = renderableType == QSurfaceFormat::OpenGL ? true : false;
    const bool isES = renderableType == QSurfaceFormat::OpenGLES ? true : false;
    const bool hasDesktop30 = isDesktop && majorVersion >= 3;
    const bool hasDesktop31 = isDesktop && (majorVersion > 3 || (majorVersion == 3 && minorVersion >= 1));
    const bool hasDesktop40 = isDesktop && majorVersion >= 4;
    const bool hasES30 = isES && majorVersion >= 3;
    const bool hasES31 = isES && (majorVersion > 3 || (majorVersion == 3 && minorVersion >= 1));

    // Reset pointers in case we fail to get the main functions
    m_qtExtraFuncs = NULL;
    m_extraFuncFlags = 0;
    m_qtFuncs_21 = NULL;
    m_qtFuncs_32 = NULL;

    m_qtFuncs = currQtContext->functions();
    if (!m_qtFuncs)
    {
        return false;
    }

    m_qtExtraFuncs = currQtContext->extraFunctions();
    if (m_qtExtraFuncs)
    {
        if (isDesktop     || hasES30)   m_extraFuncFlags |= has_glDrawBuffers;
        if (isDesktop     || hasES31)   m_extraFuncFlags |= has_glDrawRangeElements;
        if (hasDesktop30  || hasES30)   m_extraFuncFlags |= has_glRenderbufferStorageMultisample;
        if (hasDesktop30  || hasES30)   m_extraFuncFlags |= has_glBlitFramebuffer;
        if (hasDesktop30  || hasES30)   m_extraFuncFlags |= has_glVertexAttribIPointer;
        
        if (currQtContext->hasExtension("GL_EXT_framebuffer_multisample"))
        {
            m_extraFuncFlags |= has_glRenderbufferStorageMultisample;
        }
        
        if (currQtContext->hasExtension("GL_EXT_framebuffer_blit"))
        {
            m_extraFuncFlags |= has_glBlitFramebuffer;
        }
        
        if (currQtContext->hasExtension("GL_ARB_shader_storage_buffer_object"))
        {
            m_extraFuncFlags |= has_glVertexAttribIPointer;
        }
    }

    // The Qt "version functions" cannot be used on OpenGL ES
    if (isDesktop)
    {
	#if QT_VERSION >= 0x060000
		m_qtFuncs_21 = QOpenGLVersionFunctionsFactory::get<QOpenGLFunctions_2_1>(currQtContext);
		m_qtFuncs_32 = QOpenGLVersionFunctionsFactory::get<QOpenGLFunctions_3_2_Core>(currQtContext);
	#else
        m_qtFuncs_21 = currQtContext->versionFunctions<QOpenGLFunctions_2_1>();
        m_qtFuncs_32 = currQtContext->versionFunctions<QOpenGLFunctions_3_2_Core>();
	#endif
	}


    backendInfo->oglMajorVersion = majorVersion;
    backendInfo->oglMinorVersion = minorVersion;
    backendInfo->isOpenGLES = isES;

    if (m_qtFuncs->hasOpenGLFeature(QOpenGLFunctions::Framebuffers))
    {
        backendInfo->capabilityFlags |= cee::vis::OpenGLBackendInfo::FRAMEBUFFER_OBJECT;
    }

    if (hasDesktop30 || hasES30 || currQtContext->hasExtension("GL_ARB_texture_float"))
    {
        backendInfo->capabilityFlags |= cee::vis::OpenGLBackendInfo::TEXTURE_FLOAT;
    }

    if (hasDesktop30 || hasES30 || currQtContext->hasExtension("GL_ARB_texture_rg"))
    {
        backendInfo->capabilityFlags |= cee::vis::OpenGLBackendInfo::TEXTURE_RG;
    }

    if (hasDesktop40 || hasDesktop31 || currQtContext->hasExtension("GL_ARB_texture_rectangle"))
    {
        backendInfo->capabilityFlags |= cee::vis::OpenGLBackendInfo::TEXTURE_RECTANGLE;
    }

    if (hasDesktop30 || currQtContext->hasExtension("GL_EXT_gpu_shader4"))
    {
        backendInfo->capabilityFlags |= cee::vis::OpenGLBackendInfo::FLAT_INTERPOLATION_QUALIFIER;
    }
    
    if (m_extraFuncFlags & has_glBlitFramebuffer)
    {
        backendInfo->capabilityFlags |= cee::vis::OpenGLBackendInfo::BLIT_FRAMEBUFFER;
    }

    cee::Str sVersion(reinterpret_cast<const char*>(m_qtFuncs->glGetString(GL_VERSION)));
    cee::Str sVendor(reinterpret_cast<const char*>(m_qtFuncs->glGetString(GL_VENDOR)));
    cee::Str sRenderer(reinterpret_cast<const char*>(m_qtFuncs->glGetString(GL_RENDERER)));
    cee::Str sShadingLangVersion(reinterpret_cast<const char*>(m_qtFuncs->glGetString(GL_SHADING_LANGUAGE_VERSION)));
    m_logMgr->logDebug(m_loggerName, "  GL_VENDOR: " + sVendor);
    m_logMgr->logDebug(m_loggerName, "  GL_RENDERER: " + sRenderer);
    m_logMgr->logDebug(m_loggerName, "  GL_VERSION: " + sVersion);
    m_logMgr->logDebug(m_loggerName, "  GL_SHADING_LANGUAGE_VERSION: " + sShadingLangVersion);

    // For debugging, logging of all available extensions
    //m_logMgr->logDebug(m_loggerName, "Extensions reported through Qt:");
    //QList<QByteArray> extensionList = currQtContext->extensions().toList();
    //std::sort(extensionList.begin(), extensionList.end());
    //for (int i = 0; i < extensionList.size(); i++)
    //{
    //    cee::Str extStr(extensionList[i]);
    //    if (!extStr.isEmpty())
    //    {
    //        m_logMgr->logDebug(m_loggerName, "  " + cee::Str(extensionList[i]));
    //    }
    //}

    backendInfo->oglVersionString = sVersion;
    backendInfo->rendererString = sRenderer;
    backendInfo->vendorString = sVendor;
    backendInfo->backendDetailsString = "Qt backed, " + sRenderableType + ", " + sProfile + ", ceeqt_5 (Qt ver: " + QT_VERSION_STR + ")";

    return true;
}
    
//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glDepthFunc(cee::vis::GLenum func)
{
    m_qtFuncs->glDepthFunc(func);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glDepthMask(cee::vis::GLboolean flag)
{
    m_qtFuncs->glDepthMask(flag);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glEnable(cee::vis::GLenum cap)
{
    m_qtFuncs->glEnable(cap);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glDisable(cee::vis::GLenum cap)
{
    m_qtFuncs->glDisable(cap);
}
    
//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glBindBuffer(cee::vis::GLenum target, cee::vis::GLuint buffer)
{
    m_qtFuncs->glBindBuffer(target, buffer);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glViewport(cee::vis::GLint x, cee::vis::GLint y, cee::vis::GLsizei width, cee::vis::GLsizei height)
{
    m_qtFuncs->glViewport(x, y, width, height);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glClearColor(cee::vis::GLclampf red, cee::vis::GLclampf green, cee::vis::GLclampf blue, cee::vis::GLclampf alpha)
{
    m_qtFuncs->glClearColor(red, green, blue, alpha);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glClearDepth(GLclampd depth)
{
    m_qtFuncs->glClearDepthf(depth);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glStencilMask(cee::vis::GLuint mask)
{
    m_qtFuncs->glStencilMask(mask);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glClearStencil(cee::vis::GLint s)
{
    m_qtFuncs->glClearStencil(s);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glScissor(cee::vis::GLint x, cee::vis::GLint y, cee::vis::GLsizei width, cee::vis::GLsizei height)
{
    m_qtFuncs->glScissor(x, y, width, height);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glClear(cee::vis::GLbitfield mask)
{
    m_qtFuncs->glClear(mask);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glDrawBuffer(cee::vis::GLenum mode)
{
    Q_ASSERT(m_qtFuncs_21);
    if (m_qtFuncs_21)
    {
        m_qtFuncs_21->glDrawBuffer(mode);
    }
    else
    {
        m_logMgr->logDebug(m_loggerName, "OpenGL function not available: glDrawBuffer()");
    }
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glDrawBuffers(cee::vis::GLsizei n, const cee::vis::GLenum* bufs)
{
    Q_ASSERT(m_extraFuncFlags & has_glDrawBuffers);
    if (m_extraFuncFlags & has_glDrawBuffers)
    {
        m_qtExtraFuncs->glDrawBuffers(n, bufs);
    }
    else
    {
        m_logMgr->logDebug(m_loggerName, "OpenGL function not available: glDrawBuffers()");
    }
}
    
//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glBlendFunc(cee::vis::GLenum sfactor, cee::vis::GLenum dfactor)
{
    m_qtFuncs->glBlendFunc(sfactor, dfactor);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glBlendFuncSeparate(cee::vis::GLenum srcRGB, cee::vis::GLenum dstRGB, cee::vis::GLenum srcAlpha, cee::vis::GLenum dstAlpha)
{
    m_qtFuncs->glBlendFuncSeparate(srcRGB, dstRGB, srcAlpha, dstAlpha);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glBlendEquation(cee::vis::GLenum mode)
{
    m_qtFuncs->glBlendEquation(mode);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glBlendEquationSeparate(cee::vis::GLenum modeRGB, cee::vis::GLenum modeAlpha)
{
    m_qtFuncs->glBlendEquationSeparate(modeRGB, modeAlpha);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glBlendColor(cee::vis::GLfloat red, cee::vis::GLfloat green, cee::vis::GLfloat blue, cee::vis::GLfloat alpha)
{
    m_qtFuncs->glBlendColor(red, green, blue, alpha);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glColorMask(cee::vis::GLboolean red, cee::vis::GLboolean green, cee::vis::GLboolean blue, cee::vis::GLboolean alpha)
{
    m_qtFuncs->glColorMask(red, green, blue, alpha);
}
   
//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glCullFace(cee::vis::GLenum mode)
{
    m_qtFuncs->glCullFace(mode);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glFrontFace(cee::vis::GLenum mode)
{
    m_qtFuncs->glFrontFace(mode);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glLineWidth(cee::vis::GLfloat width)
{
    m_qtFuncs->glLineWidth(width);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glPointSize(cee::vis::GLfloat size)
{
    Q_ASSERT(m_qtFuncs_21);
    if (m_qtFuncs_21)
    {
        m_qtFuncs_21->glPointSize(size);
    }
    else
    {
        m_logMgr->logDebug(m_loggerName, "OpenGL function not available: glPointSize()");
    }
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glPointParameteri(cee::vis::GLenum pname, cee::vis::GLint param)
{
    Q_ASSERT(m_qtFuncs_21);
    if (m_qtFuncs_21)
    {
        m_qtFuncs_21->glPointParameteri(pname, param);
    }
    else
    {
        m_logMgr->logDebug(m_loggerName, "OpenGL function not available: glPointParameteri()");
    }
}
    
//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glPolygonMode(cee::vis::GLenum face, cee::vis::GLenum mode)
{
    Q_ASSERT(m_qtFuncs_21);
    if (m_qtFuncs_21)
    {
        m_qtFuncs_21->glPolygonMode(face, mode);
    }
    else
    {
        m_logMgr->logDebug(m_loggerName, "OpenGL function not available: glPolygonMode()");
    }
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glPolygonOffset(cee::vis::GLfloat factor, cee::vis::GLfloat units)
{
    m_qtFuncs->glPolygonOffset(factor, units);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glStencilFunc(cee::vis::GLenum func, cee::vis::GLint ref, cee::vis::GLuint mask)
{
    m_qtFuncs->glStencilFunc(func, ref, mask);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glStencilOp(cee::vis::GLenum sfail, cee::vis::GLenum dpfail, cee::vis::GLenum dppass)
{
    m_qtFuncs->glStencilOp(sfail, dpfail, dppass);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glActiveTexture(cee::vis::GLenum texture)
{
    m_qtFuncs->glActiveTexture(texture);
}


//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glTexImage2D(cee::vis::GLenum target, cee::vis::GLint level, cee::vis::GLint internalFormat, cee::vis::GLsizei width, cee::vis::GLsizei height, cee::vis::GLint border, cee::vis::GLenum format, cee::vis::GLenum type, const cee::vis::GLvoid* data)
{
    m_qtFuncs->glTexImage2D(target, level, internalFormat, width, height,border, format, type, data);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glGenerateMipmap(cee::vis::GLenum target)
{
    m_qtFuncs->glGenerateMipmap(target);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glTexParameteri(cee::vis::GLenum target, cee::vis::GLenum pname, cee::vis::GLint param)
{
    m_qtFuncs->glTexParameteri(target, pname, param);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glTexParameterfv(cee::vis::GLenum target, cee::vis::GLenum pname, const cee::vis::GLfloat * params)
{
    m_qtFuncs->glTexParameterfv(target, pname, params);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glTexImage2DMultisample(cee::vis::GLenum target, cee::vis::GLsizei samples, cee::vis::GLint internalformat, cee::vis::GLsizei width, cee::vis::GLsizei height, cee::vis::GLboolean fixedsamplelocations)
{
    Q_ASSERT(m_qtFuncs_32);
    if (m_qtFuncs_32)
    {
        m_qtFuncs_32->glTexImage2DMultisample(target, samples, internalformat, width, height, fixedsamplelocations);
    }
    else
    {
        m_logMgr->logDebug(m_loggerName, "OpenGL function not available: glTexImage2DMultisample()");
    }
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glBindTexture(cee::vis::GLenum target, cee::vis::GLuint texture)
{
    m_qtFuncs->glBindTexture(target, texture);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glPixelStorei(cee::vis::GLenum pname, cee::vis::GLint param)
{
    m_qtFuncs->glPixelStorei(pname, param);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glVertexAttribPointer(cee::vis::GLuint index, cee::vis::GLint size, cee::vis::GLenum type, cee::vis::GLboolean normalized, cee::vis::GLsizei stride, const cee::vis::GLvoid* pointer)
{
    m_qtFuncs->glVertexAttribPointer(index, size, type, normalized, stride, pointer);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glEnableVertexAttribArray(cee::vis::GLuint index)
{
    m_qtFuncs->glEnableVertexAttribArray(index);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glDisableVertexAttribArray(cee::vis::GLuint index)
{
    m_qtFuncs->glDisableVertexAttribArray(index);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glDrawArrays(cee::vis::GLenum mode, cee::vis::GLint first, cee::vis::GLsizei count)
{
    m_qtFuncs->glDrawArrays(mode, first, count);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glDrawElements(cee::vis::GLenum mode, cee::vis::GLsizei count, cee::vis::GLenum type, const cee::vis::GLvoid* indices)
{
    m_qtFuncs->glDrawElements(mode, count, type, indices);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glDrawRangeElements(cee::vis::GLenum mode, cee::vis::GLuint start, cee::vis::GLuint end, cee::vis::GLsizei count, cee::vis::GLenum type, const cee::vis::GLvoid* indices)
{
    if (m_extraFuncFlags & has_glDrawRangeElements)
    {
        m_qtExtraFuncs->glDrawRangeElements(mode, start, end, count, type, indices);
    }
    else
    {
        m_qtFuncs->glDrawElements(mode, count, type, indices);
    }
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glUniform1f(cee::vis::GLint location, cee::vis::GLfloat v0)
{
    m_qtFuncs->glUniform1f(location, v0);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glUniform1i(cee::vis::GLint location, cee::vis::GLint v0)
{
    m_qtFuncs->glUniform1i(location, v0);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glUniform1fv(cee::vis::GLint location, cee::vis::GLsizei count, const cee::vis::GLfloat* value)
{
    m_qtFuncs->glUniform1fv(location, count, value);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glUniform1iv(cee::vis::GLint location, cee::vis::GLsizei count, const cee::vis::GLint* value)
{
    m_qtFuncs->glUniform1iv(location, count, value);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glUniform2fv(cee::vis::GLint location, cee::vis::GLsizei count, const cee::vis::GLfloat* value)
{
    m_qtFuncs->glUniform2fv(location, count, value);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glUniform3fv(cee::vis::GLint location, cee::vis::GLsizei count, const cee::vis::GLfloat* value)
{
    m_qtFuncs->glUniform3fv(location, count, value);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glUniform4fv(cee::vis::GLint location, cee::vis::GLsizei count, const cee::vis::GLfloat* value)
{
    m_qtFuncs->glUniform4fv(location, count, value);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glUniformMatrix3fv(cee::vis::GLint location, cee::vis::GLsizei count, cee::vis::GLboolean transpose, const cee::vis::GLfloat* value)
{
    m_qtFuncs->glUniformMatrix3fv(location, count, transpose, value);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glUniformMatrix4fv(cee::vis::GLint location, cee::vis::GLsizei count, cee::vis::GLboolean transpose, const cee::vis::GLfloat* value)
{
    m_qtFuncs->glUniformMatrix4fv(location, count, transpose, value);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glGenBuffers(cee::vis::GLsizei n, cee::vis::GLuint* buffers)
{
    m_qtFuncs->glGenBuffers(n, buffers);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glDeleteBuffers(cee::vis::GLsizei n, const cee::vis::GLuint* buffers)
{
    m_qtFuncs->glDeleteBuffers(n, buffers);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glBufferData(cee::vis::GLenum target, cee::vis::GLsizeiptr size, const cee::vis::GLvoid* data, cee::vis::GLenum usage)
{
    m_qtFuncs->glBufferData(target, size, data, usage);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
cee::vis::GLuint OpenGLFunctionsBackend_qt5::glCreateShader(cee::vis::GLenum shaderType)
{
    return m_qtFuncs->glCreateShader(shaderType);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glDeleteShader(cee::vis::GLuint shader)
{
    m_qtFuncs->glDeleteShader(shader);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glShaderSource(cee::vis::GLuint shader, cee::vis::GLsizei count, const GLchar** string, const cee::vis::GLint* length)
{
    m_qtFuncs->glShaderSource(shader, count, string, length);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glCompileShader(cee::vis::GLuint shader)
{
    m_qtFuncs->glCompileShader(shader);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
cee::vis::GLboolean OpenGLFunctionsBackend_qt5::glIsShader(cee::vis::GLuint shader)
{
    return m_qtFuncs->glIsShader(shader);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glGetShaderInfoLog(cee::vis::GLuint shader, cee::vis::GLsizei maxLength, cee::vis::GLsizei* length, GLchar* infoLog)
{
    m_qtFuncs->glGetShaderInfoLog(shader, maxLength, length, infoLog);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glGetShaderiv(cee::vis::GLuint shader, cee::vis::GLenum pname, cee::vis::GLint* params)
{
    m_qtFuncs->glGetShaderiv(shader, pname, params);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
cee::vis::GLuint OpenGLFunctionsBackend_qt5::glCreateProgram()
{
    return m_qtFuncs->glCreateProgram();
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glDeleteProgram(cee::vis::GLuint program)
{
    m_qtFuncs->glDeleteProgram(program);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glAttachShader(cee::vis::GLuint program, cee::vis::GLuint shader)
{
    m_qtFuncs->glAttachShader(program, shader);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glLinkProgram(cee::vis::GLuint program)
{
    m_qtFuncs->glLinkProgram(program);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glValidateProgram(cee::vis::GLuint program)
{
    m_qtFuncs->glValidateProgram(program);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
cee::vis::GLboolean OpenGLFunctionsBackend_qt5::glIsProgram(cee::vis::GLuint program)
{
    return m_qtFuncs->glIsProgram(program);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glUseProgram(cee::vis::GLuint program)
{
    m_qtFuncs->glUseProgram(program);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glGetProgramInfoLog(cee::vis::GLuint program, cee::vis::GLsizei maxLength, cee::vis::GLsizei* length, GLchar* infoLog)
{
    m_qtFuncs->glGetProgramInfoLog(program, maxLength, length, infoLog);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glGetProgramiv(cee::vis::GLuint program, cee::vis::GLenum pname, cee::vis::GLint* params)
{
    m_qtFuncs->glGetProgramiv(program, pname, params);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
cee::vis::GLint OpenGLFunctionsBackend_qt5::glGetUniformLocation(cee::vis::GLuint program, const GLchar* name)
{
    return m_qtFuncs->glGetUniformLocation(program, name);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glGetActiveUniform(cee::vis::GLuint program, cee::vis::GLuint index, cee::vis::GLsizei bufSize, cee::vis::GLsizei* length, cee::vis::GLint* size, cee::vis::GLenum* type, GLchar* name)
{
    m_qtFuncs->glGetActiveUniform(program, index, bufSize, length, size, type, name);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
cee::vis::GLint OpenGLFunctionsBackend_qt5::glGetAttribLocation(cee::vis::GLuint program, const GLchar* name)
{
    return m_qtFuncs->glGetAttribLocation(program, name);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glBindAttribLocation(cee::vis::GLuint program, cee::vis::GLuint index, const GLchar* name)
{
    m_qtFuncs->glBindAttribLocation(program, index, name);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glGenTextures(cee::vis::GLsizei n, cee::vis::GLuint* textures)
{
    m_qtFuncs->glGenTextures(n, textures);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glDeleteTextures(cee::vis::GLsizei n, const cee::vis::GLuint* textures)
{
    m_qtFuncs->glDeleteTextures(n, textures);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
cee::vis::GLenum OpenGLFunctionsBackend_qt5::glGetError()
{
    return m_qtFuncs->glGetError();
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
cee::vis::GLboolean OpenGLFunctionsBackend_qt5::glIsEnabled(cee::vis::GLenum cap)
{
    return m_qtFuncs->glIsEnabled(cap);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glGetIntegerv(cee::vis::GLenum pname, cee::vis::GLint* params)
{
    m_qtFuncs->glGetIntegerv(pname, params);
}
    
//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glGenRenderbuffers(cee::vis::GLsizei n, cee::vis::GLuint* renderbuffers)
{
    m_qtFuncs->glGenRenderbuffers(n, renderbuffers);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glDeleteRenderbuffers(cee::vis::GLsizei n, cee::vis::GLuint* renderbuffers)
{
    m_qtFuncs->glDeleteRenderbuffers(n, renderbuffers);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glBindRenderbuffer(cee::vis::GLenum target, cee::vis::GLuint renderbuffer)
{
    m_qtFuncs->glBindRenderbuffer(target, renderbuffer);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glRenderbufferStorage(cee::vis::GLenum target, cee::vis::GLenum internalformat, cee::vis::GLsizei width, cee::vis::GLsizei height)
{
    m_qtFuncs->glRenderbufferStorage(target, internalformat, width, height);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glRenderbufferStorageMultisample(cee::vis::GLenum target, cee::vis::GLsizei samples, cee::vis::GLenum internalformat, cee::vis::GLsizei width, cee::vis::GLsizei height)
{
    Q_ASSERT(m_extraFuncFlags & has_glRenderbufferStorageMultisample);
    if (m_extraFuncFlags & has_glRenderbufferStorageMultisample)
    {
        m_qtExtraFuncs->glRenderbufferStorageMultisample(target, samples, internalformat, width, height);
    }
    else
    {
        m_logMgr->logDebug(m_loggerName, "OpenGL function not available: glRenderbufferStorageMultisample()");
    }
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glGenFramebuffers(cee::vis::GLsizei n, cee::vis::GLuint* ids)
{
    m_qtFuncs->glGenFramebuffers(n, ids);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glDeleteFramebuffers(cee::vis::GLsizei n, cee::vis::GLuint* framebuffers)
{
    m_qtFuncs->glDeleteFramebuffers(n, framebuffers);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glFramebufferRenderbuffer(cee::vis::GLenum target, cee::vis::GLenum attachment, cee::vis::GLenum renderbuffertarget, cee::vis::GLuint renderbuffer)
{
    m_qtFuncs->glFramebufferRenderbuffer(target, attachment, renderbuffertarget, renderbuffer);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glFramebufferTexture2D(cee::vis::GLenum target, cee::vis::GLenum attachment, cee::vis::GLenum textarget, cee::vis::GLuint texture, cee::vis::GLint level)
{
    m_qtFuncs->glFramebufferTexture2D(target, attachment, textarget, texture, level);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glBindFramebuffer(cee::vis::GLenum target, cee::vis::GLuint framebuffer)
{
    m_qtFuncs->glBindFramebuffer(target, framebuffer);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
cee::vis::GLenum OpenGLFunctionsBackend_qt5::glCheckFramebufferStatus(cee::vis::GLenum target)
{
    return m_qtFuncs->glCheckFramebufferStatus(target);
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glBlitFramebuffer(cee::vis::GLint srcX0, cee::vis::GLint srcY0, cee::vis::GLint srcX1, cee::vis::GLint srcY1, cee::vis::GLint dstX0, cee::vis::GLint dstY0, cee::vis::GLint dstX1, cee::vis::GLint dstY1, cee::vis::GLbitfield mask, cee::vis::GLenum filter)
{
    Q_ASSERT(m_extraFuncFlags & has_glBlitFramebuffer);
    if (m_extraFuncFlags & has_glBlitFramebuffer)
    {
        m_qtExtraFuncs->glBlitFramebuffer(srcX0, srcY0, srcX1, srcY1, dstX0, dstY0, dstX1, dstY1, mask, filter);
    }
    else
    {
        m_logMgr->logDebug(m_loggerName, "OpenGL function not available: glBlitFramebuffer()");
    }
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glVertexAttribIPointer(cee::vis::GLuint index, cee::vis::GLint size, cee::vis::GLenum type, cee::vis::GLsizei stride, const cee::vis::GLvoid* pointer)
{
    Q_ASSERT(m_extraFuncFlags & has_glVertexAttribIPointer);
    if (m_extraFuncFlags & has_glVertexAttribIPointer)
    {
        m_qtExtraFuncs->glVertexAttribIPointer(index, size, type, stride, pointer);
    }
    else
    {
        m_logMgr->logDebug(m_loggerName, "OpenGL function not available: glVertexAttribIPointer()");
    }
}

//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void OpenGLFunctionsBackend_qt5::glReadPixels(cee::vis::GLint x, cee::vis::GLint y, cee::vis::GLsizei width, cee::vis::GLsizei height, cee::vis::GLenum format, cee::vis::GLenum type, cee::vis::GLvoid* data)
{
    m_qtFuncs->glReadPixels(x, y, width, height, format, type, data);
}



}
}


