//##################################################################################################
//
//  Ceetron Desktop Components
//  Platform integration for Qt
//  
//  --------------------------------------------------------------------------------------------
//  Copyright (C) 2015, Ceetron AS
//  This is UNPUBLISHED PROPRIETARY SOURCE CODE of Ceetron AS. The contents of this file may 
//  not be disclosed to third parties, copied or duplicated in any form, in whole or in part,
//  without the prior written permission of Ceetron AS.
//##################################################################################################

#pragma once

#include "CeeVisualization/MouseEvent.h"
#include "CeeVisualization/WheelEvent.h"
#include "CeeCore/Image.h"

#include <QWidget>
#include <QMouseEvent>

#if QT_VERSION >= 0x050000
#include <QWindow>
#include <QOpenGLContext>
#endif

namespace cee {
namespace qt {



//==================================================================================================
//
// Static helper class for interop between Ceetron Desktop Components and Qt 
//
//==================================================================================================
class UtilsGui
{
public:
    
    //--------------------------------------------------------------------------------------------------
    /// 
    //--------------------------------------------------------------------------------------------------
    static QImage toQImage(const cee::Image& image)
    {
        int width = static_cast<int>(image.width());
        int height = static_cast<int>(image.height());
        if (width <= 0 || height <= 0)
        {
            return QImage();
        }

        const unsigned char* rgbData = image.rawPointer();
        QImage qimg(width, height, QImage::Format_RGB888);
        for (int y = 0; y < height; y++)
        {
            for (int x = 0; x < width; x++)
            {
                int srcIdx = 4*y*width + 4*x;
                qimg.setPixel(x, height - y - 1, qRgb(rgbData[srcIdx], rgbData[srcIdx + 1], rgbData[srcIdx + 2]));
            }
        }

        return qimg;
    }

    //--------------------------------------------------------------------------------------------------
    /// 
    //--------------------------------------------------------------------------------------------------
    static void toImage(const QImage& qImage, cee::Image* image)
    {
        Q_ASSERT(image);

        int width = qImage.width();
        int height = qImage.height();
        if (width <= 0 || height <= 0)
        {
            return;
        }

        unsigned char* rgbData = new unsigned char[4*width*height];
        for (int y = 0; y < height; y++)
        {
            for (int x = 0; x < width; x++)
            {
                int dstIdx = 4*y*width + 4*x;
                QRgb pixel = qImage.pixel(x, height - y - 1);
                rgbData[dstIdx] = qRed(pixel);
                rgbData[dstIdx+1] = qGreen(pixel);
                rgbData[dstIdx+2] = qBlue(pixel);
                rgbData[dstIdx+3] = qAlpha(pixel);
            }
        }

        image->setFromRgba(rgbData, width, height);

        delete[] rgbData;
    }

    //--------------------------------------------------------------------------------------------------
    /// 
    //--------------------------------------------------------------------------------------------------
    static QColor toQColor(const cee::Color3f& color)
    {
        QColor qtColor;
        qtColor.setRedF(color.r());
        qtColor.setGreenF(color.g());
        qtColor.setBlueF(color.b());

        return qtColor;
    }

    //--------------------------------------------------------------------------------------------------
    /// 
    //--------------------------------------------------------------------------------------------------
    static cee::Color3f toColor3f(const QColor& color)
    {
        return cee::Color3f(color.redF(), color.greenF(), color.blueF());
    }

    //--------------------------------------------------------------------------------------------------
    /// Translates mouse event from a QMouseEvent to a cee::vis::MouseEvent
    //--------------------------------------------------------------------------------------------------
    static cee::vis::MouseEvent toMouseEvent(const QWidget& widget, const QMouseEvent& event)
    {
        int xPos = event.x();
        int yPos = widget.height() - event.y();
        
        const double devicePixelRatio = determineWidgetDevicePixelRatio(widget);
        xPos = static_cast<int>(xPos*devicePixelRatio);
        yPos = static_cast<int>(yPos*devicePixelRatio);

        Qt::MouseButtons qtButtons = event.buttons();
        vis::MouseButtons buttons = vis::NoButton;
        if (qtButtons & Qt::LeftButton)     buttons |= vis::LeftButton;
#if QT_VERSION >= QT_VERSION_CHECK(5,15,4)
        if (qtButtons & Qt::MiddleButton)      buttons |= vis::MiddleButton;
#else
        if (qtButtons & Qt::MidButton)      buttons |= vis::MiddleButton;
#endif

        if (qtButtons & Qt::RightButton)    buttons |= vis::RightButton;

        Qt::KeyboardModifiers qtModifiers = event.modifiers();
        vis::KeyboardModifiers modifiers = vis::NoModifier;
        if (qtModifiers & Qt::ShiftModifier)      modifiers |= vis::ShiftModifier;
        if (qtModifiers & Qt::ControlModifier)    modifiers |= vis::ControlModifier;

        return cee::vis::MouseEvent(xPos, yPos, buttons, modifiers);

    }
    
    //--------------------------------------------------------------------------------------------------
    /// Translates mouse button from a Qt::MouseButton to a vis::MouseButton
    //--------------------------------------------------------------------------------------------------
    static vis::MouseButton toMouseButton(const Qt::MouseButton& button)
    {
        if (button == Qt::LeftButton)    return vis::LeftButton;
#if QT_VERSION >= QT_VERSION_CHECK(5,15,4)
        if (button & Qt::MiddleButton)      return vis::MiddleButton;
#else
        if (button & Qt::MidButton)      return vis::MiddleButton;
#endif 
        if (button == Qt::RightButton)   return vis::RightButton;

        return vis::NoButton;
    }
    
    
    //--------------------------------------------------------------------------------------------------
    /// Translates mouse wheel event from a QWheelEvent to a cee::vis::WheelEvent
    //--------------------------------------------------------------------------------------------------
    static vis::WheelEvent toWheelEvent(const QWidget& widget, const QWheelEvent& event)
    {
#if QT_VERSION >= QT_VERSION_CHECK(5,15,4)
        int xPos = event.position().x();
        int yPos = widget.height() - event.position().y();
#else
        int xPos = event.x();
        int yPos = widget.height() - event.y();
#endif
        
        const double devicePixelRatio = determineWidgetDevicePixelRatio(widget);
        xPos = static_cast<int>(xPos*devicePixelRatio);
        yPos = static_cast<int>(yPos*devicePixelRatio);

        Qt::MouseButtons qtButtons = event.buttons();
        vis::MouseButtons buttons = vis::NoButton;
        if (qtButtons & Qt::LeftButton)     buttons |= vis::LeftButton;
#if QT_VERSION >= QT_VERSION_CHECK(5,15,4)
        if (qtButtons & Qt::MiddleButton)      buttons |= vis::MiddleButton;
#else
        if (qtButtons & Qt::MidButton)      buttons |= vis::MiddleButton;
#endif 
        if (qtButtons & Qt::RightButton)    buttons |= vis::RightButton;

        Qt::KeyboardModifiers qtModifiers = event.modifiers();
        vis::KeyboardModifiers modifiers = vis::NoModifier;
        if (qtModifiers & Qt::ShiftModifier)      modifiers |= vis::ShiftModifier;
        if (qtModifiers & Qt::ControlModifier)    modifiers |= vis::ControlModifier;

#if QT_VERSION >= QT_VERSION_CHECK(5,15,4)
        return vis::WheelEvent(event.angleDelta().y(), xPos, yPos, buttons, modifiers);
#else
        return vis::WheelEvent(event.delta(), xPos, yPos, buttons, modifiers);
#endif
    }


    //--------------------------------------------------------------------------------------------------
    /// Translates mouse wheel event from a QWheelEvent to a cee::vis::WheelEvent
    //--------------------------------------------------------------------------------------------------
    static double determineWidgetDevicePixelRatio(const QWidget& widget)
    {
        double dpr = 1;

#if QT_VERSION >= 0x050600

        dpr = widget.devicePixelRatioF();

#elif QT_VERSION >= 0x050000

        // We can get a floating point value from the native window if it exists
        if (widget.windowHandle())
        {
            dpr = widget.windowHandle()->devicePixelRatio();
        }
        else
        {
            // use the widget's integer version
            dpr = widget.devicePixelRatio();
        }

#else
        // Just touch it to avoid unused warnings
        (void)(widget);

#endif

        return dpr;
    }


    //--------------------------------------------------------------------------------------------------
    /// Determine if Qt's currently loaded OpenGL implementation is OpenGLES
    /// 
    /// Note that this function requires that the QGuiApplication instance is already created.
    //--------------------------------------------------------------------------------------------------
    static bool isCurrentQtOpenGLImplementationES()
    {
        bool moduleTypeIsOpenGLES = false;

    #if QT_VERSION >= 0x050300
        // Note that the query for module type must be after QApplication has been initialized
        if (QOpenGLContext::openGLModuleType() == QOpenGLContext::LibGLES) moduleTypeIsOpenGLES = true;
    #endif

        return moduleTypeIsOpenGLES;
    }


    //--------------------------------------------------------------------------------------------------
    /// Try and guess if Qt's currently loaded OpenGL implementation is Mesa software only.
    ///
    /// The function works by interrogating the file name of the currently loaded OpenGL implementation
    /// and will return true if the file name is opengl32sw.dll.
    /// 
    /// Note that this function requires that the QGuiApplication instance is already created.
    /// Note that this function only works on Windows.
    //--------------------------------------------------------------------------------------------------
    static bool isCurrentQtOpenGLImplementationMesaWin()
    {
        QString sOpenGLModuleFileName = getCurrentQtOpenGLModuleFileNameWin();
    
        if (sOpenGLModuleFileName.contains("opengl32sw.dll", Qt::CaseInsensitive))
        {
            return true;
        }
        else
        {
            return false;
        }
    }


    //--------------------------------------------------------------------------------------------------
    /// Get the module file name of Qt's currently loaded OpenGL implementation
    ///
    /// Note that this function requires that the QGuiApplication instance is already created.
    /// Note that this function only works on Windows.
    //--------------------------------------------------------------------------------------------------
    static QString getCurrentQtOpenGLModuleFileNameWin()
    {
        QString sOpenGLModuleFileName;

    #if defined(WIN32) && QT_VERSION >= 0x050300
        #if QT_VERSION >= 0x060000
            HMODULE hOpenGLModule = QNativeInterface::QWGLContext::openGLModuleHandle();
        #elif QT_VERSION >= 0x050300
            HMODULE hOpenGLModule = static_cast<HMODULE>(QOpenGLContext::openGLModuleHandle());
        #endif
            if (hOpenGLModule)
            {
                char szFileNameBuffer[MAX_PATH];
                if (::GetModuleFileNameA(hOpenGLModule, szFileNameBuffer, sizeof(szFileNameBuffer)))
                {
                    sOpenGLModuleFileName = szFileNameBuffer;
                }
            }
    #endif

        return sOpenGLModuleFileName;
    }
};


} // namespace qt
} // namespace cee

