//##################################################################################################
//
//  Ceetron Desktop Components
//  Platform integration for WxWidgets
//  
//  --------------------------------------------------------------------------------------------
//  Copyright (C) 2014, Ceetron AS
//  This is UNPUBLISHED PROPRIETARY SOURCE CODE of Ceetron AS. The contents of this file may 
//  not be disclosed to third parties, copied or duplicated in any form, in whole or in part,
//  without the prior written permission of Ceetron AS.
//##################################################################################################

#include "ceewxViewerCanvas.h"

#include "wx/wx.h"

#include "CeeVisualization/OpenGLContextGroup.h"
#include "CeeVisualization/Camera.h"

namespace cee {
namespace wx {




//==================================================================================================
//
// 
//
//==================================================================================================

BEGIN_EVENT_TABLE(ViewerCanvas, wxGLCanvas)
    EVT_PAINT(ViewerCanvas::onPaint)
END_EVENT_TABLE()

//--------------------------------------------------------------------------------------------------
/// Creates a viewer canvas. 
/// Use this constructor for the first or only ViewerCanvas in your application.
//--------------------------------------------------------------------------------------------------
ViewerCanvas::ViewerCanvas(cee::vis::OpenGLContextGroup* contextGroup, wxWindow* parent, int* attribList)
:   wxGLCanvas(parent, wxID_ANY, attribList, wxDefaultPosition, wxDefaultSize, wxFULL_REPAINT_ON_RESIZE, L"ViewerCanvas", wxNullPalette),
    cee::vis::ViewerUIFramework(contextGroup),
    m_viewerIsInitialized(false),
    m_viewerHasBeenShutdown(false),
    m_wxGLContext(new wxGLContext(this, NULL))
{
    // Constructor should only be used for the first viewer in the context group
    wxASSERT(contextGroup);
    wxASSERT(contextGroup->contextCount() == 1);
}


//--------------------------------------------------------------------------------------------------
/// Creates a viewer sharing the OpenGL resources with the specified viewer canvas.
/// 
/// By specifying the shareCanvas, all OpenGL resources can be shared between the viewers and thus 
/// the same Models can be shown in multiple views at the same time.
//--------------------------------------------------------------------------------------------------
ViewerCanvas::ViewerCanvas(cee::vis::OpenGLContextGroup* contextGroup, wxWindow* parent, int* attribList, ViewerCanvas* shareCanvas)
:   wxGLCanvas(parent, wxID_ANY, attribList, wxDefaultPosition, wxDefaultSize, wxFULL_REPAINT_ON_RESIZE, L"ViewerCanvas", wxNullPalette),
    cee::vis::ViewerUIFramework(contextGroup),
    m_viewerIsInitialized(false),
    m_viewerHasBeenShutdown(false),
    m_wxGLContext(new wxGLContext(this, shareCanvas->m_wxGLContext))
{
    wxASSERT(shareCanvas);

    // Constructor should only be used for the first viewer in the context group
    wxASSERT(contextGroup);
    wxASSERT(contextGroup->contextCount() > 1);
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
ViewerCanvas::~ViewerCanvas()
{
    // By the time we get here, shutdownViewer() should already be called and as a consequence
    // the view should already be detached and the viewer should either be in the uninitialized or shutdown state
    wxASSERT(!m_viewerIsInitialized || m_viewerHasBeenShutdown);

    delete m_wxGLContext;
    m_wxGLContext = NULL;
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
bool ViewerCanvas::initializeViewer()
{
    if (m_viewerIsInitialized)
    {
        return true;
    }

    // Never resurrect a viewer that has been shut down
    if (m_viewerHasBeenShutdown)
    {
        return false;
    }

    if (!tryMakeGLContextCurrentForCanvas())
    {
        return false;
    }

    if (!initializeComponentInternalOpenGL())
    {
        return false;
    }

    m_viewerIsInitialized = true;
    return true;
}



//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void ViewerCanvas::shutdownViewer()
{
    // Do nothing if we're already shut down
    if (m_viewerHasBeenShutdown)
    {
        return;
    }

    setView(NULL);

    if (m_viewerIsInitialized)
    {
        bool makeCurrentSucceeded = tryMakeGLContextCurrentForCanvas();
        wxASSERT(makeCurrentSucceeded);
        if (makeCurrentSucceeded)
        {
            shutdownComponentInternalOpenGL();
        }
    }

    m_viewerHasBeenShutdown = true;
}


//--------------------------------------------------------------------------------------------------
/// Set the view to use in this viewer
/// 
/// \warn A view cannot be associated with more than one viewer. It is illegal to set the same 
//        view in multiple viewers at the same time. 
//--------------------------------------------------------------------------------------------------
void ViewerCanvas::setView(cee::vis::View* aView)
{
    if (m_view)
    {
        m_view->detachFromViewer();
    }

    if (aView)
    {
        // It is illegal to set the same view in multiple viewers
        wxASSERT(!aView->attachedViewer());

        m_view = aView;
        m_view->attachToViewer(this);

        int width, height;
        GetClientSize(&width, &height);
        m_view->camera().setViewport(0, 0, width, height);
    }
    else
    {
        m_view = NULL;
    }
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
cee::vis::View* ViewerCanvas::view()
{
    return m_view.get();

}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void ViewerCanvas::doRequestRedraw()
{
    Refresh();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void ViewerCanvas::doRequestImmediateRedraw()
{
    Refresh();
	Update();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
bool ViewerCanvas::doMakeCurrent()
{
    return tryMakeGLContextCurrentForCanvas();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
bool ViewerCanvas::tryMakeGLContextCurrentForCanvas()
{
    // On X11, according to comments inside wxWidgets, we cannot make the OpenGL context current before the windows has been shown.
    // This includes trying to make it current from the canvas' constructor.
    
    // Is this the best call to use in order to classify the window as being 'shown'? One other alternative would be IsShown()
    if (!IsShownOnScreen())
    {
        return false;
    }

    wxASSERT(m_wxGLContext);

#if (wxMAJOR_VERSION >= 3)
    if (!SetCurrent(*m_wxGLContext))
    {
        return false;
    }
#else
    // No return value for older versions
    SetCurrent(*m_wxGLContext);
#endif

    return true;
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
bool ViewerCanvas::doIsCurrent()
{
#ifdef WIN32
    if (m_wxGLContext)
    {
        if (wglGetCurrentContext() == m_wxGLContext->GetGLRC())
        {
            return true;
        }
    }
    
    return false;

#else

    // Currently unable to determine current state on Linux.
    // For now only thing to do is always return true
    return true;

#endif
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
unsigned int ViewerCanvas::doGetDefaultFramebufferObject()
{
    return 0;
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void ViewerCanvas::onPaint(wxPaintEvent& /*event*/)
{
    // See comment in initializeViewer()
    if (!IsShownOnScreen())
    {
        return;
    }

    // Do lazy initialization of the viewer since on Linux/X11, the
    // canvas needs to be shown before the OpenGL context can be made current
    bool canDoComponentDrawing = true;
    if (!m_viewerIsInitialized)
    {
        if (!initializeViewer())
        {
            canDoComponentDrawing = false;
        }
    }

    wxPaintDC dc(this);

    if (m_view && canDoComponentDrawing)
    {
        const wxSize clientSize = GetClientSize();
        m_view->camera().setViewport(0, 0, clientSize.x, clientSize.y);

        // The OpenGL context will be made current as part of the render call-sequence
        m_view->renderViewInAttachedViewer();

        glFinish();
        SwapBuffers();
    }
    else
    {
        dc.SetBackground(*wxGREY_BRUSH);
        dc.Clear();
    }
}


} // namespace wx
} // namespace gc

