//##################################################################################################
//
//  Ceetron Desktop Components
//  WxTutorialRunner
//  
//  --------------------------------------------------------------------------------------------
//  Copyright (C) 2014, Ceetron AS
//  This is UNPUBLISHED PROPRIETARY SOURCE CODE of Ceetron AS. The contents of this file may 
//  not be disclosed to third parties, copied or duplicated in any form, in whole or in part,
//  without the prior written permission of Ceetron AS.
//##################################################################################################

#include "TutorialRunnerMainWindow.h"
#include "TutorialRunnerViewerCanvas.h"

#include "wx/wx.h"

#include "CeeVisualization/VisualizationComponent.h"
#include "CeeVisualization/View.h"
#include "CeeVisualization/Camera.h"

#include "CeeUnstructGrid/DataPart.h"
#include "CeeUnstructGrid/DataState.h"




//==================================================================================================
//
// 
//
//==================================================================================================

enum 
{
    ID_UnstructGridModel,
    ID_UnstructGridModelResults,
    ID_LoadVTFx,
    ID_ScalarSettings,
    ID_VectorSettings,
    ID_PartSettings,
    ID_CuttingPlane,
    ID_Isosurface,
    ID_Isovolume,
    ID_Highlighting,
    ID_MarkupModel,
    ID_GeometryModel,
    ID_GeometryModelTexture,
    ID_GeometryModelPrimitives,
    ID_CustomReader,
    ID_CustomNavigation,
    ID_ElementSets,
    ID_Report,
	ID_Plot,

    ID_DummyLastID
};

BEGIN_EVENT_TABLE(TutorialRunnerMainWindow, wxFrame)
    EVT_SHOW(TutorialRunnerMainWindow::onShow)
    EVT_CLOSE(TutorialRunnerMainWindow::onClose)

    EVT_MENU_RANGE(ID_UnstructGridModel, ID_DummyLastID, TutorialRunnerMainWindow::onTutorialMenuItemSelected)
    EVT_MENU(wxID_EXIT, TutorialRunnerMainWindow::onExit)
END_EVENT_TABLE()

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
TutorialRunnerMainWindow::TutorialRunnerMainWindow()
:   wxFrame(NULL, wxID_ANY, wxT("WxTutorialRunner")),
    m_contextGroup(cee::vis::VisualizationComponent::createOpenGLContextGroup())
{
    wxMenu* menu = new wxMenu;
    menu->Append(ID_UnstructGridModel,          L"UnstructGridModel: Create two triangles");
    menu->Append(ID_UnstructGridModelResults,   L"UnstructGridModel: Create triangle with results");
    menu->Append(ID_LoadVTFx,                   L"Load vtfx file");
    menu->Append(ID_ScalarSettings,             L"Scalar settings");
    menu->Append(ID_VectorSettings,             L"Vector settings");
    menu->Append(ID_PartSettings,               L"Part settings");
    menu->Append(ID_CuttingPlane,               L"Cutting plane");
    menu->Append(ID_Isosurface,                 L"Isosurface");
    menu->Append(ID_Isovolume,                  L"Isovolume");
    menu->Append(ID_Highlighting,               L"Highlighting");
    menu->Append(ID_MarkupModel,                L"MarkupModel");
    menu->Append(ID_GeometryModel,              L"GeometryModel");
    menu->Append(ID_GeometryModelTexture,       L"GeometryModel: Texture");
    menu->Append(ID_GeometryModelPrimitives,    L"GeometryModel: Primitives");
    menu->Append(ID_CustomReader,               L"Custom reader");
    menu->Append(ID_CustomNavigation,           L"Custom navigation");
    menu->Append(ID_ElementSets,                L"UnstructGridModel: DataElementSets");
    menu->Append(ID_Report,                     L"Report");
    menu->Append(ID_Plot,                     	L"Plot2d");
    menu->AppendSeparator();
    menu->Append(wxID_EXIT);

    wxMenuBar* menuBar = new wxMenuBar;
    menuBar->Append(menu, wxT("Tutorials"));
    SetMenuBar(menuBar);

    CreateStatusBar();

    // According to docs, passing NULL for attributes should work, but for now it seems we need to
    // pass an explicit list in order to get a double buffered context on Linux
    int i = 0;
    int attribList[10];
    attribList[i++] = WX_GL_RGBA;
    attribList[i++] = WX_GL_DOUBLEBUFFER;
    attribList[i++] = WX_GL_DEPTH_SIZE;     attribList[i++] = 1;
    attribList[i++] = 0;

    // Need to check that creation will be OK before actually doing it, 
    // otherwise we get into trouble while constructing the canvas below.
    // In production code this scenario must be handled and/or reported to the user
    wxASSERT(wxGLCanvas::IsDisplaySupported(attribList));

    m_viewer = new TutorialRunnerViewerCanvas(m_contextGroup.get(), this, attribList);
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
TutorialRunnerMainWindow::~TutorialRunnerMainWindow()
{
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void TutorialRunnerMainWindow::onShow(wxShowEvent&)
{
    // Our window should be shown by now, so try and initialize the viewer and add view
    wxASSERT(m_viewer);
    m_viewer->initializeViewer();

    cee::vis::View* gcView = new cee::vis::View;
    m_viewer->setView(gcView);

    loadVtfAndSetResults();
}


//--------------------------------------------------------------------------------------------------
/// Intercepts default handling of the close event to do cleanup of the viewer
//--------------------------------------------------------------------------------------------------
void TutorialRunnerMainWindow::onClose(wxCloseEvent&)
{
    // Shut down the viewer before destroying windows
    m_viewer->setView(NULL);
    m_viewer->shutdownViewer();

    Destroy();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
cee::vis::View* TutorialRunnerMainWindow::getTutorialView()
{
    return m_viewer->view();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void TutorialRunnerMainWindow::onTutorialMenuItemSelected(wxCommandEvent& event)
{
    const int itemId = event.GetId();
    switch (itemId)
    {
        case ID_UnstructGridModel:          createTriangleGeometry();   break;
        case ID_UnstructGridModelResults:   createTriangleGeometryWithAllResults(); break;
        case ID_LoadVTFx:                   loadVtfAndSetResults(); break;
        case ID_ScalarSettings:             setScalarSettings(); break;
        case ID_VectorSettings:             setVectorSettings(); break;
        case ID_PartSettings:               setPartSettings(); break;
        case ID_CuttingPlane:               addCuttingPlane(); break;
        case ID_Isosurface:                 addIsosurface(); break;
        case ID_Isovolume:                  addIsovolume(); break;
        case ID_Highlighting:               highlighting(); break;
        case ID_MarkupModel:                createMarkup(); break;
        case ID_GeometryModel:              createGeometryModel(); break;
        case ID_GeometryModelTexture:       createGeometryModelWithTexture(); break;
        case ID_GeometryModelPrimitives:    createGeometryModelWithPrimitives(); break;
        case ID_CustomReader:               createCustomDataReader(); break;
        case ID_CustomNavigation:           customNavigation(); break;
        case ID_ElementSets:                addDataElementSets(); break;
        case ID_Report:                     createReport(); break;
        case ID_Plot:                     	createPlot(); break;
    }

}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void TutorialRunnerMainWindow::onExit(wxCommandEvent&)
{
    // Just call close and let onClose() event handler do cleanup
    Close(true);
}

