//##################################################################################################
//
//  Ceetron Desktop Components
//  Component: Core
//  
//  --------------------------------------------------------------------------------------------
//  Copyright (C) 2011, Ceetron AS
//  This is UNPUBLISHED PROPRIETARY SOURCE CODE of Ceetron AS. The contents of this file may 
//  not be disclosed to third parties, copied or duplicated in any form, in whole or in part,
//  without the prior written permission of Ceetron AS.
//##################################################################################################

namespace cee {


//==================================================================================================
///
/// \class cee::PtrRef
/// \ingroup Core
///
/// Smart pointer class used for handling reference counted objects (that derive from Object). 
///
/// The PtrRef<T> class encapsulates reference counting by calling Object::addRef() and Object::release() 
/// on the internally stored object pointer. 
///
/// \internal Inspired by boost's intrusive_ptr class.
//==================================================================================================

//--------------------------------------------------------------------------------------------------
/// Constructs an object from naked pointer 
/// 
/// Will call addRef() on the passed object.
//--------------------------------------------------------------------------------------------------
template<typename T>
PtrRef<T>::PtrRef(T* object)
{
    if (object)
    {
        object->addRef();
        m_object = object;
    }
    else
    {
        m_object = NULL;
    }
}

//--------------------------------------------------------------------------------------------------
/// Construct and object as a copy of \a other.
///  
/// Copies the internal pointer from the passed %PtrRef object and calls addRef() on the object.
//--------------------------------------------------------------------------------------------------
template<typename T>
PtrRef<T>::PtrRef(const PtrRef& other)
{
    if (other.m_object)
    {
        other.m_object->addRef();
        m_object = other.m_object;
    }
    else
    {
        m_object = NULL;
    }
}


//--------------------------------------------------------------------------------------------------
/// Constructs from related. 
/// 
/// Copies the internal pointer from the passed %PtrRef object and calls addRef() on the object.
//--------------------------------------------------------------------------------------------------
template<typename T>
template<typename T2>
PtrRef<T>::PtrRef(const PtrRef<T2>& other)
{
    if (other)
    {
        other->addRef();

        // Escape const
        m_object = const_cast<T2*>(other.get());
    }
    else
    {
        m_object = NULL;
    }
}


//--------------------------------------------------------------------------------------------------
/// Destroys the object. 
/// 
/// Will call release() on the internal pointer.
//--------------------------------------------------------------------------------------------------
template<typename T>
PtrRef<T>::~PtrRef()
{
    if (m_object)
    {
        m_object->release();
        m_object = NULL;
    }
}


//--------------------------------------------------------------------------------------------------
/// Assigns from raw pointer. 
//--------------------------------------------------------------------------------------------------
template<typename T>
PtrRef<T>& PtrRef<T>::operator=(T* rhs)
{
    // Copy-and-swap (on temporary obj)
    PtrRef<T>(rhs).swap(*this);
    return *this;
}


//--------------------------------------------------------------------------------------------------
/// Assigns from \a rhs
/// 
/// Copies the internal pointer from \a rhs and calls addRef().
/// If we're already storing an internal pointer, this pointer will be release()'ed. 
//--------------------------------------------------------------------------------------------------
template<typename T>
PtrRef<T>& PtrRef<T>::operator=(PtrRef rhs)
{
    // Copy-and-swap (copy already done since parameter is passed by value)
    rhs.swap(*this);
    return *this;
}


//--------------------------------------------------------------------------------------------------
/// \fn PtrRef<T>::operator=(const PtrRef<T2> &rhs)
/// Assigns from related.
/// 
/// Copies the internal pointer from \a rhs and calls addRef().
/// If we're already storing an internal pointer, this pointer will be release()'ed. 
//--------------------------------------------------------------------------------------------------
template<typename T>
template<typename T2>
PtrRef<T>& PtrRef<T>::operator=(const PtrRef<T2>& rhs)
{
    // Copy-and-swap (on temporary obj)
    PtrRef<T>(rhs).swap(*this);
    return *this;
}


//--------------------------------------------------------------------------------------------------
/// Returns the naked pointer this object is associated with. 
/// 
/// Added to be able to write the same code for smart pointers as for normal naked pointers.
//--------------------------------------------------------------------------------------------------
template<typename T>
inline T* PtrRef<T>::operator->()
{
    //CVF_TIGHT_ASSERT(m_object); 
    return m_object;
}


//--------------------------------------------------------------------------------------------------
/// Returns naked const pointer this object is associated with. 
//--------------------------------------------------------------------------------------------------
template<typename T>
inline const T* PtrRef<T>::operator->() const
{
    //CVF_TIGHT_ASSERT(m_object); 
    return m_object;
}


//--------------------------------------------------------------------------------------------------
/// Dereference operator returning a modifiable reference to the associated object. 
/// 
/// Added to be able to write the same code for smart pointers as for normal naked pointers.
/// \code
/// PtrRef<MyObject> myObject = new MyObject;
/// *myObject.doSomething();
/// \endcode
//--------------------------------------------------------------------------------------------------
template<typename T>
inline T& PtrRef<T>::operator*()
{
    //CVF_TIGHT_ASSERT(m_object); 
    return *m_object;
}


//--------------------------------------------------------------------------------------------------
/// Dereference operator returning a const reference to the associated object. 
//--------------------------------------------------------------------------------------------------
template<typename T>
inline const T& PtrRef<T>::operator*() const
{
    //CVF_TIGHT_ASSERT(m_object); 
    return *m_object;
}


//--------------------------------------------------------------------------------------------------
/// Returns the naked pointer
//--------------------------------------------------------------------------------------------------
template<typename T>
inline T* PtrRef<T>::get()
{
    return m_object;
}


//--------------------------------------------------------------------------------------------------
/// Returns naked const pointer
//--------------------------------------------------------------------------------------------------
template<typename T>
inline const T* PtrRef<T>::get() const
{
    return m_object;
}


//--------------------------------------------------------------------------------------------------
/// Does less than comparison of two %PtrRef objects.
/// 
/// Returns true if this object's internal pointer is less than the internal pointer of \a rhs.
/// 
/// The comparison is done by comparing the internal pointer values ( this.p() < rhs.p() )
/// This operator is used in several STL collections (e.g. std::set) as well as in many STL algorithms
/// (e.g. std::sort() ).
//--------------------------------------------------------------------------------------------------
template<typename T>
bool PtrRef<T>::operator<(const PtrRef& rhs) const
{
    //CVF_TIGHT_ASSERT((m_object != NULL) && (rhs.p() != NULL));

    return m_object < rhs.get();
}


//--------------------------------------------------------------------------------------------------
/// Operator for supporting tests if the objects internal pointer is NULL or not
/// 
/// Using the safe bool idiom, %PtrRef allows for code like
/// 
/// \code
/// PtrRef<cee::Image> myImage = new cee::Image;
/// ...
/// if (myImage)
/// {
///     // Do something
/// }
/// \endcode
/// without defining the operator bool(), which have many potential side effects.
//--------------------------------------------------------------------------------------------------
template<typename T>
PtrRef<T>::operator unspecified_bool_type() const
{
    return m_object == 0 ? 0 : &PtrRef::m_object;
}


//--------------------------------------------------------------------------------------------------
/// Exchanges the contents of the two smart pointers.
/// 
/// \a other is a modifiable reference to the %PtrRef object that should have its contents swapped.
/// 
/// Swap the associated pointer in this and the passed %PtrRef object. Will not modify the reference 
/// count of any of the associated objects.
//--------------------------------------------------------------------------------------------------
template<typename T>
void PtrRef<T>::swap(PtrRef& other)
{
    T* tmp = other.m_object;
    other.m_object = m_object; 
    m_object = tmp;
}


//--------------------------------------------------------------------------------------------------
/// Detach the %PtrRef from the internal object WITHOUT calling release on the object.
/// 
/// After calling, *this == NULL and associated object will have the same refCount as before the call.
/// 
/// \note Use with caution! The associated object will not be release()'ed and it is the callers
///       responsibility to ensure proper life cycle management of the object.
//--------------------------------------------------------------------------------------------------
template <typename T>
T* cee::PtrRef<T>::detach()
{
    T* tmp = m_object;
    m_object = NULL;
    return tmp;
}

}  // namespace cee
