//##################################################################################################
//
//  Ceetron Desktop Components
//  Component: UnstructGrid
//  
//  --------------------------------------------------------------------------------------------
//  Copyright (C) 2012, Ceetron AS
//  This is UNPUBLISHED PROPRIETARY SOURCE CODE of Ceetron AS. The contents of this file may 
//  not be disclosed to third parties, copied or duplicated in any form, in whole or in part,
//  without the prior written permission of Ceetron AS.
//##################################################################################################

#pragma once

#include "CeeCore/Base.h"
#include "CeeCore/PtrRef.h"
#include "CeeCore/RefCountedObject.h"
#include "CeeUnstructGrid/CustomResultFunction.h"

#include <vector>

namespace cee {
namespace ug {
    
class DataResultGroup;
class DataResultScalar;
class DataResultVector;
class DataResultDisplacement;
class DataResultSymmetricTensor;

class DataSource;
class DataState;
class ResultInfo;

typedef ResultInfo CustomResultArgument;

//==================================================================================================
//
//  
//  
//==================================================================================================
class CEE_UG_EXPORT CustomResultGenerator : public RefCountedObject
{

public:

    //--------------------------------------------------------------------------------------------------
    /// Input data results
    //--------------------------------------------------------------------------------------------------
    struct InputDataResults
    {
        std::vector<const DataResultScalar*>            scalars;            ///< scalars[i] is the DataResultScalar object of argument i if the argument is a scalar, NULL if not
        std::vector<const DataResultVector*>            vectors;            ///< vectors[i] is the DataResultVector object of argument i if the argument is a vector, NULL if not
        std::vector<const DataResultDisplacement*>      displacements;      ///< displacements[i] is the DataResultDisplacement object of argument i if the argument is a displacement, NULL if not
        std::vector<const DataResultSymmetricTensor*>   symmetricTensors;   ///< symmetricTensors[i] is the DataResultSymmetricTensor object of argument i if the argument is a symmetric tensor, NULL if not

    };

    //--------------------------------------------------------------------------------------------------
    /// Input value data
    //--------------------------------------------------------------------------------------------------
    struct InputValueData
    {
        std::vector<const double*>                      values;             ///< values[i] is the raw pointer to the value array of argument i
        std::vector<size_t>                             dimensions;         ///< dimensions[i] is the dimension of argument i (1 for scalar, 3 for vector or displacement, 6 for symmetric tensor)
        std::vector<size_t>                             counts;             ///< counts[i] is the number of items (nodes, elements, element-nodes) of argument i
    };

    //--------------------------------------------------------------------------------------------------
    /// Output data result
    //--------------------------------------------------------------------------------------------------
    struct OutputDataResult
    {
        DataResultScalar*                               scalar;             ///< The DataResultScalar object of the generated output result if it is a scalar, NULL if not
        DataResultVector*                               vector;             ///< The DataResultVector object of the generated output result if it is a vector, NULL if not
        DataResultDisplacement*                         displacement;       ///< The DataResultDisplacement object of the generated output result if it is a displacement, NULL if not
        DataResultSymmetricTensor*                      symmetricTensor;    ///< The DataResultSymmetricTensor object of the generated output result if it is a symmetric tensor, NULL if not
    };

    //--------------------------------------------------------------------------------------------------
    /// Output value data
    //--------------------------------------------------------------------------------------------------
    struct OutputValueData
    {
        double*                                         values;             ///< The raw pointer to the value array of the output result
        size_t                                          dimension;          ///< The dimension of the output result (1 for scalar, 3 for vector or displacement, 6 for symmetric tensor)
        size_t                                          count;              ///< The number of items of the output result (nodes, elements, element-nodes)
    };

public:
    CustomResultGenerator();
    CustomResultGenerator(size_t argumentCount);
    virtual ~CustomResultGenerator();

    bool                    operator==(const CustomResultGenerator& rhs) const;
    bool                    operator!=(const CustomResultGenerator& rhs) const;

    size_t argumentCount() const;

    ResultType                  argumentType(size_t index) const;
    void                        setArgumentType(size_t index, ResultType resultType);

    void                        setFunction(CustomResultFunction* function);

    bool                        hasFunction() const;
    const CustomResultFunction* function() const;

    virtual bool                generate(const DataSource* dataSource, const ResultInfo& result, const std::vector<CustomResultArgument>& arguments, DataState* state) const;

protected:
    bool                        checkArguments(const std::vector<CustomResultArgument>& arguments) const;
    bool                        computeCustomArguments(const DataSource* dataSource, const std::vector<CustomResultArgument>& arguments, DataState* state) const;
    bool                        gatherInputDataResults(const DataResultGroup* resultGroup, const std::vector<CustomResultArgument>& arguments, size_t partCount, InputDataResults* inputDataResults) const;
    bool                        gatherInputValuePointers(const std::vector<CustomResultArgument>& arguments, size_t partIndex, const InputDataResults& inputDataResults, InputValueData* inputValueData) const;
    bool                        getOuputDataResult(DataResultGroup* resultGroup, const ResultInfo& result, OutputDataResult* outputDataResult) const;
    bool                        getOutputValuePointer(size_t partIndex, const OutputDataResult& outputDataResult, const InputValueData& inputValueData, size_t outputItemCount, OutputValueData* outputValueData) const;

private:
    CEE_PRIVATE_IMPL(CustomResultGenerator);
    CEE_DISALLOW_COPY_AND_ASSIGN(CustomResultGenerator);
};

}  // namespace ug
}  // namespace cee
