//##################################################################################################
//
//  Ceetron Desktop Components
//  Component: UnstructGrid
//  
//  --------------------------------------------------------------------------------------------
//  Copyright (C) 2011, Ceetron AS
//  This is UNPUBLISHED PROPRIETARY SOURCE CODE of Ceetron AS. The contents of this file may 
//  not be disclosed to third parties, copied or duplicated in any form, in whole or in part,
//  without the prior written permission of Ceetron AS.
//##################################################################################################

#pragma once

#include "CeeUnstructGrid/Base.h"

#include "CeeCore/RefCountedObject.h"
#include "CeeCore/Color3f.h"
#include "CeeCore/Str.h"
#include "CeeUnstructGrid/ColorMapper.h"
#include "CeeUnstructGrid/ContourLines.h"


namespace cee {
namespace ug {

class ColorMapper;

//==================================================================================================
//
// 
//
//==================================================================================================
class CEE_UG_EXPORT ScalarSettings : public RefCountedObject
{
public:
    /// Position of the custom markers to show above/below the color legend.
    enum MarkerPosition 
    {
        ABOVE,          ///< Show the marker above the color legend.
        BELOW           ///< Show the marker below the color legend
    };

    /// Extent of the automatic range computation
    enum AutoRangeMode
    {
        OFF,            ///< Scalar range is not computed automatically
        ALL_ITEMS,      ///< Scalar range is computed automatically from scalar result values of all items in the model
        VISIBLE_ITEMS   ///< Scalar range is computed automatically from scalar result values of visible items in the model
    };
    
    /// Control the visibility of the color legend
    enum LegendVisibilityMode
    {
        NEVER,          ///< Always hide the color legend for this result
        AUTO,           ///< Show the color legend for this result if used by any visible parts in the view (default)
        ALWAYS          ///< Always show the color legend for this result, no matter if it is in use or not
    };

    /// Position of the legend tick marks
    /// 
    /// LEFT/RIGHT are only for vertical legends. TOP/BOTTOM are only for horizontal legends. Default is FOLLOW_LAYOUT. Follow layout
    /// will have the tick marks on the "inside" of the legend.
    enum TickMarksPosition
    {
        LEFT,           ///< Draw tick marks on the left side of the color legend
        RIGHT,          ///< Draw tick marks on the right side of the color legend
        TOP,            ///< Draw tick marks above the color legend
        BOTTOM,         ///< Draw tick marks below the color legend
        FOLLOW_LAYOUT   ///< Automatically position the tick marks left or right based on the color legend's alignment in the view (default)
    };

    /// Select the orientation of the legend. 
    ///
    /// Default is VERTICAL. 
    enum LegendDirection
    {
        HORIZONTAL,     ///< Draw a horizontal legend
        VERTICAL        ///< Draw a vertical legend
    };

public:
    virtual ~ScalarSettings();

    double                  rangeMinimum() const;
    double                  rangeMaximum() const;
    void                    setRange(double min, double max);

    AutoRangeMode           autoRangeMode() const;
    void                    setAutoRangeMode(AutoRangeMode mode);

    const ColorMapper&      colorMapper() const;
    ColorMapper&            colorMapper();

    const ContourLines&     contourLines() const;
    ContourLines&           contourLines();

    bool                    logarithmicMapping() const;
    void                    setLogarithmicMapping(bool logMapper);

    bool                    nodeAveragedValues() const;
    void                    setNodeAveragedValues(bool useNodeAveraging);
 
    LegendVisibilityMode    legendVisibilityMode() const;
    void                    setLegendVisibilityMode(LegendVisibilityMode legendVisibilityMode);
    Color3f                 legendTextColor() const;
    void                    setLegendTextColor(const Color3f& color);
    Color3f                 legendLineColor() const;
    void                    setLegendLineColor(const Color3f& color);
    unsigned int            legendLineWidth() const;
    void                    setLegendLineWidth(unsigned int lineWidth);
    unsigned int            legendBarWidth() const;
    void                    setLegendBarWidth(unsigned int widthInPixels);
    Str                     customLegendTitle() const;
    void                    setCustomLegendTitle(const Str& title);
    bool                    forceTickMarkLinesToCoverLegend() const;
    void                    setForceTickMarkLinesToCoverLegend(bool alwaysCover);
    bool                    drawLegendWithMinimumValueOnTop() const;
    void                    setDrawLegendWithMinimumValueOnTop(bool minOnTop);
    bool                    skipOverlappingTickMarkLabels() const;
    void                    setSkipOverlappingTickMarkLabels(bool skip);
    LegendVisibilityMode    aboveBelowRangeColorMarkerVisibilityMode() const;
    void                    setAboveBelowRangeColorMarkerVisibilityMode(LegendVisibilityMode mode);
    TickMarksPosition       tickMarksPosition() const;
    void                    setTickMarksPosition(TickMarksPosition position);
    LegendDirection         legendDirection() const;
    void                    setLegendDirection(LegendDirection direction);

    bool                    isSingleColorZeroRangeModeEnabled(Color3f* color = NULL) const;
    void                    setSingleColorZeroRangeMode(const Color3f& color);
    void                    disableSingleColorZeroRangeMode();

    bool                    forceUniformTickMarkSpacing() const;
    void                    setForceUniformTickMarkSpacing(bool forceUniformSpacing);

    bool                    forceUniformTextureLevels() const;
    void                    setForceUniformTextureLevels(bool forceUniformLevels);

    char                    numericFormat() const;
    void                    setNumericFormat(char format);
    int                     numericPrecision() const;
    void                    setNumericPrecision(int precision);
                            
    double                  numericScalingConstant() const;
    double                  numericScalingFirstDegreeFactor() const;
    void                    setNumericScaling(double constantTerm, double firstDegreeTerm);

    void                    addColorMarker(MarkerPosition position, const Color3f& color, const Str& description);
    size_t                  colorMarkerCount() const;
    MarkerPosition          colorMarkerPosition(size_t index) const;
    Color3f                 colorMarkerColor(size_t index) const;
    Str                     colorMarkerDescription(size_t index) const;
    void                    clearColorMarkers();

    float                   colorMarkerBoxHeight() const;
    void                    setColorMarkerBoxHeight(float height);
    float                   colorMarkerMargin() const;
    void                    setColorMarkerMargin(float margin);

    bool                    isFringesElementFilteringEnabled() const;
    void                    disableFringesElementFiltering();
    double                  fringesElementFilteringVisibleRangeMinimum() const;
    double                  fringesElementFilteringVisibleRangeMaximum() const;
    void                    setFringesElementFilteringVisibleRange(double min, double max);
    bool                    filterUndefinedElements() const;
    void                    setFilterUndefinedElements(bool filter);

private:
    ScalarSettings();
    CEE_PRIVATE_F(UnstructGridModel);
    CEE_PRIVATE_IMPL(ScalarSettings);
    CEE_DISALLOW_COPY_AND_ASSIGN(ScalarSettings);
};

}  // namespace ug
}  // namespace cee

