//##################################################################################################
//
//  Ceetron Desktop Components
//  Component: UnstructGrid
//  
//  --------------------------------------------------------------------------------------------
//  Copyright (C) 2011, Ceetron AS
//  This is UNPUBLISHED PROPRIETARY SOURCE CODE of Ceetron AS. The contents of this file may 
//  not be disclosed to third parties, copied or duplicated in any form, in whole or in part,
//  without the prior written permission of Ceetron AS.
//##################################################################################################

#pragma once

#include "CeeUnstructGrid/Base.h"

#include "CeeCore/RefCountedObject.h"
#include "CeeVisualization/Model.h"
#include "CeeCore/Color3f.h"
#include "CeeCore/BoundingBox.h"
#include "CeeCore/Str.h"
#include "CeeCore/Vec2f.h"
#include "CeeUnstructGrid/ModelSettings.h"
#include "CeeUnstructGrid/ResultInfo.h"

class VTFrameSet;

namespace cee {

class Ray;

namespace ug {

class HitItem;
class HitItemCollection;
class PartHitItems;
class ScalarSettings;
class VectorSettings;
class DisplacementSettings;
class DataSource;
class ModelSpec;
class PartSettings;
class CuttingPlane;
class Isosurface;
class Isovolume;
class ParticleTraceGroup;
class MirrorSettings;
class Error;


//==================================================================================================
//
// 
//
//==================================================================================================
class Progress
{
public:
    Progress() {}
    virtual ~Progress() {}

    virtual void update(int pctCompl) = 0;      ///< Updates progress status in percent
    virtual bool aborted() = 0;                 ///< Returns true if the operation should be aborted
};


//==================================================================================================
//
// 
//
//==================================================================================================
class CEE_UG_EXPORT UnstructGridModel : public vis::Model
{
public:
    /// Update action. Used in updateVisualization() to specify level of update needed in order to optimize the update.
    enum UpdateAction
    {
        NORMAL,                 ///< Default update. Settings and data are inspected and any needed update is performed
        HIGHLIGHT,              ///< Use this if you only used the setHighlight() method on Parts, CuttingPlane, Isosurface or Isovolume
        VISIBILITY,             ///< Use this if you only used the setVisible() method on Parts, CuttingPlane, Isosurface or Isovolume
        COLOR_AND_OPACITY,      ///< Use this if you only used the set*Color() or setOpacity() on Parts, CuttingPlane, Isosurface or Isovolume
        SCALAR_SETTINGS,        ///< Use this if you only used the ScalarSettings and ColorMapper methods
        COLOR_LEGEND,           ///< Visual configuration of the color legend (visible, text color, etc)
        NODE_POSITIONS,         ///< Only the positions of the nodes have changed. Note: Only use this when the position of the nodes have changed, but not the number of nodes or element topology.
        TRANSFORMATION_RESULTS, ///< Only the transformation results have changed, if any.
        CUTTING_PLANE,          ///< Use this if you only changed cutting plane parameters (apart from mapped result). 
                                ///  This will recompute the cutting plane in the current frame. All other frames will be 
                                ///  computed whenever set visible using the setCurrentFrameIndex() method.
        ISOSURFACE,             ///< Use this if you only changed isosurface parameters (apart from mapped result). 
                                ///  This will recompute the isosurface in the current frame. All other frames will be 
                                ///  computed whenever set visible using the setCurrentFrameIndex() method.
        ISOVOLUME,              ///< Use this if you only changed isovolume parameters (apart from mapped result). 
                                ///  This will recompute the isovolume in the current frame. All other frames will be 
                                ///  computed whenever set visible using the setCurrentFrameIndex() method.
        PARTICLE_TRACE,         ///< Use this if you only changed particle trace parameters (apart from mapped result). 
                                ///  This will recompute the particle trace in the current frame. All other frames will be 
                                ///  computed whenever set visible using the setCurrentFrameIndex() method.
        REGENERATE_ALL          ///< Force recreation of the supporting display model. Also forces reading of 
                                ///  data from an interface (if in use).
    };

public:
    UnstructGridModel();
    virtual ~UnstructGridModel();

    const DataSource*           dataSource() const;
    DataSource*                 dataSource();
    void                        setDataSource(DataSource* dataSource);

    const ModelSpec&            modelSpec() const;
    ModelSpec&                  modelSpec();

    virtual bool                updateVisualization(UpdateAction updateAction = NORMAL, Progress* progress = NULL, Error* error = NULL);
    void                        clearVisualization();

    virtual size_t              startUpdateVisualizationPerFrame(Error* error = NULL);
    virtual bool                updateVisualizationForFrame(size_t frameIndex, Error* error = NULL);

    const ModelSettings&        modelSettings() const;
    ModelSettings&              modelSettings();

    bool                        rayIntersect(int x, int y, const vis::View& view, HitItem* hitItem) const;
    bool                        rayIntersectAllHits(int x, int y, const vis::View& view, HitItemCollection* hitItems) const;
    bool                        regionIntersect(int x, int y, unsigned int width, unsigned int height, const vis::View& view, bool acceptPartiallyContainedItems, bool includeElementIndices, std::vector<PartHitItems>* itemList) const;
    bool                        polygonIntersect(const std::vector<Vec2f>& polygonInWindowCoordinates, const vis::View& view, bool acceptPartiallyContainedItems, bool includeElementIndices, std::vector<PartHitItems>* itemList) const;
    bool                        rayIntersect(const Ray& ray, HitItem* hitItem) const;
    bool                        rayIntersectAllHits(const Ray& ray, HitItemCollection* hitItems) const;

    size_t                      frameCount() const;
    size_t                      currentFrameIndex() const;
    int                         currentStateId() const;
    void                        setCurrentFrameIndex(size_t frameIndex);
    double                      particleTraceTime() const;
    void                        setParticleTraceTime(double time);
    virtual BoundingBox         boundingBox();
    virtual BoundingBox         frameBoundingBox(size_t frameIndex);
    virtual BoundingBox         partBoundingBox(size_t frameIndex, size_t globalGeometryIndex, int partId);
    virtual Str                 modelInfo() const;

    PartSettings*               partSettings(size_t globalGeometryIndex, int partId);     
    const PartSettings*         existingPartSettings(size_t globalGeometryIndex, int partId) const;
    ScalarSettings*             scalarSettings(int resultId);  
    const ScalarSettings*       existingScalarSettings(int resultId) const;  
    VectorSettings*             vectorSettings(int resultId);  
    const VectorSettings*       existingVectorSettings(int resultId) const;
    DisplacementSettings*       displacementSettings(int resultId);  
    const DisplacementSettings* existingDisplacementSettings(int resultId) const;  

    size_t                      cuttingPlaneCount() const;
    const CuttingPlane*         cuttingPlane(size_t index) const;
    CuttingPlane*               cuttingPlane(size_t index);
    void                        addCuttingPlane(CuttingPlane* cuttingplane);
    void                        removeCuttingPlane(const CuttingPlane* cuttingplane);
    void                        removeAllCuttingPlanes();

    size_t                      isosurfaceCount() const;
    const Isosurface*           isosurface(size_t index) const;
    Isosurface*                 isosurface(size_t index);
    void                        addIsosurface(Isosurface* isosurf);
    void                        removeIsosurface(const Isosurface* isosurf);
    void                        removeAllIsosurfaces();

    size_t                      isovolumeCount() const;
    const Isovolume*            isovolume(size_t index) const;
    Isovolume*                  isovolume(size_t index);
    void                        addIsovolume(Isovolume* isosurf);
    void                        removeIsovolume(const Isovolume* volume);
    void                        removeAllIsovolumes();

    size_t                      particleTraceGroupCount() const;
    const ParticleTraceGroup*   particleTraceGroup(size_t index) const;
    ParticleTraceGroup*         particleTraceGroup(size_t index);
    void                        addParticleTraceGroup(ParticleTraceGroup* group);
    void                        removeParticleTraceGroup(const ParticleTraceGroup* group);
    void                        removeAllParticleTraceGroups();
    bool                        particleTraceTimeRange(size_t frameIndex, double* minimumTime, double* maximumTime) const;

    bool                        scalarRange(int resultId, double* min, double* max) const;
    bool                        vectorRange(int resultId, double* min, double* max) const;
    bool                        displacementRange(double* min, double* max) const;
    
    cee::Color3f                elementSetColor(int setId) const;
    void                        setElementSetColor(int setId, const Color3f& color);

    MirrorSettings&             mirrorSettings();
    const MirrorSettings&       mirrorSettings() const;

public:
    VTFrameSet*                                     internal_vtFrameSet();
    void                                            internal_debug(int code);

private:
    virtual cvf::Model*                             internal_cvfModel();
    virtual std::vector<vis::OverlayItem*>          internal_overlayItems();
    virtual void                                    internal_setViewCapabilities(const cvf::OpenGLCapabilities& capabilities, bool viewDoesHighlighting);
    virtual bool                                    internal_canAddToViewWithCapabilities(const cvf::OpenGLCapabilities& viewCapabilities) const;
    virtual NotifyAction                            internal_detectAndConfigureViewClippingInModel(const vis::View& triggeringView, cvf::OpenGLContext* cvfContext);
    virtual void                                    internal_applyViewClippingPlanesBeforeRendering(const cvf::ClipPlaneSet* viewClipPlaneSet, const cvf::Plane* dropShadowClipPlane);
    virtual void                                    internal_applyCameraTransformationToClippingPlanes(const cvf::Camera& camera);

    CEE_BASE_F(DataSourceVTFx);
    CEE_BASE_F(DataSourceVTF);
    CEE_BASE_F(CuttingPlaneAnimation);
    CEE_PRIVATE_IMPL(UnstructGridModel);
    CEE_DISALLOW_COPY_AND_ASSIGN(UnstructGridModel);
};

}  // namespace ug
}  // namespace cee
