//##################################################################################################
//
//  Ceetron Desktop Components
//  Component: Visualization
//  
//  --------------------------------------------------------------------------------------------
//  Copyright (C) 2011, Ceetron AS
//  This is UNPUBLISHED PROPRIETARY SOURCE CODE of Ceetron AS. The contents of this file may 
//  not be disclosed to third parties, copied or duplicated in any form, in whole or in part,
//  without the prior written permission of Ceetron AS.
//##################################################################################################

#pragma once

#include "CeeVisualization/Base.h"
#include "CeeCore/RefCountedObject.h"
#include "CeeCore/Ray.h"
#include "CeeCore/Plane.h"

namespace cee {
class Vec3d;
class BoundingBox;

namespace vis {
class View;
class CameraInputHandler;

//==================================================================================================
//
// 
//
//==================================================================================================
class CEE_VIS_EXPORT Camera : public RefCountedObject
{
public:
    /// Projection type
    enum ProjectionType
    {
        PERSPECTIVE,        ///< Perspective projection
        ORTHO               ///< Orthographic (parallel) projection
    };
    /// Rotation screen axis
    enum ScreenAxis
    {
        SCREEN_X_AXIS,      ///< Horizontal axis to the right
        SCREEN_Y_AXIS,      ///< Vertical axis to the up
        SCREEN_Z_AXIS       ///< Normal direction leaving the screen
    };

public:
    virtual ~Camera();
    
    void            setFromLookAt(const Vec3d& eye, const Vec3d& vrp, const Vec3d& up);
    void            toLookAt(Vec3d* eye, Vec3d* vrp, Vec3d* up) const;
    void            fitView(const BoundingBox& boundingBox, const Vec3d& dir, const Vec3d& up, double adjustmentFactor = 0.9);
    void            fitViewOrtho(const BoundingBox& boundingBox, double eyeDist, const Vec3d& dir, const Vec3d& up, double adjustmentFactor = 0.9);
    void            rotateGlobalAxis(double radians, const Vec3d& axis);
    void            rotateScreenAxis(double radians, ScreenAxis dir);
    void            rubberbandZoom(int x, int y, unsigned int width, unsigned int height);

    void            setProjectionAsPerspective(double fieldOfViewYDeg, double nearPlane, double farPlane);
    void            setProjectionAsOrtho(double height, double nearPlane, double farPlane);

    void            enableAutoClipMinimumNearDistance(double minNearDistance);
    void            enableAutoClipFixedNearDistance(double fixedNearDistance);
    void            disableAutoClip();

    ProjectionType  projection() const;
    double          fieldOfViewYDegrees() const;
    double          nearPlane() const;
    double          farPlane() const;
    double          frontPlaneFrustumHeight() const;
    Vec3d           computeFitViewEyePosition(const BoundingBox& boundingBox, const Vec3d& dir, const Vec3d& up, double adjustmentFactor = 0.9, double fieldOfViewYDeg = 40.0) const;

    unsigned int    viewportWidth() const;
    unsigned int    viewportHeight() const;
    int             viewportPositionX() const;
    int             viewportPositionY() const;
    void            setViewport(int x, int y, unsigned int width, unsigned int height);

    Ray             rayFromWindowCoordinates(int x, int y) const;
    Plane           planeFromWindowCoordinates(int x1, int y1, int x2, int y2) const;
    bool            unproject(const Vec3d& coord, Vec3d* out) const;
    bool            project(const Vec3d& point, Vec3d* out) const;

    Mat4d           viewMatrix() const;
    void            setViewMatrix(Mat4d mat);

    CameraInputHandler*         inputHandler();
    const CameraInputHandler*   inputHandler() const;
    void                        setInputHandler(CameraInputHandler* handler);

private:
    Camera(View* view);

    CEE_PRIVATE_F(View);
    CEE_PRIVATE_F(OverlayAxisCross);
    CEE_PRIVATE_F(OverlayNavigationCube);
    CEE_PRIVATE_IMPL(Camera);
    CEE_DISALLOW_COPY_AND_ASSIGN(Camera);
};

} // namespace vis
} // namespace cee
