//##################################################################################################
//
//  Ceetron Desktop Components
//  Component: Visualization
//  
//  --------------------------------------------------------------------------------------------
//  Copyright (C) 2014, Ceetron AS
//  This is UNPUBLISHED PROPRIETARY SOURCE CODE of Ceetron AS. The contents of this file may 
//  not be disclosed to third parties, copied or duplicated in any form, in whole or in part,
//  without the prior written permission of Ceetron AS.
//##################################################################################################

#pragma once

#include "CeeVisualization/Base.h"
#include "CeeCore/RefCountedObject.h"
#include "CeeVisualization/MouseEvent.h"
#include "CeeVisualization/WheelEvent.h"

namespace cee {
class Vec3d;

namespace vis {
class Camera;

//==================================================================================================
//
// Base class for CameraInputHandler
//
//==================================================================================================
class CEE_VIS_EXPORT CameraInputHandler : public RefCountedObject
{
public:
    /// Navigation types
    enum NavigationType
    {
        NONE,               ///< No navigation
        PAN,                ///< Pan. Move the model in the screen plane.
        WALK,               ///< Walk. Move the camera position (eye) forward or backward in the view direction.
        ZOOM,               ///< Adjust the Field of view (perspective projection) or frustum height (orthographic projection) to resize the model.
        ROTATE,             ///< Rotate the model around the current rotation point.
        ROLL                ///< Roll. Rotate the model around view direction.
    };
    
public:
    CameraInputHandler();
    virtual ~CameraInputHandler();
    
    virtual void            mousePressEvent(MouseButton buttonPressed, const MouseEvent& mouseEvent);
    virtual bool            mouseMoveEvent(const MouseEvent& mouseEvent);
    virtual void            mouseReleaseEvent(MouseButton buttonReleased, const MouseEvent& mouseEvent);
    virtual void            wheelEvent(const WheelEvent& theWheelEvent);
    virtual void            wheelEvent(const WheelEvent& theWheelEvent, const cee::Vec3d& position);

    Vec3d                   rotationPoint() const;
    virtual void            setRotationPoint(const Vec3d& rotationPoint);
    double                  rotationSensitivity() const;
    void                    setRotationSensitivity(double sensitivity);
    double                  rollSensitivity() const;
    void                    setRollSensitivity(double sensitivity);
    double                  walkSensitivity() const;
    void                    setWalkSensitivity(double sensitivity);

    NavigationType          activeNavigation() const;
    void                    startNavigation(NavigationType navigationType, int x, int y);
    bool                    updateNavigation(int x, int y);
    void                    endNavigation();

    virtual NavigationType  navigationTypeFromInputState(MouseButtons mouseButtons, KeyboardModifiers keyboardModifiers) const = 0; ///< Returns the navigation type based on the state of the input event. Must be implemented in all derived classes.
    virtual bool            reverseZoom() const;
    
protected:
    virtual NavigationType  wheelNavigationType() const;
    
    Camera*                 camera();
    void                    setMinimumWalkTargetDistance(double distance);
    
private:
    CEE_BASE_F(Camera);
    CEE_PRIVATE_IMPL(CameraInputHandler);
    CEE_DISALLOW_COPY_AND_ASSIGN(CameraInputHandler);
};
    

    
//==================================================================================================
//
// Standard Ceetron style Zoom navigation
//
//==================================================================================================
class CEE_VIS_EXPORT CameraInputHandlerZoom : public CameraInputHandler
{
public:
    CameraInputHandlerZoom(bool updateEyePositionOnSetRotationPoint = true);
    
    virtual void    wheelEvent(const WheelEvent& theWheelEvent);
    virtual void    wheelEvent(const WheelEvent& theWheelEvent, const cee::Vec3d& pos);
    virtual void    setRotationPoint(const Vec3d& rotationPoint);
    
protected:
    virtual NavigationType  navigationTypeFromInputState(MouseButtons mouseButtons, KeyboardModifiers keyboardModifiers) const;
    virtual NavigationType  wheelNavigationType() const;
    
private:
    bool            m_updateEyeOnNewRotationPoint;
};

    
    
//==================================================================================================
//
// Standard Ceetron style Walk navigation
//
//==================================================================================================
class CEE_VIS_EXPORT CameraInputHandlerWalk : public CameraInputHandler
{
public:
    CameraInputHandlerWalk();
    
    virtual void    wheelEvent(const WheelEvent& theWheelEvent);
    virtual void    wheelEvent(const WheelEvent& theWheelEvent, const cee::Vec3d& pos);

    void            setMinimumTargetDistance(double minDistance);

protected:
    virtual NavigationType  navigationTypeFromInputState(MouseButtons mouseButtons, KeyboardModifiers keyboardModifiers) const;
    virtual NavigationType  wheelNavigationType() const;
};

} // namespace vis
} // namespace cee
