//##################################################################################################
//
//  Ceetron Desktop Components
//  Component: Visualization
//  
//  --------------------------------------------------------------------------------------------
//  Copyright (C) 2014, Ceetron AS
//  This is UNPUBLISHED PROPRIETARY SOURCE CODE of Ceetron AS. The contents of this file may 
//  not be disclosed to third parties, copied or duplicated in any form, in whole or in part,
//  without the prior written permission of Ceetron AS.
//##################################################################################################

#pragma once

#include "CeeVisualization/Base.h"
#include "CeeCore/RefCountedObject.h"
#include "CeeVisualization/MarkupPart.h"
#include "CeeCore/Color3f.h"
#include "CeeCore/Str.h"
#include "CeeCore/Vec3d.h"
#include "CeeCore/PtrRef.h"

namespace cee {
class BoundingBox;

namespace vis {
class Font;

//==================================================================================================
//
// 
//
//==================================================================================================
class CEE_VIS_EXPORT MarkupPartLabels : public MarkupPart
{
public:
    /// Enum specifying the way to draw the labels. Controls the visibility of the label.
    enum LabelDrawMode
    {
        DRAW_ONLY_VISIBLE,          ///< Label is visible if the position of the label (position provided in the add()) is visible
        DRAW_WITH_ZBUFFER_TEST,     ///< Draw the label as geometry with z buffer test
        DRAW_WITHOUT_ZBUFFER_TEST   ///< Draw the label without z buffer test. The label will always be visible.
    };

    /// Enum specifying the direction to draw the labels. 
    enum Orientation
    {
        HORIZONTAL,     ///< Horizontal text direction
        VERTICAL_UP,    ///< Vertical direction, upward text
        VERTICAL_DOWN   ///< Vertical direction, downward text
    };

    /// Horizontal alignment of the text when drawing multiple lines
    enum HorizontalTextAlignment
    {
        LEFT,       ///< Align left
        CENTER,     ///< Align center
        RIGHT       ///< Align right
    };


public:
    MarkupPartLabels();
    MarkupPartLabels(LabelDrawMode labelDrawMode, const Color3f& textColor, const Color3f& borderColor, const Color3f& backgroundColor);
    MarkupPartLabels(LabelDrawMode labelDrawMode, const Color3f& textColor, const Color3f& borderColor, const Color3f& backgroundColor, const Color3f& markerColor);
    virtual ~MarkupPartLabels();

    virtual PartType                type() const;
    virtual BoundingBox             boundingBox() const;

    virtual void                    setPriority(int priority);
    virtual int                     priority() const;

    virtual const Mat4d&            transformation() const;
    virtual void                    setTransformation(const Mat4d& matrix);

    void                            setAbsoluteOffset(int x, int y);
    int                             absoluteOffsetX() const;
    int                             absoluteOffsetY() const;
                                    
    void                            setRelativeOffset(double x, double y);
    double                          relativeOffsetX() const;
    double                          relativeOffsetY() const;

    LabelDrawMode                   drawMode() const;
    void                            setDrawMode(LabelDrawMode mode);

    Orientation                     orientation() const;
    void                            setOrientation(Orientation direction);

    HorizontalTextAlignment         horizontalTextAlignment() const;
    void                            setHorizontalTextAlignment(HorizontalTextAlignment alignment);

    unsigned int                    maximumNumberOfLabelsToDraw() const;
    void                            setMaximumNumberOfLabelsToDraw(unsigned int maximumNumberToDraw);

    float                           eyeLiftFactor() const;
    void                            setEyeLiftFactor(float factor);
    
    const Font*                     font() const;
    void                            setFont(Font* labelFont);

    const Color3f&                  textColor() const;
    void                            setTextColor(const Color3f& color);

    const Color3f&                  backgroundColor() const;
    void                            setBackgroundColor(const Color3f& color);

    const Color3f&                  borderColor() const;
    void                            setBorderColor(const Color3f& color);

    const Color3f&                  markerColor() const;
    void                            setMarkerColor(const Color3f& color);
    
    float                           borderWidth() const;
    void                            setBorderWidth(float width);

    float                           markerPointSize() const;
    void                            setMarkerPointSize(float pointSize);

    bool                            drawBackground() const;
    void                            setDrawBackground(bool drawBackground);

    bool                            drawBorder() const;
    void                            setDrawBorder(bool drawBorder);

    bool                            drawMarker() const;
    void                            setDrawMarker(bool drawMarker);
    
    bool                            offsetLabelToFitInView() const;
    void                            setOffsetLabelToFitInView(bool offsetLabel);
    
    size_t                          count() const;
    const cee::Str&                 text(size_t index) const;
    const cee::Vec3d&               position(size_t index) const;
    unsigned int                    width(size_t index);
    unsigned int                    height(size_t index);
    size_t                          add(const Vec3d& position, const Str& text);
    void                            set(size_t index, const Vec3d& position, const Str& text);
    void                            remove(size_t index);
    void                            removeAll();

    static PtrRef<MarkupPartLabels> create(const Vec3d& position, const Str& text, LabelDrawMode drawMode = DRAW_WITHOUT_ZBUFFER_TEST);
    static PtrRef<MarkupPartLabels> createWithMarker(const Vec3d& position, const Str& text, const Color3f& markerColor, LabelDrawMode drawMode = DRAW_WITHOUT_ZBUFFER_TEST);

protected:
    virtual bool                    internal_preRenderUpdatePart(bool forceUpdate, cvf::OpenGLContext* cvfContext, MarkupClippingData* clippingData);
    virtual cvf::Part*              internal_cvfPart();
    virtual cvf::Part*              internal_cvfOverlayPart();
    virtual cvf::Part*              internal_cvfOverlayDetectVisiblePart();

private:
    CEE_PRIVATE_F(MarkupModel);
    CEE_PRIVATE_IMPL(MarkupPartLabels);
    CEE_DISALLOW_COPY_AND_ASSIGN(MarkupPartLabels);
};

} // namespace vis
} // namespace cee
