//##################################################################################################
//
//  Ceetron Desktop Components
//  Component: Visualization
//  
//  --------------------------------------------------------------------------------------------
//  Copyright (C) 2018, Ceetron AS
//  This is UNPUBLISHED PROPRIETARY SOURCE CODE of Ceetron AS. The contents of this file may 
//  not be disclosed to third parties, copied or duplicated in any form, in whole or in part,
//  without the prior written permission of Ceetron AS.
//##################################################################################################

#pragma once

#include "CeeCore/Str.h"
#include "CeeVisualization/Base.h"

namespace cee {
namespace vis {

class OpenGLFixedFunctionsBackend;

typedef unsigned int    GLenum;
typedef unsigned char   GLboolean;
typedef unsigned int    GLbitfield;
typedef signed char     GLbyte;         // 1-byte signed
typedef short           GLshort;        // 2-byte signed
typedef int             GLint;          // 4-byte signed
typedef int             GLsizei;        // 4-byte signed
typedef unsigned char   GLubyte;        // 1-byte unsigned
typedef unsigned short  GLushort;       // 2-byte unsigned
typedef unsigned int    GLuint;         // 4-byte unsigned
typedef float           GLfloat;        // single precision float
typedef float           GLclampf;       // single precision float in [0,1]
typedef double          GLdouble;       // double precision float
typedef double          GLclampd;       // double precision float in [0,1]
typedef void            GLvoid;
typedef char            GLchar;
typedef ptrdiff_t       GLintptr;
typedef ptrdiff_t       GLsizeiptr;



//==================================================================================================
//
// 
//
//==================================================================================================
struct CEE_VIS_EXPORT OpenGLBackendInfo
{
public:
    enum Capability
    {
        FRAMEBUFFER_OBJECT              = 0x00000001,   // Supports framebuffer objects
        TEXTURE_FLOAT                   = 0x00000002,   // Floating-point internal texture formats
        TEXTURE_RG                      = 0x00000004,   // One- and two- component internal texture formats
        TEXTURE_RECTANGLE               = 0x00000008,   // Rectangular textures
        FLAT_INTERPOLATION_QUALIFIER    = 0x00000010,   // Supports the flat interpolation qualifier in GLSL
        BLIT_FRAMEBUFFER                = 0x00000020    // Supports glBlitFramebuffer
    };

public:
    // Capabilities
    unsigned int    oglMajorVersion;
    unsigned int    oglMinorVersion;
    bool            isOpenGLES;
    unsigned int    capabilityFlags;

    // OpenGL info
    Str     oglVersionString;
    Str     vendorString;
    Str     rendererString;
    Str     backendDetailsString;

public:
    OpenGLBackendInfo()
     :  oglMajorVersion(0),
        oglMinorVersion(0),
        isOpenGLES(false),
        capabilityFlags(0)
    {}
};



//==================================================================================================
//
// 
//
//==================================================================================================
class CEE_VIS_EXPORT OpenGLFunctionsBackend
{
public:
    virtual ~OpenGLFunctionsBackend() {}

    virtual bool                            initializeBackend(OpenGLBackendInfo* backendInfo) = 0;

    // Optional; can supply a backend for fixed function OpenGL as well
    virtual OpenGLFixedFunctionsBackend*    fixedFunctionsBackend() { return NULL; }


    // ##############################################################################################################################
    // Baseline OpenGL 2.1
    virtual void        glDisable(GLenum cap) = 0;
    virtual void        glEnable(GLenum cap) = 0;
    virtual GLenum      glGetError() = 0;
    virtual GLboolean   glIsEnabled(GLenum cap) = 0;
    virtual void        glGetIntegerv(GLenum pname, GLint* params) = 0;

    virtual void        glViewport(GLint x, GLint y, GLsizei width, GLsizei height) = 0;
    virtual void        glClearColor(GLclampf red, GLclampf green, GLclampf blue, GLclampf alpha) = 0;
    virtual void        glClearDepth(GLclampd depth) = 0;
    virtual void        glStencilMask(GLuint mask) = 0;
    virtual void        glClearStencil(GLint s) = 0;
    virtual void        glScissor(GLint x, GLint y, GLsizei width, GLsizei height) = 0;
    virtual void        glClear(GLbitfield mask) = 0;

    virtual void        glDrawBuffer(GLenum mode) = 0;
    virtual void        glDrawBuffers(GLsizei n, const GLenum* bufs) = 0;

    virtual void        glVertexAttribPointer(GLuint index, GLint size, GLenum type, GLboolean normalized, GLsizei stride, const GLvoid* pointer) = 0;
    virtual void        glEnableVertexAttribArray(GLuint index) = 0;
    virtual void        glDisableVertexAttribArray(GLuint index) = 0;

    virtual void        glDrawArrays(GLenum mode, GLint first, GLsizei count) = 0;
    virtual void        glDrawElements(GLenum mode, GLsizei count, GLenum type, const GLvoid* indices) = 0;
    virtual void        glDrawRangeElements(GLenum mode, GLuint start, GLuint end, GLsizei count, GLenum type, const GLvoid* indices) = 0;

    virtual void        glUniform1f(GLint location, GLfloat v0) = 0;
    virtual void        glUniform1i(GLint location, GLint v0) = 0;
    virtual void        glUniform1fv(GLint location, GLsizei count, const GLfloat* value) = 0;
    virtual void        glUniform1iv(GLint location, GLsizei count, const GLint* value) = 0;
    virtual void        glUniform2fv(GLint location, GLsizei count, const GLfloat* value) = 0;
    virtual void        glUniform3fv(GLint location, GLsizei count, const GLfloat* value) = 0;
    virtual void        glUniform4fv(GLint location, GLsizei count, const GLfloat* value) = 0;
    virtual void        glUniformMatrix3fv(GLint location, GLsizei count, GLboolean transpose, const GLfloat* value) = 0;
    virtual void        glUniformMatrix4fv(GLint location, GLsizei count, GLboolean transpose, const GLfloat* value) = 0;

    virtual void        glBindBuffer(GLenum target, GLuint buffer) = 0;
    virtual void        glGenBuffers(GLsizei n, GLuint* buffers) = 0;
    virtual void        glDeleteBuffers(GLsizei n, const GLuint* buffers) = 0;
    virtual void        glBufferData(GLenum target, GLsizeiptr size, const GLvoid* data, GLenum usage) = 0;

    virtual GLuint      glCreateShader(GLenum shaderType) = 0;
    virtual void        glDeleteShader(GLuint shader) = 0;
    virtual void        glShaderSource(GLuint shader, GLsizei count, const GLchar** string, const GLint* length) = 0;
    virtual void        glCompileShader(GLuint shader) = 0;
    virtual GLboolean   glIsShader(GLuint shader) = 0;
    virtual void        glGetShaderInfoLog(GLuint shader, GLsizei maxLength, GLsizei* length, GLchar* infoLog) = 0;
    virtual void        glGetShaderiv(GLuint shader, GLenum pname, GLint* params) = 0;

    virtual GLuint      glCreateProgram() = 0;
    virtual void        glDeleteProgram(GLuint program) = 0;
    virtual void        glAttachShader(GLuint program, GLuint shader) = 0;
    virtual void        glLinkProgram(GLuint program) = 0;
    virtual void        glValidateProgram(GLuint program) = 0;
    virtual GLboolean   glIsProgram(GLuint program) = 0;
    virtual void        glUseProgram(GLuint program) = 0;
    virtual void        glGetProgramInfoLog(GLuint program, GLsizei maxLength, GLsizei* length, GLchar* infoLog) = 0;
    virtual void        glGetProgramiv(GLuint program, GLenum pname, GLint* params) = 0;
    virtual GLint       glGetUniformLocation(GLuint program, const GLchar* name) = 0;
    virtual void        glGetActiveUniform(GLuint program, GLuint index, GLsizei bufSize, GLsizei* length, GLint* size, GLenum* type, GLchar* name) = 0;
    virtual GLint       glGetAttribLocation(GLuint program, const GLchar* name) = 0;
    virtual void        glBindAttribLocation(GLuint program, GLuint index, const GLchar* name) = 0;

    virtual void        glGenTextures(GLsizei n, GLuint* textures) = 0;
    virtual void        glDeleteTextures(GLsizei n, const GLuint* textures) = 0;
    virtual void        glTexImage2D(GLenum target, GLint level, GLint internalFormat, GLsizei width, GLsizei height, GLint border, GLenum format, GLenum type, const GLvoid* data) = 0;
    virtual void        glGenerateMipmap(GLenum target) = 0;
    virtual void        glTexParameteri(GLenum target, GLenum pname, GLint param) = 0;
    virtual void        glTexParameterfv(GLenum target, GLenum pname, const GLfloat* params) = 0;
    virtual void        glBindTexture(GLenum target, GLuint texture) = 0;
    virtual void        glActiveTexture(GLenum texture) = 0;

    virtual void        glReadPixels(GLint x, GLint y, GLsizei width, GLsizei height, GLenum format, GLenum type, GLvoid* data) = 0;
    virtual void        glPixelStorei(GLenum pname, GLint param) = 0;

    virtual void        glBlendFunc(GLenum sfactor, GLenum dfactor) = 0;
    virtual void        glBlendFuncSeparate(GLenum srcRGB, GLenum dstRGB, GLenum srcAlpha, GLenum dstAlpha) = 0;
    virtual void        glBlendEquation(GLenum mode) = 0;
    virtual void        glBlendEquationSeparate(GLenum modeRGB, GLenum modeAlpha) = 0;
    virtual void        glBlendColor(GLfloat red, GLfloat green, GLfloat blue, GLfloat alpha) = 0;

    virtual void        glDepthFunc(GLenum func) = 0;
    virtual void        glDepthMask(GLboolean flag) = 0;
    virtual void        glColorMask(GLboolean red, GLboolean green, GLboolean blue, GLboolean alpha) = 0;
    virtual void        glCullFace(GLenum mode) = 0;
    virtual void        glFrontFace(GLenum mode) = 0;
    virtual void        glLineWidth(GLfloat width) = 0;
    virtual void        glPointSize(GLfloat size) = 0;
    virtual void        glPointParameteri(GLenum pname, GLint param) = 0;
    virtual void        glPolygonMode(GLenum face, GLenum mode) = 0;
    virtual void        glPolygonOffset(GLfloat factor, GLfloat units) = 0;
    virtual void        glStencilFunc(GLenum func, GLint ref, GLuint mask) = 0;
    virtual void        glStencilOp(GLenum sfail, GLenum dpfail, GLenum dppass) = 0;


    // ##############################################################################################################################
    // OpenGL3+ and extensions
    virtual void        glGenRenderbuffers(GLsizei n, GLuint* renderbuffers) = 0;
    virtual void        glDeleteRenderbuffers(GLsizei n, GLuint* renderbuffers) = 0;
    virtual void        glBindRenderbuffer(GLenum target, GLuint renderbuffer) = 0;
    virtual void        glRenderbufferStorage(GLenum target, GLenum internalformat, GLsizei width, GLsizei height) = 0;
    virtual void        glRenderbufferStorageMultisample(GLenum target, GLsizei samples, GLenum internalformat, GLsizei width, GLsizei height) = 0;

    virtual void        glGenFramebuffers(GLsizei n, GLuint* ids) = 0;
    virtual void        glDeleteFramebuffers(GLsizei n, GLuint* framebuffers) = 0;
    virtual void        glFramebufferRenderbuffer(GLenum target, GLenum attachment, GLenum renderbuffertarget, GLuint renderbuffer) = 0;
    virtual void        glFramebufferTexture2D(GLenum target, GLenum attachment, GLenum textarget, GLuint texture, GLint level) = 0;
    virtual void        glBindFramebuffer(GLenum target, GLuint framebuffer) = 0;
    virtual GLenum      glCheckFramebufferStatus(GLenum target) = 0;
    virtual void        glBlitFramebuffer(GLint srcX0, GLint srcY0, GLint srcX1, GLint srcY1, GLint dstX0, GLint dstY0, GLint dstX1, GLint dstY1, GLbitfield mask, GLenum filter) = 0;

    virtual void        glVertexAttribIPointer(GLuint index, GLint size, GLenum type, GLsizei stride, const GLvoid* pointer) = 0;

    virtual void        glTexImage2DMultisample(GLenum target, GLsizei samples, GLint internalformat, GLsizei width, GLsizei height, GLboolean fixedsamplelocations) = 0;
};



//==================================================================================================
//
// 
//
//==================================================================================================
class OpenGLFixedFunctionsBackend
{
public:
    virtual ~OpenGLFixedFunctionsBackend() {}

    virtual void        glBegin(GLenum mode) = 0;
    virtual void        glEnd() = 0;
    virtual void        glVertex3f(GLfloat x, GLfloat y, GLfloat z) = 0;
    virtual void        glVertex3fv(const GLfloat* v) = 0;
    virtual void        glNormal3fv(const GLfloat* v) = 0;
    virtual void        glColor3ubv(const GLubyte* v) = 0;
    virtual void        glColor3f(GLfloat red, GLfloat green, GLfloat blue) = 0;
    virtual void        glColor3fv(const GLfloat* v) = 0;
    virtual void        glColor4f(GLfloat red, GLfloat green, GLfloat blue, GLfloat alpha) = 0;
    virtual void        glColor4fv(const GLfloat* v) = 0;
    virtual void        glTexCoord2fv(const GLfloat* v) = 0;

    virtual void        glLightModeli(GLenum pname, GLint param) = 0;
    virtual void        glLightModelfv(GLenum pname, const GLfloat* params) = 0;
    virtual void        glMaterialf(GLenum face, GLenum pname, GLfloat param) = 0;
    virtual void        glMaterialfv(GLenum face, GLenum pname, const GLfloat* params) = 0;
    virtual void        glTexEnvi(GLenum target, GLenum pname, GLint param) = 0;
    virtual void        glTexGeniv(GLenum coord, GLenum pname, const GLint* params) = 0;
    virtual void        glClipPlane(GLenum plane, const GLdouble* equation) = 0;
    virtual void        glShadeModel(GLenum mode) = 0;
    virtual void        glLightfv(GLenum light, GLenum pname, const GLfloat* params) = 0;
    virtual void        glLightf(GLenum light, GLenum pname, GLfloat param) = 0;

    virtual void        glVertexPointer(GLint size, GLenum type, GLsizei stride, const GLvoid* pointer) = 0;
    virtual void        glNormalPointer(GLenum type, GLsizei stride, const GLvoid* pointer) = 0;
    virtual void        glTexCoordPointer(GLint size, GLenum type, GLsizei stride, const GLvoid* pointer) = 0;
    virtual void        glColorPointer(GLint size, GLenum type, GLsizei stride, const GLvoid* pointer) = 0;
    virtual void        glEnableClientState(GLenum cap) = 0;
    virtual void        glDisableClientState(GLenum cap) = 0;

    virtual void        glColorMaterial(GLenum face, GLenum mode) = 0;
    virtual void        glLineStipple(GLint factor, GLushort pattern) = 0;

    virtual void        glMatrixMode(GLenum mode) = 0;
    virtual void        glLoadIdentity() = 0;
    virtual void        glLoadMatrixf(const GLfloat* m) = 0;
    virtual void        glLoadMatrixd(const GLdouble* m) = 0;
    virtual void        glMultMatrixf(const GLfloat* m) = 0;
    virtual void        glPushMatrix() = 0;
    virtual void        glPopMatrix() = 0;
    virtual void        glOrtho(GLdouble left, GLdouble right, GLdouble bottom, GLdouble top, GLdouble nearVal, GLdouble farVal) = 0;

    virtual void        glPushAttrib(GLbitfield mask) = 0;
    virtual void        glPopAttrib() = 0;
    virtual void        glPushClientAttrib(GLbitfield mask) = 0;
    virtual void        glPopClientAttrib() = 0;

    // Uses GLU. We should get rid of this one, but slightly complex to replace
    virtual GLint       gluBuild2DMipmaps(GLenum target, GLint internalFormat, GLsizei width, GLsizei height, GLenum format, GLenum type, const void* data) = 0;
};


} // namespace vis
} // namespace cee
